/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"

char def_dev_name[] = "mlx5_0";
struct test_info t_info = {0};
struct test_struct *testp;

static struct test_struct tests[] = {
	INIT_TEST(all, test_all, false),
	INIT_TEST(basic, test_basic, true),
	INIT_TEST(range, test_range, true),
	INIT_TEST(resize, test_resize, true),
	INIT_TEST(resource, test_resource, true),
	INIT_TEST(resource_aso_meter, test_resource_aso_meter, true),
	INIT_TEST(resource_argument, test_resource_argument, true),
	INIT_TEST(resource_argument_bulk, test_resource_argument_block, true),
	INIT_TEST(resource_aso_counter, test_resource_aso_counter, true),
	INIT_TEST(resource_block_alloc, test_resource_block_alloc, true),
	INIT_TEST(compare, test_compare, true),
	INIT_TEST(conv_func, test_conv_func, true),
	//INIT_TEST(pool, test_pool, true),
	INIT_TEST(multi_threaded, test_multi_threaded, true),
	INIT_TEST(dest, test_dest, true),
	INIT_TEST(actions, test_actions, true),
	INIT_TEST(metric, test_metric, true),
	INIT_TEST(add_rm_hdr, test_add_rm_hdr, true),
	INIT_TEST(parser, test_parser, true),
	INIT_TEST(inline_action, test_inline_action, true),
	INIT_TEST(isolated_matcher, test_isolated_matcher, true),
	INIT_TEST(ct, test_ct, true),
	INIT_TEST(ct_multi, test_multi_threaded_ct, false),
};

static int run_test(struct ibv_context *ibv_ctx, struct test_struct *test)
{
	int ret;

	ret = test->run(ibv_ctx, &t_info);
	if (ret)
		printf("Test %s: Failed\n", test->name);
	else
		printf("Test %s: Passed\n", test->name);

	return ret;
}

int test_all(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	int i, num_of_tests = 0, ret = 0, count = 0;

	for (i = 0; i < ARRAY_SIZE(tests); i++) {
		if (!tests[i].run_all)
			continue;

		printf("\n*** Running %s test ***\n", tests[i].name);
		num_of_tests++;
		ret = run_test(ibv_ctx, &tests[i]);
		if (!ret)
			count++;
	}

	printf("\n***** Summary *****\n");
	printf("%u / %u Tests passed\n", count, num_of_tests);
	printf("*******************\n");

	return count != num_of_tests;
}

static bool test_selection(char *test_name)
{
	bool found = false;
	size_t i;

	for (i = 0; i < sizeof(tests)/ sizeof(tests[0]); i++) {
		if (strlen(test_name) != strlen(tests[i].name))
			continue;

		if (memcmp(tests[i].name, test_name, strlen(tests[i].name)))
			continue;

		found = true;
		testp = &tests[i];
		break;
	}

	return found;
}

int min(int a, int b)
{
	return a < b ? a : b;
}

static int args_parser(int argc, char **argv, char **dev_name)
{
	bool test_found;
	char *endptr;
	int i;

	/* Set default values in case user did not pass any arguments */
	*dev_name = def_dev_name;
	testp = &tests[0];

	for (i = 0; i < argc - 1; i++) {
		if (!memcmp("-d", argv[i], 2)) {
			*dev_name = argv[i + 1];
		} else if (!memcmp("-t", argv[i], 2)) {
			test_found = test_selection(argv[i + 1]);
			if (!test_found) {
				TEST_LOG("Test: %s not found", argv[i + 1]);
				errno = EINVAL;
				return -EINVAL;
			}
		} else if (!memcmp("-c", argv[i], 2)) {
			t_info.cpu_mask = strtoll(argv[i + 1], &endptr, 0);
		}
	}

	printf("Device selected: %s\n", *dev_name);
	printf("Test selected: %s\n", testp->name);

	return 0;
}

int main(int argc, char **argv)
{
	struct mlx5dv_context_attr dv_ctx_attr = {};
	struct ibv_device **dev_list = NULL;
	struct ibv_context *ibv_ctx = NULL;
	struct ibv_device *dev = NULL;
	char *dev_name;
	int i, ret;

	if (args_parser(argc, argv, &dev_name))
		return -errno;

	dev_list = ibv_get_device_list(NULL);
	if (!dev_list) {
		TEST_LOG("Failed to get IB devices list");
		return -errno;
	}

	for (i = 0; dev_list[i]; i++) {
		if (!strcmp(ibv_get_device_name(dev_list[i]), dev_name)) {
			dev = dev_list[i];
			break;
		}
	}

	if (!dev) {
		TEST_LOG("IB device %s not found", dev_name);
		ret = -ENOENT;
		goto free_dev_list;
	}

	dv_ctx_attr.flags = 1;//MLX5DV_CONTEXT_FLAGS_DEVX; TODO
	ibv_ctx = mlx5dv_open_device(dev, &dv_ctx_attr);
	if (!ibv_ctx) {
		TEST_LOG("Couldn't get context for %s", ibv_get_device_name(dev));
		ret = -EINVAL;
		goto free_dev_list;
	}

	/* errno gets set to ENOSPC inside a function called
	 * by mlx5dv_open_device.
	 */
	errno = 0;

	ret = run_test(ibv_ctx, testp);
	if (ret)
		goto close_ib_dev;

	if (ibv_ctx)
		ibv_close_device(ibv_ctx);
	if (dev_list)
		ibv_free_device_list(dev_list);

	return 0;

close_ib_dev:
	if (ibv_ctx)
		ibv_close_device(ibv_ctx);
free_dev_list:
	if (dev_list)
		ibv_free_device_list(dev_list);

	return ret;
}
