/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#ifndef _HWS_TEST_H_
#define _HWS_TEST_H_

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <stdbool.h>
#include <assert.h>
#include <stddef.h>
#include <infiniband/verbs.h>
#include <infiniband/mlx5dv.h>
#include "../mlx5dv_hws.h"
#include "../hws_mlx5_ifc.h"

/***** Defines & Macros *****/
#define INIT_TEST(str, func, include_in_run_all) \
	{ .name = #str, .run = func, .run_all = include_in_run_all}

#define TEST_LOG(format, arg...) \
	mlx5dv_hws_test_log_print(format, ##arg)

#define mlx5dv_hws_test_log_print(format, arg...) \
do { \
	int tmp = errno; \
	fprintf(stderr, "MLX5DV_HWS_TEST[%s:%d]: " format "\n", __func__, __LINE__, ##arg); \
	errno = tmp; \
} while (0)

#define CALC_BIT_OFF(typ, fld) ((unsigned int)(uintptr_t) (&(__devx_nullp(typ)->fld)))

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
#endif

#define IS_BIT_SET(_value, _bit) ((_value) & (1ULL << (_bit)))

#define BASIC_ITEM_NUM_OF_FIELDS 3

/***** Structs *****/
struct test_info {
	uint64_t cpu_mask;
};

struct test_struct {
	int (*run)(struct ibv_context *, struct test_info *);
	const char *name;
	bool run_all;
};

struct mlx5_ifc_my_basic_item_bits {
	uint8_t sip[0x20];
	uint8_t dip[0x20];
	uint8_t sport[0x10];
	uint8_t dport[0x10];
	uint8_t version[0x4];
	uint8_t reserved[0x1C];/* Some of the DEVX macros divide by 8 */
	uint8_t sampler[0x20];
};

/***** Common functions headers *****/

int dump_all(struct mlx5dv_hws_context *ctx);

int create_basic_item(struct mlx5dv_hws_item *item);

void destroy_basic_item(struct mlx5dv_hws_item *item);

int process_gettime(struct timespec *tp);

int thread_gettime(struct timespec *t);

double timediff(const struct timespec *start, const struct timespec *end);

int set_cpu_cores(cpu_set_t *set, uint64_t core_bits);

int min(int a, int b);

/***** Test functions headers *****/

/**
 * @brief Run all tests.
 *
 * This will run all tests under tests directory.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_all(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Create context and table.
 *
 * This is a basic test which creates only context and table.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_basic(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Range match test
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_range(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Compare match test
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_compare(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Conversion function match test
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_conv_func(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Resize from small to large matcher
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_resize(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Check pool creation and operation.
 *
 * Will create a one resource with fixed sized objects pool,
 * and will get and put chunks and see if it operates in correct
 * ordering.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_pool(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Test resource management operations.
 *
 * This function performs a series of tests on the resource management layer,
 * including resource allocation, querying, configuration, and multithreaded
 * operations. It ensures that the resource management functions operate
 * correctly.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non-zero otherwise with errno/fs_errno set.
 */
int test_resource(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Insert rules via multiple threads.
 *
 * This test inserts rules via multiple threads.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_multi_threaded(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Check creation and destruction of destination actions.
 *
 * @param ibv_ctx
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_dest(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Check creation and destruction of actions.
 *       Including action-templates etc.
 *
 * @param ctx
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_actions(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Test ASO meter resource.
 *
 * This function configures and allocates an ASO meter resource.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non-zero otherwise with errno/fs_errno set.
 */
int test_resource_aso_meter(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Test argument resource.
 *
 * This function configures and allocates an argument resource.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non-zero otherwise with errno/fs_errno set.
 */
int test_resource_argument(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Test argument resource with block write.
 *
 * This function configures and allocates an argument resource, then tests allocating
 * and writing to multiple consecutive offsets in a single operation.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non-zero otherwise with errno/fs_errno set.
 */
int test_resource_argument_block(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Test ASO counter resource.
 *
 * This function configures and allocates an ASO counter resource,
 * and then uses this counter in HWS rules creation.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non-zero otherwise with errno/fs_errno set.
 */
int test_resource_aso_counter(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Check metric query functions
 *
 * This is a basic metric test which creates a table and a matcher
 * and checks if the metric info was set correctly.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */

int test_metric(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Test Add Header, Remove Header action.
 *
 * This function creates insert header action with argument and rule creation,
 * remove the header action with the rule creation.
 * Verifying that creation and polling succeed.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non-zero otherwise with errno/fs_errno set.
 */
int test_add_rm_hdr(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Test parse graph creation and rule insertion using sampler.
 *
 * The test defines a custom IP-like protocol parser node, configures
 * transitions to TCP and UDP nodes based on a "next header" field, sets up a
 * sampler for the first 32 bits of the custom header, and finally binds the
 * graph. Then there is a use creation using a match on the graph sampler
 * field.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_parser(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Test resource offset block allocation.
 *
 * This test uses various patterns of allocating resource offsets to ensure
 * correctness and performance. We allocate uniform blocks and ensure we can use
 * all of them, then in a different test we allocate random sized blocks and
 * print usage information (since buddy allocation does not guarantee 100%
 * usage).
 *
 * The test also includes a benchmark that is equipped to compare buddy offset
 * allocation to bitmap on a single thread.
 */
int test_resource_block_alloc(struct ibv_context *ibv_ctx,
			      struct test_info *t_info);

/**
 * @brief Test inline action creation.
 *
 * The test uses the mlx5dv_hws_action_create_inline API call to create all
 * supported inline actions, SET, ADD, ADD_FIELD, COPY, and INSERT.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_inline_action(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Test isolated matcher functionality
 *
 * This test tests the isolated matcher functionality.
 *
 * @param[in] ibv_ctx - ibv context supporting devx.
 * @return Zero on success, non zero otherwise with errno/fs_errno set.
 */
int test_isolated_matcher(struct ibv_context *ibv_ctx, struct test_info *t_info);

/**
 * @brief Test CT operations
 *
 * The test crates all the relevant objects for CT and does the CT operations:
 * Mach CT fields, create rule and destroy.
 * @param ibv_ctx
 * @param t_info
 * @return int
 */
int test_ct(struct ibv_context *ibv_ctx, struct test_info *t_info);

int test_multi_threaded_ct(struct ibv_context *ibv_ctx, struct test_info *t_info);
#endif /* _HWS_TEST_H_ */
