/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"
#include "hws_test_resource.h"

#include "../mlx5dv_hws.h"

#define NUM_OF_FIELDS 2
#define NUM_OF_QUEUES 16
#define ROW_LOG 20
#define NUM_OF_RULES (1 << ROW_LOG)
#define QUEUE_SIZE 256
#define BURST_TH 32

struct mlx5_ifc_my_item_bits {
	uint8_t sip[0x20];
	uint8_t dip[0x20];
	uint8_t sport[0x10];
	uint8_t dport[0x10];
};

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			printf("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

static int create_item(struct mlx5dv_hws_item *item)
{
	struct mlx5dv_hws_field *fields;
	void *mask, *match, *buff;

	fields = calloc(NUM_OF_FIELDS, sizeof(struct mlx5dv_hws_field));
	if (!fields) {
		TEST_LOG("Failed to allocate fields");
		return ENOMEM;
	}

	buff = calloc(2, DEVX_ST_SZ_BYTES(my_item));
	if (!buff) {
		TEST_LOG("Failed to allocate mask and match");
		free(fields);
		return ENOMEM;
	}

	match = buff;
	mask = buff + DEVX_ST_SZ_BYTES(my_item);

	item->num_of_fields = NUM_OF_FIELDS;
	item->fields = fields;
	item->data.mask = mask;
	item->data.match = match;

	fields[0].fname = MLX5DV_HWS_FNAME_IPV4_SRC;
	fields[0].header = MLX5DV_HWS_HEADER_OUTER;
	fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[0].value.bit_off = CALC_BIT_OFF(my_item, sip);
	DEVX_SET(my_item, mask, sip, -1);
	DEVX_SET(my_item, match, sip, 0);

	fields[1].fname = MLX5DV_HWS_FNAME_IPV4_DST;
	fields[1].header = MLX5DV_HWS_HEADER_OUTER;
	fields[1].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[1].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[1].value.bit_off = CALC_BIT_OFF(my_item, dip);
	DEVX_SET(my_item, mask, dip, -1);
	DEVX_SET(my_item, match, dip, 0x05060708);

	return 0;
}

static void destroy_item(struct mlx5dv_hws_item *item)
{
	free((void *)item->data.match);
	free(item->fields);
}


int test_nat64_action(struct mlx5dv_hws_context *ctx)
{
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_at *at, *at_root;
	struct mlx5dv_hws_at *all_at[3];
	struct mlx5dv_hws_mt *mt, *r_mt;
	struct mlx5dv_hws_action *actions[7];
	struct mlx5dv_hws_matcher *matcher, *root_matcher;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_action *tag;
	struct mlx5dv_hws_action *drop;
	struct mlx5dv_hws_action *dest_ft;
	struct mlx5dv_hws_action *dest_miss;
	struct mlx5dv_hws_action_nat64_attr nat64_attr;
	struct mlx5dv_hws_action *nat64;
	struct mlx5dv_hws_table *tbl, *root_tbl;
	struct mlx5dv_hws_at_attr attr_at = {0};
	uint8_t registers[4] = {82, 83, 84};
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	struct mlx5dv_hws_action_data rule_actions[5] = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_item root_item = {0};
	int ret;
	uint32_t pending_rules = 0;
	uint32_t miss_count = 0;
	uint16_t queue_id = 0;
	uint32_t hw_bp = 0;


	dest_miss = mlx5dv_hws_action_create_dest_default_miss(ctx, MLX5DV_HWS_ACTION_FLAG_NIC_TX);
	if (!dest_miss) {
		TEST_LOG("Failed to create dest_miss action");
		return -1;
	}

	/* Create Root Table */
	tbl_attr.level = 0;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_TX;
	root_tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!root_tbl) {
		TEST_LOG("Failed to create root table");
		return -1;
	}
	printf("Root Table created\n");

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_TX;

	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		return -1;
	}

	printf("Table created\n");


	mt_attr.comp_mask = 0;
	r_mt = mlx5dv_hws_match_template_create(ctx, NULL, 0, &mt_attr);
	if (!r_mt) {
		TEST_LOG("Failed to create root match template");
		return 1;
	}

	at_root = mlx5dv_hws_action_template_create(NULL, 0, &attr_at);
	if (!at_root) {
		TEST_LOG("Failed to create at root");
		return -1;
	}

	printf("Root Match template created\n");

	dest_ft = mlx5dv_hws_action_create_dest_table(ctx, tbl,  MLX5DV_HWS_ACTION_FLAG_ROOT | MLX5DV_HWS_ACTION_FLAG_NIC_TX);
	if (!dest_ft) {
		TEST_LOG("Failed to create table action");
		ret = -1;
		return ret;
	}

	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = 4;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;
	root_matcher = mlx5dv_hws_matcher_create(root_tbl, &r_mt, 1, &at_root, 1, &matcher_attr);
	if (!root_matcher) {
		TEST_LOG("Failed to create root matcher");
			return -1;
	}

	printf("Root Matcher created\n");

	rules = calloc(NUM_OF_RULES, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		return -1;
	}

	rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules);
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.queue_id = queue_id;

	ret = mlx5dv_hws_rule_create(root_matcher, 0, &root_item.data, 0,
				     rule_actions, dest_ft, &rule_attr, rule);
	if (ret)
		TEST_LOG("Failed to create root rule");

	pending_rules = 1;
	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, 1, &miss_count, &hw_bp, true);
	pending_rules = 0;

	drop = mlx5dv_hws_action_create_dest_drop(ctx, MLX5DV_HWS_ACTION_FLAG_NIC_TX);
	if (!drop) {
		TEST_LOG("Failed to create drop action");
		return -1;
	}
	printf("Drop created\n");

	tag = mlx5dv_hws_action_create_tag(ctx, MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!tag) {
		TEST_LOG("Failed to create tag action");
		return -1;
	}
	printf("action tag created\n");

	nat64_attr.num_of_registers = 3;
	nat64_attr.flags = (MLX5DV_HWS_ACTION_NAT64_V4_TO_V6 |
			    MLX5DV_HWS_ACTION_NAT64_BACKUP_ADDR);
	nat64_attr.registers = registers;

	nat64 = mlx5dv_hws_action_create_nat64(ctx, &nat64_attr, MLX5DV_HWS_ACTION_FLAG_NIC_TX);
	if (!nat64) {
		TEST_LOG("Failed to create nat64 action");
		return -1;
	}

	printf("action nat64 created\n");

	actions[0] = nat64;

	/* Create action template */
	at = mlx5dv_hws_action_template_create(actions, 1, &attr_at);
	if (!at) {
		TEST_LOG("Failed to create at");
		return -1;
	}

	printf("action template created\n");

	ret = create_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		return -1;
	}

	printf("Item created\n");

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, NULL, 0, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		return -1;
	}

	printf("Match template created\n");

	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = 8;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;

	all_at[0] = at;
	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, all_at, 1, &matcher_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher");
		return -1;
	}

	printf("Matcher created\n");

	/* insert some rules */
	rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules);
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.queue_id = 0;


	dump_all(ctx);

	rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules + 1 *
					mlx5dv_hws_rule_get_handle_size());
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.rule_idx = 0;

	ret = mlx5dv_hws_rule_create(matcher, 0, &item.data, 0,
				     rule_actions, dest_miss, &rule_attr, rule);
	if (ret)
		TEST_LOG("Failed to create rule");
	pending_rules = 1;

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, 1,
			      &miss_count, &hw_bp, true);

	TEST_LOG("Created NAT64 rule");

	ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
	if (ret)
		TEST_LOG("Failed to destroy rule");

	pending_rules = 1;

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, 1,
			      &miss_count, &hw_bp, true);

	TEST_LOG("Destroyed NAT64 rule");

	free(rules);

	mlx5dv_hws_action_destroy(drop);
	mlx5dv_hws_action_destroy(tag);
	mlx5dv_hws_action_destroy(nat64);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_match_template_destroy(mt);
	mlx5dv_hws_matcher_destroy(matcher);
	return 0;
}

int test_basic_actions(struct mlx5dv_hws_context *ctx)
{
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	uint8_t data[20] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0xa, 0xb, 0xc, 0xd, 0x1, 0x2, 3, 4};
	struct aso_cseg_input aso_cseg_in = {0};
	struct aso_wqe aso_wqe = {0};
	uint16_t wqe_len;
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_action_hdr_data mhd = {0};
	struct mlx5dv_hws_resource *all_resources[4];
	struct mlx5dv_hws_resource *resource, *resource2;
	struct mlx5dv_hws_resource *resource_countr, *resource_countr2;
	struct mlx5dv_hws_resource *resource_aso_ct;
	struct mlx5dv_hws_resource_attr res_attr = {0};
	struct mlx5dv_hws_action_data rule_actions[5] = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_at *at, *at2, *at_root;
	struct mlx5dv_hws_at *all_at[3];
	struct mlx5dv_hws_mt *mt, *r_mt;
	struct mlx5dv_hws_action *actions[7];
	struct mlx5dv_hws_matcher *matcher, *root_matcher, *matcher2;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_item root_item = {0};
	struct mlx5dv_hws_action *push;
	struct mlx5dv_hws_action *tag, *r_tag;
	struct mlx5dv_hws_action *pop, *pop2;
	struct mlx5dv_hws_action *reformat_tl2_l2;
	struct mlx5dv_hws_action *reformat_l2_tl2;
	struct mlx5dv_hws_action *reformat_l2_tl3;
	struct mlx5dv_hws_action *reformat_tl3_l2, *r_reformat_tl3_l2;
	struct mlx5dv_hws_action *aso_counter;
	struct mlx5dv_hws_action *aso_ct;
	struct mlx5dv_hws_action *drop;
	struct mlx5dv_hws_table *tbl, *root_tbl;
	struct mlx5dv_hws_at_attr attr_at = {0};
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	uint32_t pending_rules = 0;
	uint32_t miss_count = 0;
	uint16_t queue_id = 0;
	uint32_t hw_bp = 0;
	int ret = -1, i;

	drop = mlx5dv_hws_action_create_dest_drop(ctx, 1);
	if (!drop) {
		TEST_LOG("Failed to create drop action");
		return -1;
	}

	printf("action goto drop created\n");

	/* Create Root Table */
	tbl_attr.level = 0;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	root_tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!root_tbl) {
		TEST_LOG("Failed to create root table");
		return -1;
	}

	printf("Root Table created\n");

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		return -1;
	}

	printf("Table created\n");

	tag = mlx5dv_hws_action_create_tag(ctx, 1);
	if (!tag) {
		TEST_LOG("Failed to create tag action");
		return -1;
	}

	printf("action tag created\n");

	r_tag = mlx5dv_hws_action_create_tag(ctx, MLX5DV_HWS_ACTION_FLAG_NIC_RX |
						  MLX5DV_HWS_ACTION_FLAG_ROOT);
	if (!r_tag) {
		TEST_LOG("Failed to create root tag action");
		return -1;
	}

	printf("action root tag created\n");

	pop = mlx5dv_hws_action_create_pop_vlan(ctx, 1);
	if (!pop) {
		TEST_LOG("Failed to create pop action");
		goto clean_tag;
	}

	printf("action pop created\n");

	pop2 = mlx5dv_hws_action_create_pop_vlan(ctx, 1);
	if (!pop2) {
		TEST_LOG("Failed to create pop2 action");
		goto clean_tag;
	}

	printf("action pop2 created\n");

	push = mlx5dv_hws_action_create_push_vlan(ctx, 1);
	if (!push) {
		TEST_LOG("Failed to create push action");
		goto clean_pop;
	}

	printf("action push created\n");

	res_attr.bulk_log_size  = 10;
	res_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B;
	resource = mlx5dv_hws_resource_alloc(ctx, &res_attr);
	if (!resource) {
		TEST_LOG("Failed to create resource ARG!!");
		goto clean_push;
	}
	res_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_128B;
	resource2 = mlx5dv_hws_resource_alloc(ctx, &res_attr);
	if (!resource) {
		TEST_LOG("Failed to create resource 2 ARG!!");
		return -1;
	}

	printf("ARG resource created\n");

	res_attr.bulk_log_size  = 10;
	res_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER;
	resource_countr = mlx5dv_hws_resource_alloc(ctx, &res_attr);
	if (!resource_countr) {
		TEST_LOG("Failed to create resource CTR!!");
		goto clean_push;
	}

	printf("resource CTR created\n");
	res_attr.bulk_log_size  = 12;
	res_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER;
	resource_countr2 = mlx5dv_hws_resource_alloc(ctx, &res_attr);
	if (!resource_countr2) {
		TEST_LOG("Failed to create resource CTR 2!!");
		return -1;
	}
	printf("resource CTR 2 created\n");

	all_resources[0] = resource_countr;
	all_resources[1] = resource_countr2;
	aso_counter = mlx5dv_hws_action_create_counter(ctx, all_resources, 2, 1);
	if (!aso_counter) {
		TEST_LOG("Failed to create resource CTR!!");
		return -1;
	}

	res_attr.bulk_log_size = 10;
	res_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK;
	resource_aso_ct = mlx5dv_hws_resource_alloc(ctx, &res_attr);
	if (!resource_aso_ct) {
		TEST_LOG("Failed to create resource ASO CT!!");
		return -1;
	}

	printf("resource ASO CT created\n");
	all_resources[0] = resource_aso_ct;
	aso_ct = mlx5dv_hws_action_create_aso_ct(ctx, all_resources, 1, 1, 1);
	if (!aso_ct) {
		TEST_LOG("Failed to create CT action!!");
		return -1;
	}

	reformat_tl2_l2 = mlx5dv_hws_action_create_reformat_tnl_l2_to_l2(ctx, 1);
	if (!reformat_tl2_l2) {
		TEST_LOG("Failed to create reformat_tl2_l2 action!!");
		return -1;
	}

	mhd.sz = 14;
	mhd.hdr_data = data;
	all_resources[0] = resource;
	all_resources[1] = resource2;
	reformat_l2_tl2 = mlx5dv_hws_action_create_reformat_l2_to_tnl_l2(ctx, all_resources,
									 2, &mhd, 1);
	if (!reformat_l2_tl2) {
		TEST_LOG("Failed to create reformat_l2_tl2 action!!");
		return -1;
	}

	reformat_l2_tl3 = mlx5dv_hws_action_create_reformat_l2_to_tnl_l3(ctx, all_resources,
									 2, &mhd, 1);
	if (!reformat_l2_tl3) {
		TEST_LOG("Failed to create reformat_l2_tl3 action!!");
		return -1;
	}

	reformat_tl3_l2 = mlx5dv_hws_action_create_reformat_tnl_l3_to_l2(ctx, all_resources,
									 2, &mhd, 1);
	if (!reformat_tl3_l2) {
		TEST_LOG("Failed to create reformat_tl3_l2 action!!");
		return -1;
	}

	r_reformat_tl3_l2 =
		mlx5dv_hws_action_create_reformat_tnl_l3_to_l2(ctx, NULL, 0, &mhd,
							       MLX5DV_HWS_ACTION_FLAG_NIC_RX |
							       MLX5DV_HWS_ACTION_FLAG_ROOT);
	if (!r_reformat_tl3_l2) {
		TEST_LOG("Failed to create r_reformat_tl3_l2 action!!");
		return -1;
	}

	actions[0] = tag;
	actions[1] = reformat_l2_tl3;//reformat_tl3_l2;
	actions[2] = pop;
	actions[3] = aso_counter;

	/* Create action template */
	at = mlx5dv_hws_action_template_create(actions, 4, &attr_at);
	if (!at) {
		TEST_LOG("Failed to create at");
		goto clean_push;
	}

	printf("action template created\n");

	actions[0] = r_tag;
	actions[0] = r_reformat_tl3_l2;
	at_root = mlx5dv_hws_action_template_create(actions, 1, &attr_at);
	if (!at) {
		TEST_LOG("Failed to create at root");
		goto clean_push;
	}

	printf("action template created\n");

	actions[0] = tag;
	actions[1] = reformat_tl3_l2;
	actions[2] = push;
	actions[3] = aso_counter;
	actions[4] = aso_ct;

	at2 = mlx5dv_hws_action_template_create(actions, 5, &attr_at);
	if (!at2) {
		TEST_LOG("Failed to create at");
		return -1;
	}

	printf("action template no 2 created\n");

	/* Read_enable = 0 (Write Only) */
	aso_cseg_in.operand_masks =
			(ASO_BYTEWISE_64BYTE << ASO_CSEG_DATA_MASK_MODE_OFFSET) |
			(ASO_OPER_LOGICAL_OR << ASO_CSEG_COND_OPER_OFFSET) |
			(ASO_OP_ALWAYS_TRUE << ASO_CSEG_COND_1_OPER_OFFSET) |
			(ASO_OP_ALWAYS_TRUE << ASO_CSEG_COND_0_OPER_OFFSET);
	aso_cseg_in.data_mask = UINT64_MAX;
	for (i = 0; i < sizeof(aso_wqe.bytewise_data); i++)
		aso_wqe.bytewise_data[i] = i;
	aso_fill_send_cseg(0, 0, &aso_wqe.aso_ctrl, false);
	aso_fill_condition_cseg(&aso_cseg_in, &aso_wqe.aso_ctrl);
	wqe_len = sizeof(struct aso_wqe); /* ASO Control + ASO Data */


	rule_actions[0].tag.value = htobe32(4);
	rule_actions[1].reformat_l2_to_tnl_l3.hdr_data = data;
	rule_actions[1].reformat_l2_to_tnl_l3.resource_idx = 1;
	rule_actions[1].reformat_l2_to_tnl_l3.resource_offset = 3;
	rule_actions[3].counter.resource_offset = 8;
	rule_actions[3].counter.resource_idx = 1;
	rule_actions[3].counter.aso_wqe = (__be32 *) &aso_wqe;
	rule_actions[3].counter.wqe_size = wqe_len;
	rule_actions[4].aso_ct.resource_offset = 3;
	rule_actions[4].aso_ct.resource_idx = 0;
	rule_actions[4].aso_ct.direction = MLX5DV_HWS_ACTION_ASO_CT_DIRECTION_INITIATOR;

	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto clean_at;
	}

	printf("Table created\n");


	ret = create_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto tbl_destroy;
	}

	printf("Item created\n");

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		goto item_destroy;
	}

	printf("Match template created\n");

	mt_attr.comp_mask = 0;
	r_mt = mlx5dv_hws_match_template_create(ctx, NULL, 0, &mt_attr);
	if (!r_mt) {
		TEST_LOG("Failed to create root match template");
		return 1;
	}

	printf("Root Match template created\n");

	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;
	root_matcher = mlx5dv_hws_matcher_create(root_tbl, &r_mt, 1, &at_root, 1, &matcher_attr);
	if (!root_matcher) {
		TEST_LOG("Failed to create root matcher");
			return -1;
	}

	printf("Root Matcher created\n");

	matcher_attr.optimize_using_rule_idx = 1;
	all_at[0] = at;
	all_at[1] = at2;
	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, all_at, 2, &matcher_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher");
		goto mt_destroy;
	}

	printf("Matcher created\n");

	matcher2 = mlx5dv_hws_matcher_create(tbl, &mt, 1, all_at, 2, &matcher_attr);
	if (!matcher2) {
		TEST_LOG("Failed to create matcher");
		goto mt_destroy;
	}

	printf("Matcher  2 created\n");

	mlx5dv_hws_matcher_destroy(matcher2);

	rules = calloc(NUM_OF_RULES, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		return -1;
	}

	rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules);
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.queue_id = queue_id;

	ret = mlx5dv_hws_rule_create(root_matcher, 0, &root_item.data, 0,
				     rule_actions, drop, &rule_attr, rule);
	if (ret)
		TEST_LOG("Failed to create root rule");

	pending_rules = 1;
	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, 1, &miss_count, &hw_bp, true);
	pending_rules = 0;

	dump_all(ctx);

	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules + i *
							mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		DEVX_SET(my_basic_item, (void *)item.data.match, sip, i);
		ret = mlx5dv_hws_rule_create(matcher, 0, &item.data, i % 2 ? 0 : 1,
					     rule_actions, drop, &rule_attr, rule);
		if (ret)
			TEST_LOG("Failed to create rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH,
			      &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count,
		      &hw_bp, true);
	printf("queue [%d] Insertion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules + i *
						mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
		if (ret)
			TEST_LOG("Failed to destroy rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH,
			      &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH,
		      &miss_count, &hw_bp, true);
	printf("queue [%d] Deletion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	free(rules);

	mlx5dv_hws_match_template_destroy(mt);
	destroy_item(&item);
	mlx5dv_hws_action_destroy(pop);
	mlx5dv_hws_action_destroy(tag);
	mlx5dv_hws_action_destroy(push);
	mlx5dv_hws_action_destroy(r_tag);
	mlx5dv_hws_action_destroy(aso_counter);
	mlx5dv_hws_action_destroy(aso_ct);
	mlx5dv_hws_action_destroy(reformat_tl2_l2);
	mlx5dv_hws_action_destroy(reformat_l2_tl2);
	mlx5dv_hws_action_destroy(reformat_l2_tl3);
	mlx5dv_hws_action_destroy(r_reformat_tl3_l2);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_matcher_destroy(matcher);
	mlx5dv_hws_matcher_destroy(root_matcher);
	mlx5dv_hws_table_destroy(tbl);
	return 0;

mt_destroy:
	mlx5dv_hws_match_template_destroy(mt);
item_destroy:
	destroy_item(&item);
tbl_destroy:
	mlx5dv_hws_table_destroy(tbl);
clean_at:
	mlx5dv_hws_action_template_destroy(at);
clean_push:
	mlx5dv_hws_action_destroy(push);
clean_pop:
	mlx5dv_hws_action_destroy(pop);
clean_tag:
	mlx5dv_hws_action_destroy(tag);

	return ret;
}

int test_actions(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_context *ctx;
	int ret;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	}

	printf("Context created\n");

	ret = test_basic_actions(ctx);
	if (!ret)
		ret = test_nat64_action(ctx);

	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return ret;

dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	return -errno;
}
