/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test_resource.h"
#include <arpa/inet.h>

#define NUM_OF_QUEUES 16
#define ROW_LOG 0
#define NUM_OF_RULES (1 << ROW_LOG)
#define ARG_BULK_LOG ROW_LOG
#define NUM_OF_ARGS (1 << ARG_BULK_LOG)
#define QUEUE_SIZE 256
#define BURST_TH 1
#define ARG_QUEUE_ID 1

#define VLAN_TPID 0x8100
#define VLAN_PRIO 5
#define VLAN_CFI 1
#define VLAN_ID 0xC0C
#define PACKET_START 0x0
#define MAC_START 0x1
#define FIRST_VLAN_START 0x2
#define INNER_MAC 0x13
#define VLAN_START_IDX 12

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			printf("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

static int test_add_hdr(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_resource_attr resource_attr = {0};
	struct mlx5dv_hws_action_data actions_data[1] = {0};
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_action_insert_header hdr = {0};
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_resource *all_resources[4];
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_action *dest_action;
	struct mlx5dv_hws_resource *resource;
	struct mlx5dv_hws_matcher *matcher;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	struct mlx5dv_hws_action *mh;
	struct mlx5dv_hws_table *tbl;
	uint32_t pending_rules = 0;
	struct mlx5dv_hws_at *at;
	struct mlx5dv_hws_mt *mt;
	uint32_t miss_count = 0;
	uint16_t queue_id = 0;
	uint16_t vlan_hdr[2];
	uint32_t hw_bp = 0;
	uint16_t vlan_tci;
	int ret;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	}

	/* Allocate argument resource */
	resource_attr.bulk_log_size = ARG_BULK_LOG;
	resource_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B;
	resource = mlx5dv_hws_resource_alloc(ctx, &resource_attr);
	if (!resource) {
		TEST_LOG("Failed to allocate resource");
		goto ctx_close;
	}

	/* Construct VLAN TCI (Tag Control Information) */
	vlan_tci = (VLAN_PRIO << 13) | (VLAN_CFI << 12) | VLAN_ID;
	vlan_hdr[0] = htons(VLAN_TPID);
	vlan_hdr[1] = htons(vlan_tci);

	hdr.anchor = MLX5DV_HWS_ACTION_ANCHOR_PACKET_START;
	hdr.offset = VLAN_START_IDX;
	hdr.encap = 0;
	hdr.hdr_data.hdr_data = (uint8_t *)vlan_hdr;
	hdr.hdr_data.sz = sizeof(vlan_hdr);
	all_resources[0] = resource;

	mh = mlx5dv_hws_action_create_insert_header(ctx, all_resources, 1, &hdr,
						    MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!mh) {
		TEST_LOG("Failed to create insert header action");
		goto resource_destroy;
	}

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto action_destroy;
	}

	ret = create_basic_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto tbl_destroy;
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		goto item_destroy;
	}

	at = mlx5dv_hws_action_template_create(&mh, 1, &at_attr);
	if (!at) {
		TEST_LOG("Failed to create dummy action template");
		goto mt_destroy;
	}

	/* Create HWS Matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;
	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &matcher_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher");
		goto at_destroy;
	}

	dest_action = mlx5dv_hws_action_create_dest_default_miss(ctx,
								 MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!dest_action) {
		TEST_LOG("Failed to create default miss action\n");
		goto matcher_destroy;
	}

	/* Dump once all control data created */
	dump_all(ctx);

	rules = calloc(NUM_OF_RULES, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		goto default_miss_destroy;
	}

	rule_attr.queue_id = queue_id;

	rule = (struct mlx5dv_hws_rule *)(uint8_t *)rules;
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.rule_idx = 0;

	actions_data[0].insert_header.resource_offset = 0;
	actions_data[0].insert_header.hdr_data = (uint8_t *)vlan_hdr;

	DEVX_SET(my_basic_item, (void *)item.data.match, sip, 0);
	ret = mlx5dv_hws_rule_create(matcher, 0, &item.data, 0, actions_data, dest_action,
					&rule_attr, rule);
	if (ret) {
		TEST_LOG("Failed to create rule %d", 0);
		goto default_miss_destroy;
	}

	pending_rules++;

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Insertion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	rule = (struct mlx5dv_hws_rule *)(uint8_t *)rules;
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.rule_idx = 0;

	ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
	if (ret) {
		TEST_LOG("Failed to destroy rule %d", 0);
		goto default_miss_destroy;
	}

	pending_rules++;

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Deletion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	free(rules);
	mlx5dv_hws_action_destroy(dest_action);
	mlx5dv_hws_matcher_destroy(matcher);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_match_template_destroy(mt);
	destroy_basic_item(&item);
	mlx5dv_hws_table_destroy(tbl);
	mlx5dv_hws_action_destroy(mh);
	mlx5dv_hws_resource_free(resource);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return ret;

default_miss_destroy:
	mlx5dv_hws_action_destroy(dest_action);
matcher_destroy:
	mlx5dv_hws_matcher_destroy(matcher);
at_destroy:
	mlx5dv_hws_action_template_destroy(at);
mt_destroy:
	mlx5dv_hws_match_template_destroy(mt);
item_destroy:
	destroy_basic_item(&item);
tbl_destroy:
	mlx5dv_hws_table_destroy(tbl);
action_destroy:
	mlx5dv_hws_action_destroy(mh);
resource_destroy:
	mlx5dv_hws_resource_free(resource);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	TEST_LOG("Test failed errno=%d", errno);
	return -errno;
}

static int test_rm_hdr(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_action_remove_header_attr rm_hdr = {0};
	struct mlx5dv_hws_action_data actions_data[1] = {0};
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_action *dest_action;
	struct mlx5dv_hws_matcher *matcher;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	struct mlx5dv_hws_action *mh;
	struct mlx5dv_hws_table *tbl;
	uint32_t pending_rules = 0;
	struct mlx5dv_hws_at *at;
	struct mlx5dv_hws_mt *mt;
	uint32_t miss_count = 0;
	uint16_t queue_id = 0;
	uint32_t hw_bp = 0;
	int ret;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	}

	rm_hdr.type = MLX5DV_HWS_ACTION_REMOVE_HEADER_TYPE_BY_HEADER;
	rm_hdr.by_header.start_anchor = MLX5DV_HWS_ACTION_ANCHOR_PACKET_START;
	rm_hdr.by_header.end_anchor = MLX5DV_HWS_ACTION_ANCHOR_INNER_MAC;
	rm_hdr.by_header.decap = true;

	mh = mlx5dv_hws_action_create_remove_header(ctx, &rm_hdr,
						    MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!mh) {
		TEST_LOG("Failed to create remove header action");
		goto ctx_close;
	}

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto action_destroy;
	}

	ret = create_basic_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto tbl_destroy;
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		goto item_destroy;
	}

	at = mlx5dv_hws_action_template_create(&mh, 1, &at_attr);
	if (!at) {
		TEST_LOG("Failed to create dummy action template");
		goto mt_destroy;
	}

	/* Create HWS Matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;
	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &matcher_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher");
		goto at_destroy;
	}

	dest_action = mlx5dv_hws_action_create_dest_default_miss(ctx,
								 MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!dest_action) {
		TEST_LOG("Failed to create default miss action\n");
		goto matcher_destroy;
	}

	/* Dump once all control data created */
	dump_all(ctx);

	rules = calloc(NUM_OF_RULES, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		goto default_miss_destroy;
	}

	rule_attr.queue_id = queue_id;

	rule = (struct mlx5dv_hws_rule *)(uint8_t *)rules;
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.rule_idx = 0;

	actions_data[0].insert_header.resource_offset = 0;

	DEVX_SET(my_basic_item, (void *)item.data.match, sip, 0);
	ret = mlx5dv_hws_rule_create(matcher, 0, &item.data, 0, actions_data, dest_action,
					&rule_attr, rule);
	if (ret) {
		TEST_LOG("Failed to create rule %d", 0);
		goto default_miss_destroy;
	}

	pending_rules++;

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Insertion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	rule = (struct mlx5dv_hws_rule *)(uint8_t *)rules;
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.rule_idx = 0;

	ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
	if (ret) {
		TEST_LOG("Failed to destroy rule %d", 0);
		goto default_miss_destroy;
	}

	pending_rules++;

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Deletion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	free(rules);
	mlx5dv_hws_action_destroy(dest_action);
	mlx5dv_hws_matcher_destroy(matcher);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_match_template_destroy(mt);
	destroy_basic_item(&item);
	mlx5dv_hws_table_destroy(tbl);
	mlx5dv_hws_action_destroy(mh);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return 0;

default_miss_destroy:
	mlx5dv_hws_action_destroy(dest_action);
matcher_destroy:
	mlx5dv_hws_matcher_destroy(matcher);
at_destroy:
	mlx5dv_hws_action_template_destroy(at);
mt_destroy:
	mlx5dv_hws_match_template_destroy(mt);
item_destroy:
	destroy_basic_item(&item);
tbl_destroy:
	mlx5dv_hws_table_destroy(tbl);
action_destroy:
	mlx5dv_hws_action_destroy(mh);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	TEST_LOG("Test failed errno=%d", errno);
	return -errno;
}

int test_add_rm_hdr(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	if (test_add_hdr(ibv_ctx, t_info)) {
		TEST_LOG("Test Insert Header action failed errno=%d", errno);
		return -errno;
	}
	TEST_LOG("Test Insert Header passed");

	if (test_rm_hdr(ibv_ctx, t_info)) {
		TEST_LOG("Test Remove Header action failed errno=%d", errno);
		return -errno;
	}
	TEST_LOG("Test Remove Header passed");

	return 0;
}
