/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#define _GNU_SOURCE
#include <sched.h>
#include <time.h>
#include "hws_test.h"

#define DUMP_FILE_PATH "/tmp/dump.csv"

int dump_all(struct mlx5dv_hws_context *ctx)
{
	FILE* file;

	file = fopen(DUMP_FILE_PATH, "w+");
	if (!file) {
		TEST_LOG("Failed to open file");
		return -errno;
	}

	if (mlx5dv_hws_debug_dump(ctx, file)) {
		TEST_LOG("Failed to dump");
		goto close_file;
	} else {
		printf("Dumped app info to %s\n", DUMP_FILE_PATH);
	}

	fclose(file);
	return 0;

close_file:
	fclose(file);
	return -errno;
}

int create_basic_item(struct mlx5dv_hws_item *item)
{
	struct mlx5dv_hws_field *fields;
	void *mask, *match;

	fields = calloc(BASIC_ITEM_NUM_OF_FIELDS, sizeof(struct mlx5dv_hws_field));
	if (!fields) {
		TEST_LOG("Failed to allocate fields");
		return ENOMEM;
	}

	match = calloc(1, DEVX_ST_SZ_BYTES(my_basic_item));
	if (!match) {
		TEST_LOG("Failed to allocate match");
		free(fields);
		return ENOMEM;
	}

	mask = calloc(1, DEVX_ST_SZ_BYTES(my_basic_item));
	if (!mask) {
		TEST_LOG("Failed to allocate mask");
		free(match);
		free(fields);
		return ENOMEM;
	}

	item->num_of_fields = BASIC_ITEM_NUM_OF_FIELDS;
	item->fields = fields;
	item->data.mask = mask;
	item->data.match = match;

	/* DON'T CHANGE FIELD ORDER BELOW ! SIP, DIP, VERSION */
	fields[0].fname = MLX5DV_HWS_FNAME_IPV4_SRC;
	fields[0].header = MLX5DV_HWS_HEADER_OUTER;
	fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[0].value.bit_off = CALC_BIT_OFF(my_basic_item, sip);
	DEVX_SET(my_basic_item, mask, sip, -1);
	DEVX_SET(my_basic_item, match, sip, 0x01010101);

	fields[1].fname = MLX5DV_HWS_FNAME_IPV4_DST;
	fields[1].header = MLX5DV_HWS_HEADER_OUTER;
	fields[1].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[1].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[1].value.bit_off = CALC_BIT_OFF(my_basic_item, dip);
	DEVX_SET(my_basic_item, mask, dip, -1);
	DEVX_SET(my_basic_item, match, dip, 0x02020202);

	fields[2].fname = MLX5DV_HWS_FNAME_IP_VERSION;
	fields[2].header = MLX5DV_HWS_HEADER_OUTER;
	fields[2].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[2].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[2].value.bit_off = CALC_BIT_OFF(my_basic_item, version);
	DEVX_SET(my_basic_item, mask, version, -1);
	DEVX_SET(my_basic_item, match, version, 0x4);

	return 0;
}

void destroy_basic_item(struct mlx5dv_hws_item *item)
{
	free((void*)item->data.match);
	free((void*)item->data.mask);
	free(item->fields);
}

int process_gettime(struct timespec *t)
{
	return clock_gettime(CLOCK_PROCESS_CPUTIME_ID, t);
}

int thread_gettime(struct timespec *t)
{
	return clock_gettime(CLOCK_THREAD_CPUTIME_ID, t);
}

double timediff(const struct timespec *start, const struct timespec *end)
{
	return (end->tv_sec - start->tv_sec) + ((end->tv_nsec - start->tv_nsec) / 1000000000.0);
}

int set_cpu_cores(cpu_set_t *set, uint64_t core_bits)
{
	int num_of_cores = 0;
	int i;

	CPU_ZERO(set);

	for (i = 0; i < 64; i++) {
		if (core_bits & (1ULL << i)) {
			CPU_SET(i, set);
			num_of_cores++;
		}
	}

	return num_of_cores;
}
