/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"
#include "../hws_mlx5_ifc.h"

#define NUM_OF_QUEUES 16
#define ROW_LOG 4
#define QUEUE_SIZE 256
#define BURST_TH 32
#define NUM_CASES 6

struct mlx5dv_hws_matcher *matchers[NUM_CASES] = {0};
struct mlx5dv_hws_mt *mts[NUM_CASES] = {0};
struct mlx5dv_hws_rule *rules;
struct mlx5dv_hws_action *drop;
struct mlx5dv_hws_context *ctx;
struct mlx5dv_hws_table *tbl;
struct mlx5dv_hws_at *at;

static int check_comp_support(struct ibv_context *ibv_ctx, bool *has_supp)
{
	uint32_t out[DEVX_ST_SZ_DW(query_hca_cap_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(query_hca_cap_in)] = {0};
	uint16_t supp_ste_format_gen_wqe;
	int ret;

	DEVX_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
	DEVX_SET(query_hca_cap_in, in, op_mod,
			MLX5_GET_HCA_CAP_OP_MOD_WQE_BASED_FLOW_TABLE |
			MLX5_HCA_CAP_OPMOD_GET_CUR);

	ret = mlx5dv_devx_general_cmd(ibv_ctx, in, sizeof(in), out, sizeof(out));
	if (ret)
		return ret;

	supp_ste_format_gen_wqe =
		DEVX_GET(query_hca_cap_out, out,
			 capability.wqe_based_flow_table_cap.ste_format_gen_wqe);

	*has_supp = supp_ste_format_gen_wqe & (1 << MLX5_IFC_RTC_STE_FORMAT_4DW_RANGE);

	return 0;
}

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool got_comp = *pending_rules >= expected_comp;
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			TEST_LOG("Failed during poll queue");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

static int
create_matcher_insert_one_rule(struct mlx5dv_hws_item *item, uint8_t idx)
{
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	uint32_t miss_count, hw_bp, pending_rules;
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_matcher *matcher;
	struct mlx5dv_hws_rule *rule;
	struct mlx5dv_hws_mt *mt;
	int ret;

	/* Create HWS SRC Matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = idx;
	matcher_attr.resizable = true;

	mt = mlx5dv_hws_match_template_create(ctx, item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		return -1;
	}

	TEST_LOG("Match template created case %d", idx);

	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &matcher_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher idx %d", idx);
		goto free_mt;
	}

	TEST_LOG("Matcher created case %d", idx);

	rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules +
						     idx * mlx5dv_hws_rule_get_handle_size());
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.rule_idx = 0;
	rule_attr.queue_id = 0;

	ret = mlx5dv_hws_rule_create(matcher, 0, &item->data, 0, NULL, drop, &rule_attr, rule);
	if (ret)
		TEST_LOG("Failed to create rule");
	else
		TEST_LOG("Rule created case %d", idx);

	pending_rules = 1;
	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);

	matchers[idx] = matcher;
	mts[idx] = mt;
	return 0;

free_mt:
	mlx5dv_hws_match_template_destroy(mt);
	return -1;
}

int test_compare(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_item item = {0};
	bool fw_supp;
	int ret, i;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	} else {
		TEST_LOG("Context created");
	}

	ret = check_comp_support(ibv_ctx, &fw_supp);
	if (ret) {
		TEST_LOG("Fail to check cap");
		errno = EINVAL;
		goto ctx_close;
	}

	if (!fw_supp) {
		TEST_LOG("Comp matching not supported");
		mlx5dv_hws_context_close(ctx);
		ibv_dealloc_pd(ctx_attr.pd);
		return 0;
	}

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto ctx_close;
	} else {
		TEST_LOG("Table created");
	}

	at = mlx5dv_hws_action_template_create(NULL, 0, &at_attr);
	if (!at) {
		TEST_LOG("Failed to create dummy action template");
		goto tbl_destroy;
	} else {
		TEST_LOG("Dummy action template created");
	}

	ret = create_basic_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto at_destroy;
	} else {
		TEST_LOG("Item created");
	}

	drop = mlx5dv_hws_action_create_dest_drop(ctx, MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!drop) {
		TEST_LOG("Failed to create drop action");
		goto item_destroy;
	}

	rules = calloc(NUM_CASES, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		goto drop_action_destroy;
	}

	/* Case 0 */
	DEVX_SET(my_basic_item, (void *)item.data.match, sip, 0x01010101);
	item.fields->op = MLX5DV_HWS_FIELD_MATCH_OP_NE;
	item.fields->type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	ret = create_matcher_insert_one_rule(&item, 0);
	if (ret) {
		TEST_LOG("Failed to run case 0");
		goto free_rules;
	} else {
		TEST_LOG("Passed case 0, SIP != 1.1.1.1");
	}

	/* Case 1 */
	DEVX_SET(my_basic_item, (void *)item.data.match, sip, 0x01010101);
	item.fields->op = MLX5DV_HWS_FIELD_MATCH_OP_GT;
	item.fields->type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	ret = create_matcher_insert_one_rule(&item, 1);
	if (ret) {
		TEST_LOG("Failed to run case 1");
		goto free_cases;
	} else {
		TEST_LOG("Passed case 1, SIP > 1.1.1.1");
	}

	/* Case 2 */
	DEVX_SET(my_basic_item, (void *)item.data.match, sip, 0x01010101);
	item.fields->op = MLX5DV_HWS_FIELD_MATCH_OP_LT;
	item.fields->type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	ret = create_matcher_insert_one_rule(&item, 2);
	if (ret) {
		TEST_LOG("Failed to run case 2");
		goto free_cases;
	} else {
		TEST_LOG("Passed case 2, SIP < 1.1.1.1");
	}

	/* Case 3 */
	item.fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_LT;
	item.fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	DEVX_SET(my_basic_item, (void *)item.data.match, sip, 0x01010101);
	item.fields[1].op = MLX5DV_HWS_FIELD_MATCH_OP_LT;
	item.fields[1].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	DEVX_SET(my_basic_item, (void *)item.data.match, dip, 0x02020202);
	ret = create_matcher_insert_one_rule(&item, 3);
	if (ret) {
		TEST_LOG("Failed to run case 3");
		goto free_cases;
	} else {
		TEST_LOG("Passed case 3, SIP < 1.1.1.1, DIP < 2.2.2.2");
	}

	/* Case 4 */
	item.fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_NE;
	item.fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_FIELD;
	item.fields[0].field.fname_b = MLX5DV_HWS_FNAME_IPV4_DST;
	item.fields[0].field.header_b = MLX5DV_HWS_HEADER_OUTER;
	ret = create_matcher_insert_one_rule(&item, 4);
	if (ret) {
		TEST_LOG("Failed to run case 4");
		goto free_cases;
	} else {
		TEST_LOG("Passed case 4, SIP != DIP");
	}

	/* Case 5 */
	item.fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_NE;
	item.fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_FIELD;
	item.fields[0].field.fname_b = MLX5DV_HWS_FNAME_IPV4_DST;
	item.fields[0].field.header_b = MLX5DV_HWS_HEADER_OUTER;
	item.fields[1].op = MLX5DV_HWS_FIELD_MATCH_OP_NE;
	item.fields[1].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_FIELD;
	item.fields[1].field.fname_b = MLX5DV_HWS_FNAME_REG_C1;
	item.fields[1].field.header_b = MLX5DV_HWS_HEADER_NONE;
	ret = create_matcher_insert_one_rule(&item, 5);
	if (ret) {
		TEST_LOG("Failed to run case 5");
		goto free_cases;
	} else {
		TEST_LOG("Passed case 5, SIP != DIP, DIP != REG_C1");
	}

	/* Dump once all control data created */
	dump_all(ctx);

	TEST_LOG("Deleting compare rules");
	for (i = 0; i < NUM_CASES; i++) {
		struct mlx5dv_hws_rule_attr rule_attr = {0};
		uint32_t miss_count, hw_bp, pending_rules;
		struct mlx5dv_hws_rule *rule;

		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules +
							     i * mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = 0;
		rule_attr.rule_idx = 0;
		rule_attr.queue_id = 0;
		ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
		if (ret) {
			TEST_LOG("Failed to destroy rule, %d", i);
			goto free_cases;
		}
		pending_rules = 1;
		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count,
			      &hw_bp, true);
	}

	TEST_LOG("Deleting compare mt & matchers");
	for (i = 0; i < NUM_CASES; i++) {
		mlx5dv_hws_match_template_destroy(mts[i]);
		mlx5dv_hws_matcher_destroy(matchers[i]);
	}

	free(rules);
	mlx5dv_hws_action_destroy(drop);
	destroy_basic_item(&item);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_table_destroy(tbl);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return 0;

free_cases:
	for (i = 0; i < NUM_CASES; i++) {
		if (mts[i])
			mlx5dv_hws_match_template_destroy(mts[i]);

		if (matchers[i])
			mlx5dv_hws_matcher_destroy(matchers[i]);
	}
free_rules:
	free(rules);
drop_action_destroy:
	mlx5dv_hws_action_destroy(drop);
item_destroy:
	destroy_basic_item(&item);
at_destroy:
	mlx5dv_hws_action_template_destroy(at);
tbl_destroy:
	mlx5dv_hws_table_destroy(tbl);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	TEST_LOG("Test failed errno=%d", errno);
	return -errno;
}
