/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"

#define CONV_ITEM_NUM_OF_FIELDS 1
#define NUM_OF_QUEUES 16
#define PORT_BIT_SIZE 16
#define QUEUE_SIZE 256
#define BITS_IN_BYTE 8
#define BITS_IN_DW 32
#define BURST_TH 1
#define ROW_LOG 2

/* Simple field 32 GET */
#define HWS_GET(p, bit_off, bit_sz) \
	((be32toh(*(const __be32 *)((uint8_t *)(p) + bit_off / BITS_IN_BYTE)) >> \
		(BITS_IN_DW - bit_sz - bit_off % BITS_IN_BYTE)) & \
		__devx_mask(bit_sz))

struct user_conv_data {
	uint32_t bit_off;
};

void user_conv_func(const struct mlx5dv_hws_item_data *item_data,
		    void *user_conv_data,
		    enum mlx5dv_hws_field_user_conv_flag conv_flag,
		    uint32_t *ret_conv_data)
{
	uint32_t bit_offset;
	uint16_t port_val;

	printf("Conversion function\n");

	if (!ret_conv_data || !item_data) {
		TEST_LOG("Bad input");
		return;
	}

	if (conv_flag == MLX5DV_HWS_FIELD_USER_CONV_FLAG_MASK) {
		*ret_conv_data = -1;
		return;
	}

	if (conv_flag != MLX5DV_HWS_FIELD_USER_CONV_FLAG_MATCH)
		return;

	bit_offset = ((struct user_conv_data *)user_conv_data)->bit_off;

	port_val = HWS_GET(item_data->match, bit_offset, PORT_BIT_SIZE);
	printf("Given port number: 0x%x\n", (unsigned int)(port_val));

	*ret_conv_data = port_val + 1;
	printf("Return converted data: 0x%x\n", (unsigned int)(*ret_conv_data));
}

static int create_conv_item(struct mlx5dv_hws_item *item)
{
	struct user_conv_data *conv_data;
	struct mlx5dv_hws_field *fields;
	void *mask, *match;

	fields = calloc(CONV_ITEM_NUM_OF_FIELDS, sizeof(struct mlx5dv_hws_field));
	if (!fields) {
		TEST_LOG("Failed to allocate fields");
		return ENOMEM;
	}

	match = calloc(1, DEVX_ST_SZ_BYTES(my_basic_item));
	if (!match) {
		TEST_LOG("Failed to allocate match");
		free(fields);
		return ENOMEM;
	}

	mask = calloc(1, DEVX_ST_SZ_BYTES(my_basic_item));
	if (!mask) {
		TEST_LOG("Failed to allocate mask");
		free(match);
		free(fields);
		return ENOMEM;
	}
	conv_data = calloc(1, sizeof(struct user_conv_data));
	if (!conv_data) {
		TEST_LOG("Failed to allocate conversion data");
		free(mask);
		free(match);
		free(fields);
		return ENOMEM;
	}

	item->num_of_fields = CONV_ITEM_NUM_OF_FIELDS;
	item->fields = fields;
	item->data.mask = mask;
	item->data.match = match;

	fields[0].fname = MLX5DV_HWS_FNAME_IP_UDP_SPORT;
	fields[0].header = MLX5DV_HWS_HEADER_OUTER;
	fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[0].value.bit_off = CALC_BIT_OFF(my_basic_item, sport);
	conv_data->bit_off = CALC_BIT_OFF(my_basic_item, sport);
	fields[0].value.user_conv_data = (void *)conv_data;
	fields[0].value.user_conv_func = user_conv_func;
	DEVX_SET(my_basic_item, mask, sport, 0xff);
	DEVX_SET(my_basic_item, match, sport, 0xaa);

	return 0;
}

static void destroy_conv_item(struct mlx5dv_hws_item *item)
{
	free((void *)item->data.match);
	free((void *)item->data.mask);
	free((void *)item->fields[0].value.user_conv_data);
	free(item->fields);
}


static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			printf("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

int test_conv_func(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_matcher *matcher, *root_matcher;
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_table *tbl, *root_tbl;
	struct mlx5dv_hws_action *dest_action;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	uint32_t pending_rules = 0;
	struct mlx5dv_hws_at *at;
	struct mlx5dv_hws_mt *mt;
	uint32_t miss_count = 0;
	uint16_t queue_id = 0;
	uint32_t hw_bp = 0;
	int ret;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	} else {
		printf("Context created\n");
	}

	/* Create Root Table */
	tbl_attr.level = 0;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	root_tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!root_tbl) {
		TEST_LOG("Failed to create root table");
		goto ctx_close;
	} else {
		TEST_LOG("Root Table created");
	}

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto root_tbl_destroy;
	} else {
		printf("Table created\n");
	}

	ret = create_conv_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto tbl_destroy;
	} else {
		printf("Item created\n");
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		goto item_destroy;
	} else {
		printf("Match template created\n");
	}

	at = mlx5dv_hws_action_template_create(NULL, 0, &at_attr);
	if (!at) {
		TEST_LOG("Failed to create dummy action template");
		goto mt_destroy;
	} else {
		printf("Dummy action template created\n");
	}

	/* Create ROOT Matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;
	root_matcher = mlx5dv_hws_matcher_create(root_tbl, &mt, 1, &at, 1, &matcher_attr);
	if (!root_matcher) {
		TEST_LOG("Failed to create root matcher");
		goto at_destroy;
	} else {
		printf("Root Matcher created\n");
	}

	/* Create HWS Matcher */
	matcher_attr.optimize_using_rule_idx = 1;
	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &matcher_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher");
		goto root_matcher_destroy;
	} else {
		printf("Matcher created\n");
	}

	dest_action = mlx5dv_hws_action_create_dest_default_miss(ctx,
								 MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!dest_action) {
		TEST_LOG("Failed to create default miss action\n");
		goto matcher_destroy;
	}

	rules = calloc(1, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		goto default_miss_destroy;
	}

	rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules);
	rule_attr.queue_id = queue_id;
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.rule_idx = 0;

	ret = mlx5dv_hws_rule_create(matcher, 0, &item.data, 0, NULL, dest_action, &rule_attr,
					 rule);

	if (ret)
		TEST_LOG("Failed to create rule");

	pending_rules = 1;
	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);

	printf("queue [%d] Insertion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		1,
		hw_bp);

	/* Dump once all control data created */
	dump_all(ctx);

	ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
	if (ret)
		TEST_LOG("Failed to destroy rule");

	pending_rules = 1;
	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);

	printf("queue [%d] Deletion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		1,
		hw_bp);

	free(rules);
	mlx5dv_hws_action_destroy(dest_action);
	mlx5dv_hws_matcher_destroy(root_matcher);
	mlx5dv_hws_matcher_destroy(matcher);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_match_template_destroy(mt);
	destroy_conv_item(&item);
	mlx5dv_hws_table_destroy(root_tbl);
	mlx5dv_hws_table_destroy(tbl);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return 0;

default_miss_destroy:
	mlx5dv_hws_action_destroy(dest_action);
matcher_destroy:
	mlx5dv_hws_matcher_destroy(matcher);
root_matcher_destroy:
	mlx5dv_hws_matcher_destroy(root_matcher);
at_destroy:
	mlx5dv_hws_action_template_destroy(at);
mt_destroy:
	mlx5dv_hws_match_template_destroy(mt);
item_destroy:
	destroy_conv_item(&item);
tbl_destroy:
	mlx5dv_hws_table_destroy(tbl);
root_tbl_destroy:
	mlx5dv_hws_table_destroy(root_tbl);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	TEST_LOG("Test failed errno=%d", errno);
	return -errno;
}
