/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"
#include "../mlx5dv_hws_ct.h"
#include "../hws_mlx5_ifc.h"

#define NUM_OF_FIELDS 2
#define NUM_OF_QUEUES 16
#define ROW_LOG 16
#define NUM_OF_RULES (1 << ROW_LOG)
#define QUEUE_SIZE 256
#define BURST_TH 32

struct mlx5_ifc_my_item_bits {
	uint8_t sip[0x20];
	uint8_t dip[0x20];
	uint8_t sport[0x10];
	uint8_t dport[0x10];
};

struct mlx5_ifc_my_ct_item_bits {
	uint8_t sip[0x20];
	uint8_t dip[0x20];
	uint8_t regc[0x20];
	uint8_t sport[0x10];
	uint8_t dport[0x10];
	uint8_t ipver[0x4];
	uint8_t prot[0x8];
};

struct mlx5_ifc_my_ct_v6_item_bits {
	uint8_t sip_0[0x20];
	uint8_t sip_1[0x20];
	uint8_t sip_2[0x20];
	uint8_t sip_3[0x20];
	uint8_t dip_0[0x20];
	uint8_t dip_1[0x20];
	uint8_t dip_2[0x20];
	uint8_t dip_3[0x20];
	uint8_t regc[0x20];
	uint8_t sport[0x10];
	uint8_t dport[0x10];
	uint8_t ipver[0x4];
	uint8_t prot[0x8];
};


static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			printf("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

static int create_ct_ipv4_item(struct mlx5dv_hws_item *item)
{
	struct mlx5dv_hws_field *fields;
	void *mask, *match, *buff;

	fields = calloc(7, sizeof(struct mlx5dv_hws_field));
	if (!fields) {
		TEST_LOG("Failed to allocate fields");
		return ENOMEM;
	}

	buff = calloc(7, DEVX_ST_SZ_BYTES(my_ct_item));
	if (!buff) {
		TEST_LOG("Failed to allocate mask and match");
		free(fields);
		return ENOMEM;
	}

	match = buff;
	mask = buff + DEVX_ST_SZ_BYTES(my_ct_item);

	item->num_of_fields = 7;
	item->fields = fields;
	item->data.mask = mask;
	item->data.match = match;

	fields[0].fname = MLX5DV_HWS_FNAME_IPV4_SRC;
	fields[0].header = MLX5DV_HWS_HEADER_OUTER;
	fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[0].value.bit_off = CALC_BIT_OFF(my_ct_item, sip);
	DEVX_SET(my_ct_item, mask, sip, -1);

	fields[1].fname = MLX5DV_HWS_FNAME_IPV4_DST;
	fields[1].header = MLX5DV_HWS_HEADER_OUTER;
	fields[1].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[1].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[1].value.bit_off = CALC_BIT_OFF(my_ct_item, dip);
	DEVX_SET(my_ct_item, mask, dip, -1);

	fields[2].fname = MLX5DV_HWS_FNAME_REG_C1;
	fields[2].header = MLX5DV_HWS_HEADER_NONE;
	fields[2].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[2].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[2].value.bit_off = CALC_BIT_OFF(my_ct_item, regc);
	DEVX_SET(my_ct_item, mask, regc, -1);

	fields[3].fname = MLX5DV_HWS_FNAME_IP_PROTOCOL;
	fields[3].header = MLX5DV_HWS_HEADER_OUTER;
	fields[3].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[3].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[3].value.bit_off = CALC_BIT_OFF(my_ct_item, prot);
	DEVX_SET(my_ct_item, mask, prot, -1);

	fields[4].fname = MLX5DV_HWS_FNAME_IP_SPORT;
	fields[4].header = MLX5DV_HWS_HEADER_OUTER;
	fields[4].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[4].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[4].value.bit_off = CALC_BIT_OFF(my_ct_item, sport);
	DEVX_SET(my_ct_item, mask, sport, -1);

	fields[5].fname = MLX5DV_HWS_FNAME_IP_DPORT;
	fields[5].header = MLX5DV_HWS_HEADER_OUTER;
	fields[5].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[5].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[5].value.bit_off = CALC_BIT_OFF(my_ct_item, dport);
	DEVX_SET(my_ct_item, mask, dport, -1);

	fields[6].fname = MLX5DV_HWS_FNAME_IP_VERSION;
	fields[6].header = MLX5DV_HWS_HEADER_OUTER;
	fields[6].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[6].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[6].value.bit_off = CALC_BIT_OFF(my_ct_item, ipver);
	DEVX_SET(my_ct_item, mask, ipver, -1);

	return 0;
}

static int create_ct_ipv6_item(struct mlx5dv_hws_item *item)
{
	struct mlx5dv_hws_field *fields;
	void *mask, *match, *buff;

	fields = calloc(13, sizeof(struct mlx5dv_hws_field));
	if (!fields) {
		TEST_LOG("Failed to allocate fields");
		return ENOMEM;
	}

	buff = calloc(13, DEVX_ST_SZ_BYTES(my_ct_v6_item));
	if (!buff) {
		TEST_LOG("Failed to allocate mask and match");
		free(fields);
		return ENOMEM;
	}

	match = buff;
	mask = buff + DEVX_ST_SZ_BYTES(my_ct_v6_item);

	item->num_of_fields = 13;
	item->fields = fields;
	item->data.mask = mask;
	item->data.match = match;

	fields[0].fname = MLX5DV_HWS_FNAME_IPV6_SRC_127_96;
	fields[0].header = MLX5DV_HWS_HEADER_OUTER;
	fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[0].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, sip_0);
	DEVX_SET(my_ct_v6_item, mask, sip_0, -1);
	DEVX_SET(my_ct_v6_item, match, sip_0, 0);

	fields[1].fname = MLX5DV_HWS_FNAME_IPV6_DST_127_96;
	fields[1].header = MLX5DV_HWS_HEADER_OUTER;
	fields[1].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[1].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[1].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, dip_0);
	DEVX_SET(my_ct_v6_item, mask, dip_0, -1);
	DEVX_SET(my_ct_v6_item, match, dip_0, 0x05060708);

	fields[2].fname = MLX5DV_HWS_FNAME_IPV6_SRC_95_64;
	fields[2].header = MLX5DV_HWS_HEADER_OUTER;
	fields[2].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[2].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[2].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, sip_1);
	DEVX_SET(my_ct_v6_item, mask, sip_1, -1);
	DEVX_SET(my_ct_v6_item, match, sip_1, 0);

	fields[3].fname = MLX5DV_HWS_FNAME_IPV6_DST_95_64;
	fields[3].header = MLX5DV_HWS_HEADER_OUTER;
	fields[3].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[3].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[3].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, dip_1);
	DEVX_SET(my_ct_v6_item, mask, dip_1, -1);
	DEVX_SET(my_ct_v6_item, match, dip_1, 0x05060708);

	fields[4].fname = MLX5DV_HWS_FNAME_IPV6_SRC_63_32;
	fields[4].header = MLX5DV_HWS_HEADER_OUTER;
	fields[4].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[4].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[4].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, sip_2);
	DEVX_SET(my_ct_v6_item, mask, sip_2, -1);


	fields[5].fname = MLX5DV_HWS_FNAME_IPV6_DST_63_32;
	fields[5].header = MLX5DV_HWS_HEADER_OUTER;
	fields[5].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[5].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[5].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, dip_2);
	DEVX_SET(my_ct_v6_item, mask, dip_2, -1);

	fields[6].fname = MLX5DV_HWS_FNAME_IPV6_SRC_31_0;
	fields[6].header = MLX5DV_HWS_HEADER_OUTER;
	fields[6].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[6].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[6].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, sip_3);
	DEVX_SET(my_ct_v6_item, mask, sip_3, -1);

	fields[7].fname = MLX5DV_HWS_FNAME_IPV6_DST_31_0;
	fields[7].header = MLX5DV_HWS_HEADER_OUTER;
	fields[7].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[7].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[7].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, dip_3);
	DEVX_SET(my_ct_v6_item, mask, dip_3, -1);


	fields[8].fname = MLX5DV_HWS_FNAME_REG_C1;
	fields[8].header = MLX5DV_HWS_HEADER_NONE;
	fields[8].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[8].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[8].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, regc);
	DEVX_SET(my_ct_v6_item, mask, regc, -1);

	fields[9].fname = MLX5DV_HWS_FNAME_IP_PROTOCOL;
	fields[9].header = MLX5DV_HWS_HEADER_OUTER;
	fields[9].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[9].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[9].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, prot);
	DEVX_SET(my_ct_v6_item, mask, prot, -1);

	fields[10].fname = MLX5DV_HWS_FNAME_IP_SPORT;
	fields[10].header = MLX5DV_HWS_HEADER_OUTER;
	fields[10].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[10].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[10].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, sport);
	DEVX_SET(my_ct_v6_item, mask, sport, -1);

	fields[11].fname = MLX5DV_HWS_FNAME_IP_DPORT;
	fields[11].header = MLX5DV_HWS_HEADER_OUTER;
	fields[11].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[11].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[11].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, dport);
	DEVX_SET(my_ct_v6_item, mask, dport, -1);

	fields[12].fname = MLX5DV_HWS_FNAME_IP_VERSION;
	fields[12].header = MLX5DV_HWS_HEADER_OUTER;
	fields[12].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[12].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[12].value.bit_off = CALC_BIT_OFF(my_ct_v6_item, ipver);
	DEVX_SET(my_ct_v6_item, mask, ipver, -1);

	return 0;
}

static void set_match_val(struct mlx5dv_hws_ct_match *match_val, int base_val,
			  bool is_v6)
{
	int i;

	if (is_v6) {
		for (i = 0; i < 16; i++) {
			match_val->ctv6.src_addr[i] = base_val + 1;
			match_val->ctv6.dst_addr[i] = base_val + 2;
		}
		match_val->ctv6.protocol = 3;
		match_val->ctv6.metadata = base_val;
		match_val->ctv6.dst_port = 0x1400;
		match_val->ctv6.src_port = 0x5000;
	} else {
		match_val->ctv4.dst_addr = base_val + 1;
		match_val->ctv4.src_addr = base_val + 2;
		match_val->ctv4.protocol = 3;
		match_val->ctv4.metadata = base_val;
	}
}

int test_basic_ct_rules(struct mlx5dv_hws_context *ctx, bool is_ipv6)
{
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	__be64 modify_action_data[128] = {0};
	uint8_t *action_ptr;
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_ct_matcher_attr matcher_ct_attr = {0};
	struct mlx5dv_hws_ct_rule_attr rule_ct_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_resource *all_resources[4];
	struct mlx5dv_hws_resource *resource;
	struct mlx5dv_hws_resource *resource_countr;
	struct mlx5dv_hws_resource_attr res_attr = {0};
	struct mlx5dv_hws_action_data rule_actions[5] = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_at *at, *at_root;
	struct mlx5dv_hws_mt *r_mt, *mt;
	struct mlx5dv_hws_action *actions[7];
	struct mlx5dv_hws_matcher *matcher, *root_matcher;
	struct mlx5dv_hws_item root_item = {0};
	struct mlx5dv_hws_action *tag, *r_tag;
	struct mlx5dv_hws_action *mh;
	struct mlx5dv_hws_action_modify_header pattern;
	struct mlx5dv_hws_action *aso_counter;
	struct mlx5dv_hws_action *drop;
	struct mlx5dv_hws_table *tbl, *root_tbl;
	struct mlx5dv_hws_at_attr attr_at = {0};
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	uint32_t pending_rules = 0;
	uint32_t miss_count = 0;
	uint16_t queue_id = 0;
	uint32_t hw_bp = 0;
	int ret = 0, i = 0;
	struct mlx5dv_hws_table *tbl3;
	struct mlx5dv_hws_action *dest_ft;
	enum mlx5dv_hws_action_flags flags = MLX5DV_HWS_ACTION_FLAG_NIC_RX;
	struct mlx5dv_hws_ct_match match_val = {0};
	struct mlx5dv_hws_ct_action_data actions_data = {0};

	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;

	tbl3 = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl3) {
		TEST_LOG("Failed to create table for action");
		return -1;
	}

	printf("Table created\n");

	dest_ft = mlx5dv_hws_action_create_dest_table(ctx, tbl3, flags);
	if (!dest_ft) {
		TEST_LOG("Failed to create table action");
		return -1;
	}

	printf("action goto table created\n");


	drop = mlx5dv_hws_action_create_dest_drop(ctx, 1);
	if (!drop) {
		TEST_LOG("Failed to create drop action");
		return -1;
	}

	printf("action goto drop created\n");

	/* Create Root Table */
	tbl_attr.level = 0;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	root_tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!root_tbl) {
		TEST_LOG("Failed to create root table");
		return -1;
	}

	printf("Root Table created\n");

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		return -1;
	}

	printf("Table created\n");

	tag = mlx5dv_hws_action_create_tag(ctx, 1);
	if (!tag) {
		TEST_LOG("Failed to create tag action");
		return -1;
	}

	printf("action tag created\n");

	r_tag = mlx5dv_hws_action_create_tag(ctx, MLX5DV_HWS_ACTION_FLAG_NIC_RX |
						  MLX5DV_HWS_ACTION_FLAG_ROOT);
	if (!r_tag) {
		TEST_LOG("Failed to create root tag action");
		return -1;
	}

	printf("action root tag created\n");

	res_attr.bulk_log_size  = 10;
	res_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B;
	resource = mlx5dv_hws_resource_alloc(ctx, &res_attr);
	if (!resource) {
		TEST_LOG("Failed to create resource ARG!!");
		return -1;
	}

	printf("ARG resource created\n");

	res_attr.bulk_log_size = 10;
	res_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER;
	resource_countr = mlx5dv_hws_resource_alloc(ctx, &res_attr);
	if (!resource_countr) {
		TEST_LOG("Failed to create resource CTR!!");
		return -1;
	}

	printf("resource CTR created\n");

	all_resources[0] = resource_countr;
	aso_counter = mlx5dv_hws_action_create_counter(ctx, all_resources, 1, 1);
	if (!aso_counter) {
		TEST_LOG("Failed to create resource CTR!!");
		return -1;
	}

	all_resources[0] = resource;
	pattern.sz = 16;
	pattern.modify_actions = modify_action_data;

	action_ptr = (uint8_t *) modify_action_data;
	DEVX_SET(set_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
	DEVX_SET(set_action_in, action_ptr, field, MLX5_MODI_OUT_SMAC_15_0);
	DEVX_SET(set_action_in, action_ptr, length, 8);
	action_ptr += 8;

	DEVX_SET(set_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_NOP);
	DEVX_SET(set_action_in, action_ptr, field, MLX5_MODI_OUT_SMAC_47_16);
	DEVX_SET(set_action_in, action_ptr, length, 0);
	DEVX_SET(set_action_in, action_ptr, data, 0xaabbccdd);

	mh = mlx5dv_hws_action_create_modify_header(ctx, all_resources, 1, &pattern, 1, 1);
	if (!mh) {
		TEST_LOG("Failed to create modify-header action!!");
		return -1;
	}

	mh = mlx5dv_hws_action_create_inline(ctx, modify_action_data[0], true,
					     MLX5DV_HWS_ACTION_FLAG_NIC_RX);

	if (!mh) {
		TEST_LOG("Failed to create modify-header inline action !!");
		return -1;
	}

	actions[0] = tag;
	actions[1] = mh;
	actions[2] = aso_counter;

	/* Create action template */
	at = mlx5dv_hws_action_template_create(actions, 3, &attr_at);
	if (!at) {
		TEST_LOG("Failed to create at");
		return -1;
	}

	printf("action template created\n");

	actions[0] = r_tag;
	at_root = mlx5dv_hws_action_template_create(actions, 1, &attr_at);
	if (!at) {
		TEST_LOG("Failed to create at root");
		return -1;
	}

	printf("action template created\n");

	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto clean_at;
	}

	printf("Table created\n");

	mt_attr.comp_mask = 0;
	r_mt = mlx5dv_hws_match_template_create(ctx, NULL, 0, &mt_attr);
	if (!r_mt) {
		TEST_LOG("Failed to create root match template");
		return 1;
	}

	printf("Root Match template created\n");

	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG / 2;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;
	root_matcher = mlx5dv_hws_matcher_create(root_tbl, &r_mt, 1, &at_root, 1, &matcher_attr);
	if (!root_matcher) {
		TEST_LOG("Failed to create root matcher");
		return -1;
	}

	printf("Root Matcher created\n");


	ret = is_ipv6 ? create_ct_ipv6_item(&item) : create_ct_ipv4_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		return -1;
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		return -1;
	}
	matcher_attr.rule.num_log = ROW_LOG;
	if (is_ipv6)
		matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 3;
	matcher_ct_attr.attr = &matcher_attr;
	matcher_ct_attr.ct_type = is_ipv6 ? MLX5DV_HWS_CT_TYPE_V6 : MLX5DV_HWS_CT_TYPE_V4;
	matcher_ct_attr.bi_direction = true;
	matcher = mlx5dv_hws_ct_matcher_create(tbl, mt, &matcher_ct_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher");
		return -1;
	}

	printf("Matcher created\n");

	actions[0] = tag;
	actions[1] = mh;
	actions[2] = aso_counter;

	ret = mlx5dv_hws_ct_actions_bind(matcher, actions, 3);
	if (ret) {
		TEST_LOG("Failed to bind actions array");
		return -1;
	}

	rules = calloc(NUM_OF_RULES, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		return -1;
	}

	rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules);
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.queue_id = queue_id;
	ret = 0;

	ret = mlx5dv_hws_rule_create(root_matcher, 0, &root_item.data, 0,
				     rule_actions, drop, &rule_attr, rule);
	if (ret)
		TEST_LOG("Failed to create root rule %d", i);

	pending_rules = 1;
	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, 1, &miss_count, &hw_bp, true);
	pending_rules = 0;

	dump_all(ctx);

	action_ptr = (uint8_t *) modify_action_data;

	actions_data.ctr.action = aso_counter;
	actions_data.ctr.offset = 2;
	actions_data.mh.mode = MLX5DV_HWS_CT_MH_MODE_DATA_OPTIMIZED;//MLX5DV_HWS_CT_MH_MODE_DATA_INLINE;
	//MLX5DV_HWS_CT_MH_MODE_DATA_OPTIMIZED//MLX5DV_HWS_CT_MH_MODE_DATA_NONE;
	//actions_data.mh.info.data.modify_data = action_ptr;
	actions_data.mh.info.optimized.modify_value = htobe32(7);
	actions_data.mh.info.optimized.modify_reverse_value = htobe32(9);
	actions_data.mh.action = mh;
	actions_data.mh.action_reverse = mh;
	//actions_data.mh.info.data.modify_reverse_data = action_ptr;
	actions_data.tag.action = tag;
	actions_data.tag.tag_value = htobe32(4);
	actions_data.dest_table = dest_ft;
	actions_data.dest_table_reverse = dest_ft;

	for (i = 0; i < NUM_OF_RULES / 4 ; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules + i *
							mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;
		rule_ct_attr.attr = &rule_attr;
		rule_ct_attr.bi_direction = true;
		if (is_ipv6)
			rule_ct_attr.attr->rule_idx = i;

		set_match_val(&match_val, i, is_ipv6);

		ret = mlx5dv_hws_ct_rule_create(matcher, &match_val, &actions_data, &rule_ct_attr, rule);
		if (ret) {
			TEST_LOG("Failed to create rule %d", i);
		} else {
			pending_rules++;
			//pending_rules = 1;
			poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, /*1*/BURST_TH,
				      &miss_count, &hw_bp, false /*true*/);
		}
	}
	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count,
		      &hw_bp, true);
	printf("queue [%d] Insertion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	for (i = 0; i < NUM_OF_RULES / 4 ; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules + i *
							mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;
		rule_ct_attr.attr = &rule_attr;
		rule_ct_attr.bi_direction = true;
		if (is_ipv6)
			rule_ct_attr.attr->rule_idx = i;

		set_match_val(&match_val, i, is_ipv6);

		ret = mlx5dv_hws_ct_rule_destroy(rule, &match_val, &rule_ct_attr);
		if (ret)
			TEST_LOG("Failed to destroy rule %d", i);
		else {
			pending_rules++;
			//pending_rules = 1;
			poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, /*1*/BURST_TH,
					&miss_count, &hw_bp, false /*true*/);
		}
	}
	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH,
		      &miss_count, &hw_bp, true);
	printf("queue [%d] Deletion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	free(rules);

	ret = mlx5dv_hws_ct_actions_unbind(matcher, actions, 3);
	if (ret) {
		TEST_LOG("Failed to UNbind actions array");
		return -1;
	}

	mlx5dv_hws_action_destroy(tag);
	mlx5dv_hws_action_destroy(r_tag);
	mlx5dv_hws_action_destroy(aso_counter);
	mlx5dv_hws_action_destroy(mh);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_matcher_destroy(matcher);
	mlx5dv_hws_matcher_destroy(root_matcher);
	mlx5dv_hws_table_destroy(tbl);
	return 0;

clean_at:
	mlx5dv_hws_action_template_destroy(at);
	return ret;
}

static int check_support_jumbo(struct ibv_context *ibv_ctx, uint32_t *supp_jumbo)
{
	uint32_t out[DEVX_ST_SZ_DW(query_hca_cap_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(query_hca_cap_in)] = {0};
	int ret;

	DEVX_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
	DEVX_SET(query_hca_cap_in, in, op_mod,
		 MLX5_GET_HCA_CAP_OP_MOD_GENERAL_DEVICE_2 |
		 MLX5_HCA_CAP_OPMOD_GET_CUR);

	ret = mlx5dv_devx_general_cmd(ibv_ctx, in, sizeof(in), out, sizeof(out));
	if (ret) {
		printf("Failed to query device caps");
		return ret;
	}

	*supp_jumbo =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap_2.format_select_dw_8_6_ext);

	return 0;
}

int test_ct(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_context *ctx;
	uint32_t supp_jumbo = 0;
	int ret;


	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	}

	printf("Context created\n");
	check_support_jumbo(ibv_ctx, &supp_jumbo);

	ret = test_basic_ct_rules(ctx, false);
	if (!ret)
		TEST_LOG("Passed CT-V4");

	if (supp_jumbo && !ret) {
		ret = test_basic_ct_rules(ctx, true);
		if (!ret)
			TEST_LOG("Passed CT-V6");
	}

	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return ret;

dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	return -errno;
}
