/* Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"

#define NUM_OF_QUEUES 16
#define QUEUE_SIZE 256

int test_dest_ft(struct mlx5dv_hws_context *ctx)
{
	struct mlx5dv_hws_table *tbl;
	struct mlx5dv_hws_action *dest_ft;
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	enum mlx5dv_hws_action_flags flags = MLX5DV_HWS_ACTION_FLAG_NIC_RX;
	int ret = 0;

	TEST_LOG("Test dest table action creation");

	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;

	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		return -1;
	}

	printf("Table created\n");

	dest_ft = mlx5dv_hws_action_create_dest_table(ctx, tbl, flags);
	if (!dest_ft) {
		TEST_LOG("Failed to create table action");
		ret = -1;
		goto out_clean_tbl;
	}

	printf("action goto table created\n");

	mlx5dv_hws_action_destroy(dest_ft);
out_clean_tbl:
	mlx5dv_hws_table_destroy(tbl);
	return ret;
}

int test_dest_drop(struct mlx5dv_hws_context *ctx)
{
	enum mlx5dv_hws_action_flags flags = MLX5DV_HWS_ACTION_FLAG_NIC_RX;
	struct mlx5dv_hws_action *dest;
	int ret = 0;

	TEST_LOG("Test dest drop action creation");

	dest = mlx5dv_hws_action_create_dest_drop(ctx, flags);
	if (!dest) {
		TEST_LOG("Failed to create drop action");
		return -1;
	}

	printf("action goto drop created\n");

	mlx5dv_hws_action_destroy(dest);
	return ret;
}

int test_dest_default_miss(struct mlx5dv_hws_context *ctx)
{
	enum mlx5dv_hws_action_flags flags = MLX5DV_HWS_ACTION_FLAG_NIC_TX;
	struct mlx5dv_hws_action *dest;
	int ret = 0;

	TEST_LOG("Test dest default miss action creation");

	dest = mlx5dv_hws_action_create_dest_default_miss(ctx, flags);
	if (!dest) {
		TEST_LOG("Failed to create action");
		return -1;
	}

	printf("action goto default - miss created\n");

	mlx5dv_hws_action_destroy(dest);
	return ret;
}

static int test_dest_root(struct mlx5dv_hws_context *ctx)
{
	enum mlx5dv_hws_action_flags flags = MLX5DV_HWS_ACTION_FLAG_NIC_RX;
	struct mlx5dv_hws_action *dest;
	int ret = 0;

	TEST_LOG("Test dest root action creation");

	dest = mlx5dv_hws_action_create_dest_root(ctx, 2, flags);
	if (!dest) {
		TEST_LOG("Failed to create action goto root");
		return -1;
	}

	printf("action goto root - action created\n");

	mlx5dv_hws_action_destroy(dest);
	return ret;
}

static int test_dest_simple(struct mlx5dv_hws_context *ctx)
{
	enum mlx5dv_hws_action_flags flags = MLX5DV_HWS_ACTION_FLAG_NIC_RX;
	struct mlx5dv_hws_action *dest_arr, *dest_arr2, *dest_tbl1, *dest_tbl2;
	struct mlx5dv_hws_action_dest_attr dests[2] = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_table *tbl1, *tbl2;
	int ret = 0;

	TEST_LOG("Test multi dest array to 2 tables");

	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;

	tbl1 = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl1) {
		TEST_LOG("Failed to create table1");
		return -1;
	}
	printf("Table1 created\n");

	tbl_attr.level = 2;
	tbl2 = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl2) {
		TEST_LOG("Failed to create table2");
		goto out_destroy_tbl1;
	}
	printf("Table2 created\n");

	dest_tbl1 = mlx5dv_hws_action_create_dest_table(ctx, tbl1, flags);
	if (!dest_tbl1) {
		TEST_LOG("Failed to create table 1 action");
		ret = -1;
		goto out_destroy_tbl2;
	}
	printf("Table1 action created\n");

	dest_tbl2 = mlx5dv_hws_action_create_dest_table(ctx, tbl2, flags);
	if (!dest_tbl2) {
		TEST_LOG("Failed to create table 1 action");
		ret = -1;
		goto out_destroy_act_tbl1;
	}
	printf("Table2 action created\n");

	dests[0].dest_action = dest_tbl1;
	dests[1].dest_action = dest_tbl2;
	dest_arr = mlx5dv_hws_action_create_dest_array(ctx, 2, dests, flags);
	if (!dest_arr) {
		TEST_LOG("Failed to create multi dest action");
		ret = -1;
		goto out_destroy_act_tbl2;
	}
	printf("action goto dest array created\n");

	dest_arr2 = mlx5dv_hws_action_create_dest_array(ctx, 2, dests, flags);
	if (!dest_arr2) {
		TEST_LOG("Failed to create multi dest action 2");
		ret = -1;
		goto out_destroy_act_tbl2;
	}

	mlx5dv_hws_action_destroy(dest_arr);
	mlx5dv_hws_action_destroy(dest_arr2);

out_destroy_act_tbl2:
	mlx5dv_hws_action_destroy(dest_tbl2);
out_destroy_act_tbl1:
	mlx5dv_hws_action_destroy(dest_tbl1);
out_destroy_tbl2:
	mlx5dv_hws_table_destroy(tbl2);
out_destroy_tbl1:
	mlx5dv_hws_table_destroy(tbl1);
	return ret;
}

#ifdef FDB_TEST
static int
test_dest_fdb(struct mlx5dv_hws_context *ctx)
{
	struct mlx5dv_hws_action *dest_arr, *dest_tbl1, *dest_vport;
	enum mlx5dv_hws_action_flags flags = MLX5DV_HWS_ACTION_FLAG_FDB;
	struct mlx5dv_hws_action_dest_attr dests[3] = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_table *tbl1;
	uint32_t uplink_vport = 4; /* Too lazy to get cap */
	int ret = 0;

	uint8_t hdr2[18] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17};
	uint8_t hdr1[14] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13};

	TEST_LOG("FDB dest array to 2 tables with reformat");

	/* This is not part of quick tests since it requires switchdev,
	 * and currently has hard coded values for uplink.
	 */

	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_FDB;

	tbl1 = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl1) {
		TEST_LOG("Failed to create table1");
		return -1;
	}
	printf("Table1 created\n");

	dest_tbl1 = mlx5dv_hws_action_create_dest_table(ctx, tbl1, flags);
	if (!dest_tbl1) {
		TEST_LOG("Failed to create table 1 action");
		ret = -1;
		goto out_destroy_tbl1;
	}
	printf("Table1 action created\n");

	dest_vport = mlx5dv_hws_action_create_dest_vport(ctx, uplink_vport, flags);
	if (!dest_vport) {
		TEST_LOG("Failed to create dest vport action");
		ret = -1;
		goto out_destroy_act_tbl1;
	}
	printf("Dest vport action created\n");

	dests[0].dest_action = dest_vport;
	dests[0].reformat.type = MLX5DV_HWS_ACTION_DEST_REFORMAT_TYPE_L2_TO_TNL_L2;
	dests[0].reformat.hdr_data.hdr_data = hdr1;
	dests[0].reformat.hdr_data.sz = 14;

	dests[1].dest_action = dest_vport;
	dests[1].reformat.type = MLX5DV_HWS_ACTION_DEST_REFORMAT_TYPE_L2_TO_TNL_L3;
	dests[1].reformat.hdr_data.hdr_data = hdr2;
	dests[1].reformat.hdr_data.sz = 18;

	dests[2].dest_action = dest_tbl1;
	dests[2].reformat.type = MLX5DV_HWS_ACTION_DEST_REFORMAT_TYPE_NONE;

	dest_arr = mlx5dv_hws_action_create_dest_array(ctx, 3, dests, flags);
	if (!dest_arr) {
		TEST_LOG("Failed to create multi dest action");
		ret = -1;
		goto out_destroy_vport;
	}
	TEST_LOG("action goto dest array created\n");

	mlx5dv_hws_action_destroy(dest_arr);

out_destroy_vport:
	mlx5dv_hws_action_destroy(dest_vport);
out_destroy_act_tbl1:
	mlx5dv_hws_action_destroy(dest_tbl1);
out_destroy_tbl1:
	mlx5dv_hws_table_destroy(tbl1);
	return ret;
}
#endif /* FDB_TEST */

int test_dest(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_context *ctx;
	int ret;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	}

	printf("Context created\n");

	ret = test_dest_ft(ctx);
	if (!ret)
		ret = test_dest_drop(ctx);

	if (!ret)
		ret = test_dest_default_miss(ctx);

	if (!ret)
		ret = test_dest_simple(ctx);

	if (!ret)
		ret = test_dest_root(ctx);

	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return ret;

dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	return -errno;
}
