/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"

#define NUM_OF_QUEUES 16
#define ROW_LOG 10
#define QUEUE_SIZE 256
#define BURST_TH 1
#define NUM_INLINE_ACTION 5
#define VLAN_TPID 0x8100
#define VLAN_PRIO 4
#define VLAN_CFI 1
#define VLAN_ID 0xc0c
#define VLAN_START_IDX 12
#define W_SIZE 2
#define MLX5_IFC_STC_ACTION_TYPE_BIT_FIELD_COPY (1ULL << (64 - 5))

#define MLX5DV_HWS_ACTION_SET(action_ptr, field_v, length_v, offset_v, data_v) do { \
	DEVX_SET(set_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET); \
	DEVX_SET(set_action_in, action_ptr, field, field_v); \
	DEVX_SET(set_action_in, action_ptr, length, length_v); \
	DEVX_SET(set_action_in, action_ptr, offset, offset_v); \
	DEVX_SET(set_action_in, action_ptr, data, data_v); \
} while (0)

#define MLX5DV_HWS_ACTION_ADD(action_ptr, field_v, length_v, offset_v, data_v) do { \
	DEVX_SET(add_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_ADD); \
	DEVX_SET(add_action_in, action_ptr, field, field_v); \
	DEVX_SET(add_action_in, action_ptr, length, length_v); \
	DEVX_SET(add_action_in, action_ptr, offset, offset_v); \
	DEVX_SET(add_action_in, action_ptr, data, data_v); \
} while (0)

#define MLX5DV_HWS_ACTION_ADD_FIELD(action_ptr, src_field_v, src_offset_v, length_v, dst_field_v, \
				    dst_left_shift_v) \
do { \
	DEVX_SET(add_field_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_ADD_FIELD); \
	DEVX_SET(add_field_in, action_ptr, src_field, src_field_v); \
	DEVX_SET(add_field_in, action_ptr, src_offset, src_offset_v); \
	DEVX_SET(add_field_in, action_ptr, length, length_v); \
	DEVX_SET(add_field_in, action_ptr, dst_field, dst_field_v); \
	DEVX_SET(add_field_in, action_ptr, dst_left_shift, dst_left_shift_v); \
} while (0)

#define MLX5DV_HWS_ACTION_COPY(action_ptr, src_field_v, src_offset_v, \
			       length_v, dst_field_v, dst_offset_v) \
do { \
	DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_COPY); \
	DEVX_SET(copy_action_in, action_ptr, src_field, src_field_v); \
	DEVX_SET(copy_action_in, action_ptr, src_offset, src_offset_v); \
	DEVX_SET(copy_action_in, action_ptr, length, length_v); \
	DEVX_SET(copy_action_in, action_ptr, dst_field, dst_field_v); \
	DEVX_SET(copy_action_in, action_ptr, dst_offset, dst_offset_v); \
} while (0)

#define MLX5DV_HWS_ACTION_INSERT(action_ptr, push_esp_v, insert_anchor_v, insert_offset_v, \
				 insert_argument_v) \
do { \
	DEVX_SET(stc_ste_param_insert, action_ptr, action_type, MLX5_MODIFICATION_TYPE_INSERT); \
	/* If encap set, inline_data must be 0x0 */ \
	DEVX_SET(stc_ste_param_insert, action_ptr, encap, false); \
	/* Setter for inline insert */ \
	DEVX_SET(stc_ste_param_insert, action_ptr, inline_data, true); \
	DEVX_SET(stc_ste_param_insert, action_ptr, push_esp, push_esp_v); \
	DEVX_SET(stc_ste_param_insert, action_ptr, insert_anchor, insert_anchor_v); \
	DEVX_SET(stc_ste_param_insert, action_ptr, insert_offset, insert_offset_v); \
	/* Insert header size, given in Words. For inline insertion must be 2 */ \
	DEVX_SET(stc_ste_param_insert, action_ptr, insert_size, 0x2 * W_SIZE); \
	DEVX_SET(stc_ste_param_insert, action_ptr, insert_argument, insert_argument_v); \
} while (0)

static int check_copy_add_field_support(struct ibv_context *ibv_ctx, bool *copy_add_supp)
{
	uint32_t out[DEVX_ST_SZ_DW(query_hca_cap_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(query_hca_cap_in)] = {0};
	uint64_t stc_action_type_63_0;
	int ret;

	DEVX_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
	DEVX_SET(query_hca_cap_in, in, op_mod,
		 MLX5_GET_HCA_CAP_OP_MOD_WQE_BASED_FLOW_TABLE |
		 MLX5_HCA_CAP_OPMOD_GET_CUR);

	ret = mlx5dv_devx_general_cmd(ibv_ctx, in, sizeof(in), out, sizeof(out));
	if (ret)
		return ret;

	stc_action_type_63_0 =
		DEVX_GET64(query_hca_cap_out, out,
			   capability.wqe_based_flow_table_cap.stc_action_type_63_0);

	*copy_add_supp = stc_action_type_63_0 & MLX5_IFC_STC_ACTION_TYPE_BIT_FIELD_COPY;

	return 0;
}

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			printf("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

static uint32_t
create_vlan_hdr(uint16_t tpi, uint8_t prio, uint8_t cfi, uint16_t vid)
{
	return ((uint32_t)tpi << 16) | ((uint32_t)prio << 13) |
	       ((uint32_t)cfi << 12) | vid;
}

int test_inline_action(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_action_data action_data[NUM_INLINE_ACTION] = {0};
	struct mlx5dv_hws_action *actions[NUM_INLINE_ACTION] = {0};
	struct mlx5dv_hws_at *root_at, *hws_at[NUM_INLINE_ACTION];
	uint64_t modify_action_data[NUM_INLINE_ACTION] = {0};
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_matcher *matcher, *root_matcher;
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_table *tbl, *root_tbl;
	struct mlx5dv_hws_action *dest_action;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	uint32_t pending_rules = 0;
	struct mlx5dv_hws_mt *mt;
	uint32_t miss_count = 0;
	uint16_t queue_id = 0;
	uint32_t vlan_header;
	uint32_t hw_bp = 0;
	bool copy_add_supp;
	int ret, i, n_actions;

	if (check_copy_add_field_support(ibv_ctx, &copy_add_supp)) {
		TEST_LOG("Failed to query support cap");
		goto out_err;
	}

	n_actions = copy_add_supp ? NUM_INLINE_ACTION : NUM_INLINE_ACTION - 2;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	} else {
		printf("Context created\n");
	}

	/* Create Root Table */
	tbl_attr.level = 0;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	root_tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!root_tbl) {
		TEST_LOG("Failed to create root table");
		goto ctx_close;
	} else {
		printf("Root Table created\n");
	}

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto root_tbl_destroy;
	} else {
		printf("Table created\n");
	}

	ret = create_basic_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto tbl_destroy;
	} else {
		printf("Item created\n");
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		goto item_destroy;
	} else {
		printf("Match template created\n");
	}

	root_at = mlx5dv_hws_action_template_create(NULL, 0, &at_attr);
	if (!root_at) {
		TEST_LOG("Failed to create dummy action template");
		goto mt_destroy;
	} else {
		printf("Dummy action template created\n");
	}

	/* Create ROOT Matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;
	root_matcher = mlx5dv_hws_matcher_create(root_tbl, &mt, 1, &root_at, 1, &matcher_attr);
	if (!root_matcher) {
		TEST_LOG("Failed to create root matcher");
		goto root_at_destroy;
	} else {
		printf("Root Matcher created\n");
	}

	MLX5DV_HWS_ACTION_SET(&modify_action_data[0], MLX5_MODI_OUT_TCP_ACK_NUM, 2, 30, 5);
	DEVX_SET(set_action_in, &action_data[0].inline_action.data, data, 5);

	MLX5DV_HWS_ACTION_ADD(&modify_action_data[1], MLX5_MODI_OUT_IPV4_TTL, 0, 0, 0xab);
	DEVX_SET(add_action_in, &action_data[1].inline_action.data, data, 0xab);

	vlan_header = create_vlan_hdr(VLAN_TPID, VLAN_PRIO, VLAN_CFI, VLAN_ID);
	MLX5DV_HWS_ACTION_INSERT(&modify_action_data[2], false,
				 MLX5DV_HWS_ACTION_ANCHOR_PACKET_START, VLAN_START_IDX,
				 vlan_header);
	DEVX_SET(stc_ste_param_insert, &action_data[2].inline_action.data, insert_argument,
		 vlan_header);

	if (copy_add_supp) {
		MLX5DV_HWS_ACTION_ADD_FIELD(&modify_action_data[3], MLX5_MODI_OUT_IPV4_IHL, 0, 4,
					    MLX5_MODI_META_REG_C_0, 1);

		MLX5DV_HWS_ACTION_COPY(&modify_action_data[4], MLX5_MODI_OUT_TCP_DATA_OFFSET, 0,
				       4, MLX5_MODI_META_REG_C_0, 0);
	}

	for (i = 0; i < n_actions; i++) {
		actions[i] = mlx5dv_hws_action_create_inline(ctx,
							     modify_action_data[i],
							     true,
							     MLX5DV_HWS_ACTION_FLAG_NIC_RX);
		if (!actions[i]) {
			TEST_LOG("Failed to create inline actions %d", i);
			goto destroy_inline_actions;
		}

		hws_at[i] = mlx5dv_hws_action_template_create(&actions[i], 1, &at_attr);
		if (!hws_at[i]) {
			TEST_LOG("Failed to create inline action template %d", i);
			mlx5dv_hws_action_destroy(actions[i]);
			i--;
			goto destroy_inline_actions;
		}
	}
	printf("inline action template created\n");

	/* Create HWS Matcher */
	matcher_attr.optimize_using_rule_idx = 1;
	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, hws_at, n_actions, &matcher_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher");
		goto destroy_inline_actions;
	} else {
		printf("Matcher created\n");
	}

	dest_action = mlx5dv_hws_action_create_dest_default_miss(ctx,
								 MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!dest_action) {
		TEST_LOG("Failed to create default miss action\n");
		goto matcher_destroy;
	}

	/* Dump once all control data created */
	dump_all(ctx);

	rules = calloc(n_actions, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		goto default_miss_destroy;
	}

	rule_attr.queue_id = queue_id;

	for (i = 0; i < n_actions; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules +
			i * mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		DEVX_SET(my_basic_item, (void *)item.data.match, sip, i);
		ret = mlx5dv_hws_rule_create(matcher, 0, &item.data, i, &action_data[i],
					     dest_action, &rule_attr, rule);
		if (ret)
			TEST_LOG("Failed to create rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count,
			      &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Insertion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		n_actions,
		hw_bp);

	for (i = 0; i < n_actions; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules +
			i * mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
		if (ret)
			TEST_LOG("Failed to destroy rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count,
			      &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count,
		      &hw_bp, true);
	printf("queue [%d] Deletion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		n_actions,
		hw_bp);

	free(rules);
	mlx5dv_hws_action_destroy(dest_action);
	mlx5dv_hws_matcher_destroy(root_matcher);
	for (i = 0; i < n_actions; i++) {
		mlx5dv_hws_action_template_destroy(hws_at[i]);
		mlx5dv_hws_action_destroy(actions[i]);
	}

	mlx5dv_hws_matcher_destroy(matcher);
	mlx5dv_hws_action_template_destroy(root_at);
	mlx5dv_hws_match_template_destroy(mt);
	destroy_basic_item(&item);
	mlx5dv_hws_table_destroy(tbl);
	mlx5dv_hws_table_destroy(root_tbl);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return 0;

default_miss_destroy:
	mlx5dv_hws_action_destroy(dest_action);
matcher_destroy:
	mlx5dv_hws_matcher_destroy(matcher);
destroy_inline_actions:
	while (i--) {
		mlx5dv_hws_action_template_destroy(hws_at[i]);
		mlx5dv_hws_action_destroy(actions[i]);
	}
	mlx5dv_hws_matcher_destroy(root_matcher);
root_at_destroy:
	mlx5dv_hws_action_template_destroy(root_at);
mt_destroy:
	mlx5dv_hws_match_template_destroy(mt);
item_destroy:
	destroy_basic_item(&item);
tbl_destroy:
	mlx5dv_hws_table_destroy(tbl);
root_tbl_destroy:
	mlx5dv_hws_table_destroy(root_tbl);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	TEST_LOG("Test failed errno=%d", errno);
	return -errno;
}
