/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"

#define NUM_OF_QUEUES 16
#define ROW_LOG 10
#define NUM_OF_RULES (1 << ROW_LOG)
#define BURST_TH 32
#define QUEUE_SIZE 256

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			printf("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

int test_isolated_matcher(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_action *dest_matcher, *default_miss;
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_matcher_attr ste_arr_attr = {0};
	struct mlx5dv_hws_rule *rules, *ste_arr_rules;
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_action_data action_data[1];
	struct mlx5dv_hws_matcher *matcher, *ste_arr;
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_rule *rule;
	struct mlx5dv_hws_table *tbl;
	uint32_t pending_rules = 0;
	struct mlx5dv_hws_at *at;
	struct mlx5dv_hws_mt *mt;
	uint32_t miss_count = 0;
	uint16_t queue_id = 0;
	uint32_t hw_bp = 0;
	int ret, i;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	} else {
		printf("Context created\n");
	}

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto ctx_close;
	} else {
		printf("Table created\n");
	}

	ret = create_basic_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto tbl_destroy;
	} else {
		printf("Item created\n");
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		goto item_destroy;
	} else {
		printf("Match template created\n");
	}

	at = mlx5dv_hws_action_template_create(NULL, 0, &at_attr);
	if (!at) {
		TEST_LOG("Failed to create dummy action template");
		goto mt_destroy;
	} else {
		printf("Dummy action template created\n");
	}

	/* Create first matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &matcher_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher");
		goto at_destroy;
	} else {
		printf("Matcher created\n");
	}

	/* Create STE array matcher */
	ste_arr_attr.comp_mask = 0;
	ste_arr_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_HTABLE;
	ste_arr_attr.insert_mode = MLX5DV_HWS_MATCHER_INSERT_BY_INDEX;
	ste_arr_attr.distribute_mode = MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH;
	ste_arr_attr.isolated = true;
	ste_arr_attr.table.sz_row_log = ROW_LOG;
	ste_arr_attr.table.sz_col_log = 0;
	ste_arr_attr.optimize_using_rule_idx = 1;
	ste_arr = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &ste_arr_attr);
	if (!ste_arr) {
		TEST_LOG("Failed to create STE array matcher");
		goto matcher_destroy;
	} else {
		printf("STE array matcher created\n");
	}

	dest_matcher = mlx5dv_hws_action_create_dest_matcher(ctx,
							     ste_arr,
							     MLX5DV_HWS_ACTION_DEST_MATCHER_BY_INDEX,
							     MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!dest_matcher) {
		TEST_LOG("Failed to create default miss action\n");
		goto ste_arr_destroy;
	}

	default_miss = mlx5dv_hws_action_create_dest_default_miss(ctx,
								  MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!default_miss) {
		TEST_LOG("Failed to create default miss action\n");
		goto dest_matcher_destroy;
	}

	/* Dump once all control data created */
	dump_all(ctx);

	/* Fill STE array */
	ste_arr_rules = calloc(NUM_OF_RULES, mlx5dv_hws_rule_get_handle_size());
	if (!ste_arr_rules) {
		TEST_LOG("Failed to create STE array rule array");
		goto default_miss_destroy;
	}

	rule_attr.queue_id = queue_id;

	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)ste_arr_rules +
						  (i * mlx5dv_hws_rule_get_handle_size()));
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		DEVX_SET(my_basic_item, (void *)item.data.match, sip, i);
		ret = mlx5dv_hws_rule_create(ste_arr, 0, &item.data, 0, NULL,
					     default_miss, &rule_attr, rule);
		if (ret)
			TEST_LOG("Failed to create rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules,
			      BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Insertion. Total misses: %u (out of: %u) HW bp %u\n",
	       queue_id,
	       miss_count,
	       NUM_OF_RULES,
	       hw_bp);

	pending_rules = 0;

	/* Create Rules */
	rules = calloc(NUM_OF_RULES, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		goto ste_arr_rules_destroy;
	}

	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules +
						  (i * mlx5dv_hws_rule_get_handle_size()));
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;
		action_data[0].dest_matcher.index = i;

		DEVX_SET(my_basic_item, (void *)item.data.match, sip, 0x10000000 + i);
		ret = mlx5dv_hws_rule_create(matcher, 0, &item.data, 0, action_data,
					     dest_matcher, &rule_attr, rule);
		if (ret)
			TEST_LOG("Failed to create rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules,
			      BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Insertion. Total misses: %u (out of: %u) HW bp %u\n",
	       queue_id,
	       miss_count,
	       NUM_OF_RULES,
	       hw_bp);

	/* Delete rules */
	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules +
						  (i * mlx5dv_hws_rule_get_handle_size()));
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
		if (ret)
			TEST_LOG("Failed to destroy rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules,
			      BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Deletion. Total misses: %u (out of: %u) HW bp %u\n",
	       queue_id,
	       miss_count,
	       NUM_OF_RULES,
	       hw_bp);

	/* Delete STE array rules */
	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)ste_arr_rules +
						  (i * mlx5dv_hws_rule_get_handle_size()));
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
		if (ret)
			TEST_LOG("Failed to destroy rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules,
			      BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Deletion. Total misses: %u (out of: %u) HW bp %u\n",
	       queue_id,
	       miss_count,
	       NUM_OF_RULES,
	       hw_bp);

	free(rules);
	free(ste_arr_rules);
	mlx5dv_hws_action_destroy(default_miss);
	mlx5dv_hws_action_destroy(dest_matcher);
	mlx5dv_hws_matcher_destroy(matcher);
	mlx5dv_hws_matcher_destroy(ste_arr);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_match_template_destroy(mt);
	destroy_basic_item(&item);
	mlx5dv_hws_table_destroy(tbl);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return 0;

ste_arr_rules_destroy:
	free(ste_arr_rules);
default_miss_destroy:
	mlx5dv_hws_action_destroy(default_miss);
dest_matcher_destroy:
	mlx5dv_hws_action_destroy(dest_matcher);
matcher_destroy:
	mlx5dv_hws_matcher_destroy(matcher);
ste_arr_destroy:
	mlx5dv_hws_matcher_destroy(ste_arr);
at_destroy:
	mlx5dv_hws_action_template_destroy(at);
mt_destroy:
	mlx5dv_hws_match_template_destroy(mt);
item_destroy:
	destroy_basic_item(&item);
tbl_destroy:
	mlx5dv_hws_table_destroy(tbl);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	TEST_LOG("Test failed errno=%d", errno);
	return -errno;
}
