/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"

#define NUM_OF_QUEUES 16
#define S_ROW_LOG 10
#define L_ROW_LOG 19
#define QUEUE_SIZE 256
#define S_MATCHER_PRIORITY 33
#define L_MATCHER_PRIORITY 37

int test_metric(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_metric_matcher s_matcher_out = {0};
	struct mlx5dv_hws_metric_matcher l_matcher_out = {0};
	struct mlx5dv_hws_matcher_attr s_matcher_attr = {0};
	struct mlx5dv_hws_matcher_attr l_matcher_attr = {0};
	struct mlx5dv_hws_matcher *s_matcher, *l_matcher;
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_metric_table tbl_out = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_action *dest_action;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_table *tbl;
	struct mlx5dv_hws_at *at;
	struct mlx5dv_hws_mt *mt;
	uint64_t flags;
	int ret;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	} else {
		printf("Context created\n");
	}

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto ctx_close;
	} else {
		printf("Table created\n");
	}

	ret = create_basic_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto tbl_destroy;
	} else {
		printf("Item created\n");
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		goto item_destroy;
	} else {
		printf("Match template created\n");
	}

	at = mlx5dv_hws_action_template_create(NULL, 0, &at_attr);
	if (!at) {
		TEST_LOG("Failed to create dummy action template");
		goto mt_destroy;
	} else {
		printf("Dummy action template created\n");
	}

	/* Create HWS small matcher */
	s_matcher_attr.comp_mask = 0;
	s_matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	s_matcher_attr.rule.num_log = S_ROW_LOG;
	s_matcher_attr.optimize_using_rule_idx = 1;
	s_matcher_attr.priority = S_MATCHER_PRIORITY;
	s_matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &s_matcher_attr);
	if (!s_matcher) {
		TEST_LOG("Failed to create small matcher");
		goto at_destroy;
	} else {
		printf("Small matcher created\n");
	}

	/* Create HWS large matcher */
	l_matcher_attr.comp_mask = 0;
	l_matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	l_matcher_attr.rule.num_log = L_ROW_LOG;
	l_matcher_attr.optimize_using_rule_idx = 1;
	l_matcher_attr.priority = L_MATCHER_PRIORITY;
	l_matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &l_matcher_attr);
	if (!l_matcher) {
		TEST_LOG("Failed to create large matcher");
		goto s_matcher_destroy;
	} else {
		printf("Large matcher created\n");
	}

	dest_action = mlx5dv_hws_action_create_dest_default_miss(ctx,
								 MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!dest_action) {
		TEST_LOG("Failed to create default miss action\n");
		goto l_matcher_destroy;
	}

	if (mlx5dv_hws_metric_query_table(tbl, MLX5DV_HWS_METRIC_FLAG_INFO, &tbl_out)) {
		TEST_LOG("Failed to query table");
		goto l_matcher_destroy;
	} else {
		assert(tbl_out.info.level == 1);
		assert(tbl_out.info.ft_fw_type == MLX5_IFC_FT_NIC_RX);
		printf("Correct table metrics\n");
	}

	flags = MLX5DV_HWS_METRIC_FLAG_INFO | MLX5DV_HWS_METRIC_FLAG_COST;
	if (mlx5dv_hws_metric_query_matcher(s_matcher, flags, &s_matcher_out)) {
		TEST_LOG("Failed to query small matcher");
		goto dest_action_destroy;
	} else {
		assert(s_matcher_out.info.priority == S_MATCHER_PRIORITY);
		assert(s_matcher_out.info.log_num_of_rows == S_ROW_LOG);
		assert(s_matcher_out.info.log_num_of_columns > 0);
		assert(s_matcher_out.info.num_of_match_stes == 1);
		/* No need for action STE as we only have termenation action */
		assert(s_matcher_out.info.num_of_action_stes == 0);
		printf("Correct small matcher metrics\n");
	}

	if (mlx5dv_hws_metric_query_matcher(l_matcher, flags, &l_matcher_out)) {
		TEST_LOG("Failed to query large matcher");
		goto dest_action_destroy;
	} else {
		assert(l_matcher_out.info.priority == L_MATCHER_PRIORITY);
		assert(l_matcher_out.info.log_num_of_rows == L_ROW_LOG);
		assert(l_matcher_out.info.log_num_of_columns > 0);
		assert(l_matcher_out.info.num_of_match_stes == 1);
		/* No need for action STE as we only have termenation action */
		assert(l_matcher_out.info.num_of_action_stes == 0);
		printf("Correct large matcher metrics\n");
	}

	assert(l_matcher_out.cost.memory > s_matcher_out.cost.memory);
	assert(l_matcher_out.cost.miss_pkt_processing > s_matcher_out.cost.miss_pkt_processing);

	printf("Correct matchers cost comparison\n");

	mlx5dv_hws_action_destroy(dest_action);
	mlx5dv_hws_matcher_destroy(l_matcher);
	mlx5dv_hws_matcher_destroy(s_matcher);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_match_template_destroy(mt);
	destroy_basic_item(&item);
	mlx5dv_hws_table_destroy(tbl);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return 0;

dest_action_destroy:
	mlx5dv_hws_action_destroy(dest_action);
l_matcher_destroy:
	mlx5dv_hws_matcher_destroy(l_matcher);
s_matcher_destroy:
	mlx5dv_hws_matcher_destroy(s_matcher);
at_destroy:
	mlx5dv_hws_action_template_destroy(at);
mt_destroy:
	mlx5dv_hws_match_template_destroy(mt);
item_destroy:
	destroy_basic_item(&item);
tbl_destroy:
	mlx5dv_hws_table_destroy(tbl);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	TEST_LOG("Test failed errno=%d", errno);
	return -errno;
}
