/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#define _GNU_SOURCE
#include <sched.h>
#include <pthread.h>
#include <stdatomic.h>
#include "hws_test.h"
#include "../mlx5dv_hws_ct.h"

#define ROW_LOG 20
#define NUM_OF_RULES (1 << ROW_LOG)
#define QUEUE_SIZE 1024
#define BURST_TH 32

#define DEFAULT_NUM_OF_CORES 4
#define THREADS_PER_CORE 1
#define QUEUE_PER_THREAD 1

#define LOOP 20
#define SNAPSHOT_INTERVAL 1000000 //In microseconds
#define K_RULES 1024

pthread_mutex_t count_mutex_ct;

atomic_int global_counter_ct;

cpu_set_t *cpu_set_ct;
int num_of_cores_ct;
int num_of_threads_ct;
struct thread_info *th_info_ct;

struct queue_info {
	uint32_t inserted_rules;
	uint32_t deleted_rules;
	uint32_t miss_count;
	uint32_t total_bp;
	uint32_t queue_id;
	uint32_t pending;
	uint32_t bp;
};

struct thread_info {
	struct queue_info queue[QUEUE_PER_THREAD];
	struct mlx5dv_hws_context *ctx;
	int thread_id;
	const char *thread_name;
	struct mlx5dv_hws_matcher *matcher;
	uint32_t curr_loop;
	struct mlx5dv_hws_action *dest_action;
	struct mlx5dv_hws_action *counter;
	struct mlx5dv_hws_action *tag;
};

struct thread_metrics {
	uint32_t inserted;
	uint32_t deleted;
	uint32_t bp;
};

struct mlx5_ifc_my_ct_item_bits {
	uint8_t sip[0x20];
	uint8_t dip[0x20];
	uint8_t regc[0x20];
	uint8_t sport[0x10];
	uint8_t dport[0x10];
	uint8_t ipver[0x4];
	uint8_t prot[0x8];
};

static int create_ct_ipv4_item(struct mlx5dv_hws_item *item)
{
	struct mlx5dv_hws_field *fields;
	void *mask, *match, *buff;

	fields = calloc(7, sizeof(struct mlx5dv_hws_field));
	if (!fields) {
		TEST_LOG("Failed to allocate fields");
		return ENOMEM;
	}

	buff = calloc(7, DEVX_ST_SZ_BYTES(my_ct_item));
	if (!buff) {
		TEST_LOG("Failed to allocate mask and match");
		free(fields);
		return ENOMEM;
	}

	match = buff;
	mask = buff + DEVX_ST_SZ_BYTES(my_ct_item);

	item->num_of_fields = 7;
	item->fields = fields;
	item->data.mask = mask;
	item->data.match = match;

	fields[0].fname = MLX5DV_HWS_FNAME_IPV4_SRC;
	fields[0].header = MLX5DV_HWS_HEADER_OUTER;
	fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[0].value.bit_off = CALC_BIT_OFF(my_ct_item, sip);
	DEVX_SET(my_ct_item, mask, sip, -1);

	fields[1].fname = MLX5DV_HWS_FNAME_IPV4_DST;
	fields[1].header = MLX5DV_HWS_HEADER_OUTER;
	fields[1].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[1].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[1].value.bit_off = CALC_BIT_OFF(my_ct_item, dip);
	DEVX_SET(my_ct_item, mask, dip, -1);

	fields[2].fname = MLX5DV_HWS_FNAME_REG_C1;
	fields[2].header = MLX5DV_HWS_HEADER_NONE;
	fields[2].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[2].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[2].value.bit_off = CALC_BIT_OFF(my_ct_item, regc);
	DEVX_SET(my_ct_item, mask, regc, -1);

	fields[3].fname = MLX5DV_HWS_FNAME_IP_PROTOCOL;
	fields[3].header = MLX5DV_HWS_HEADER_OUTER;
	fields[3].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[3].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[3].value.bit_off = CALC_BIT_OFF(my_ct_item, prot);
	DEVX_SET(my_ct_item, mask, prot, -1);

	fields[4].fname = MLX5DV_HWS_FNAME_IP_SPORT;
	fields[4].header = MLX5DV_HWS_HEADER_OUTER;
	fields[4].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[4].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[4].value.bit_off = CALC_BIT_OFF(my_ct_item, sport);
	DEVX_SET(my_ct_item, mask, sport, -1);

	fields[5].fname = MLX5DV_HWS_FNAME_IP_DPORT;
	fields[5].header = MLX5DV_HWS_HEADER_OUTER;
	fields[5].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[5].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[5].value.bit_off = CALC_BIT_OFF(my_ct_item, dport);
	DEVX_SET(my_ct_item, mask, dport, -1);

	fields[6].fname = MLX5DV_HWS_FNAME_IP_VERSION;
	fields[6].header = MLX5DV_HWS_HEADER_OUTER;
	fields[6].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[6].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[6].value.bit_off = CALC_BIT_OFF(my_ct_item, ipver);
	DEVX_SET(my_ct_item, mask, ipver, -1);

	return 0;
}

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			printf("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
			/* Give up CPU to other threads instead of busy wait */
			sched_yield();
		}

		got_comp = !!ret;
	}

	return 0;
}

static void *manager_thread(void *arg)
{
	uint32_t total_inserted, _total_inserted;
	uint32_t total_deleted, _total_deleted;
	struct thread_metrics *sm, *em;
	struct timespec start, end;
	struct queue_info *queue;
	struct thread_info *th;
	double time_diff;
	uint32_t diff;
	int i, j;

	sm = calloc(num_of_threads_ct, sizeof(*sm));
	if (!sm) {
		TEST_LOG("failed to allocate thread_metrics");
		return NULL;
	}

	em = calloc(num_of_threads_ct, sizeof(*em));
	if (!em) {
		TEST_LOG("failed to allocate thread_metrics");
		free(sm);
		return NULL;
	}

	printf("Start manager thread, core %u\n", sched_getcpu());
	while (atomic_load(&global_counter_ct) < num_of_threads_ct) {
		_total_inserted = 0;
		_total_deleted = 0;
		for (i = 0; i < num_of_threads_ct; i++) {
			sm[i].inserted = 0;
			sm[i].deleted = 0;
			sm[i].bp = 0;
			th = &th_info_ct[i];
			for (j = 0; j < QUEUE_PER_THREAD; j++) {
				queue = &th->queue[j];
				sm[i].inserted += queue->inserted_rules;
				sm[i].deleted += queue->deleted_rules;
				_total_inserted += sm[i].inserted;
				_total_deleted += sm[i].deleted;
				sm[i].bp += queue->total_bp;
			}
		}

		total_inserted = 0;
		total_deleted = 0;
		clock_gettime(CLOCK_REALTIME, &start);
		usleep(SNAPSHOT_INTERVAL);
		for (i = 0; i < num_of_threads_ct; i++) {
			em[i].inserted = 0;
			em[i].deleted = 0;
			em[i].bp = 0;
			th = &th_info_ct[i];
			for (j = 0; j < QUEUE_PER_THREAD; j++) {
				queue = &th->queue[j];
				em[i].inserted += queue->inserted_rules;
				em[i].deleted += queue->deleted_rules;
				total_inserted += em[i].inserted;
				total_deleted += em[i].deleted;
				em[i].bp += queue->total_bp;
			}

		}

		clock_gettime(CLOCK_REALTIME, &end);
		time_diff = timediff(&start, &end);
		diff = (total_inserted + total_deleted) -
		       (_total_inserted + _total_deleted);

		printf("Manager thread K-Operations/Sec:     %.2lf\n",
		       (double)(diff / K_RULES) / time_diff);

		diff = total_inserted - _total_inserted;
		printf("Manager thread K-Insertions/Sec:     %.2lf\n",
		       (double)(diff / K_RULES) / time_diff);

		diff = total_deleted - _total_deleted;
		printf("Manager thread K-Deletions/Sec:      %.2lf\n",
		       (double)(diff / K_RULES) / time_diff);

		for (i = 0; i < num_of_threads_ct; i++) {
			diff = em[i].inserted - sm[i].inserted;
			if (diff)
				printf("Thread[%u] K-Insertions/Sec:          %-10.2lf HW BP %u\n",
				       i, (double)(diff / K_RULES) / time_diff,
				       em[i].bp - sm[i].bp);

			diff = em[i].deleted - sm[i].deleted;
			if (diff)
				printf("Thread[%u] K-Deletions/Sec:           %-10.2lf HW BP %u\n",
				       i, (double)(diff / K_RULES) / time_diff,
				       em[i].bp - sm[i].bp);
		}
	}

	free(sm);
	free(em);

	return NULL;
}

static void sync_thread(struct thread_info *my_th_info)
{
	bool need_sync = true;
	int i;

	while (need_sync) {
		need_sync = false;
		pthread_mutex_lock(&count_mutex_ct);
		for (i = 0; i < num_of_threads_ct; i++) {
			if (my_th_info->curr_loop > th_info_ct[i].curr_loop)
				need_sync = true;
		}
		if (!need_sync)
			my_th_info->curr_loop++;

		pthread_mutex_unlock(&count_mutex_ct);
	}
}

static void *run_loop_ct(void *arg)
{
	struct thread_info *my_th_info = (struct thread_info *)arg;
	struct mlx5dv_hws_ct_action_data actions_data = {0};
	struct mlx5dv_hws_ct_match match_val = {0};
	struct mlx5dv_hws_ct_rule_attr rule_ct_attr = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	//struct mlx5dv_hws_action *actions[3];
	struct mlx5dv_hws_action *counter;
	struct mlx5dv_hws_action *tag;
	struct mlx5dv_hws_action *dest_ft;
	struct mlx5dv_hws_matcher *matcher;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	struct queue_info *queue;
	int ret, i, l;

	if (cpu_set_ct &&
	    pthread_setaffinity_np(pthread_self(), sizeof(*cpu_set_ct),
				   &cpu_set_ct[my_th_info->thread_id / THREADS_PER_CORE])) {
		TEST_LOG("pthread_setaffinity_np failed");
		return NULL;
	}

	printf("Starting thread %u on core: %u\n", my_th_info->thread_id, sched_getcpu());

	ret = create_ct_ipv4_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto item_destroy;
	}

	my_th_info->curr_loop = 0;
	ctx = my_th_info->ctx;
	matcher = my_th_info->matcher;
	dest_ft = my_th_info->dest_action;
	tag = my_th_info->tag;
	counter = my_th_info->counter;

	rules = calloc(NUM_OF_RULES, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		goto item_destroy;
	}

	for (i = 0; i < QUEUE_PER_THREAD; i++) {
		my_th_info->queue[i].queue_id = (my_th_info->thread_id * QUEUE_PER_THREAD) + i;
		my_th_info->queue[i].inserted_rules = 0;
		my_th_info->queue[i].deleted_rules = 0;
		my_th_info->queue[i].miss_count = 0;
		my_th_info->queue[i].total_bp = 0;
		my_th_info->queue[i].pending = 0;
		my_th_info->queue[i].bp = 0;
	}

	sync_thread(my_th_info);

	actions_data.ctr.action = counter;
	actions_data.ctr.offset = 2;
	actions_data.mh.mode = MLX5DV_HWS_CT_MH_MODE_DATA_NONE;
	actions_data.tag.action = tag;
	actions_data.tag.tag_value = htobe32(4);
	actions_data.dest_table = dest_ft;
	actions_data.dest_table_reverse = dest_ft;

	for (l = 0; l < LOOP; l++) {

		/* Create HWS rules */
		for (i = 0; i < NUM_OF_RULES; i++) {
			rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules +
				(i * mlx5dv_hws_rule_get_handle_size()));
			queue = &my_th_info->queue[i % QUEUE_PER_THREAD];

			rule_attr.burst = !((i + 1) % BURST_TH == 0);
			rule_attr.queue_id = queue->queue_id;
			rule_attr.user_data = rule;
			queue->inserted_rules++;
			match_val.ctv4.dst_addr = i;
			//match_val.ctv4.src_addr = i + 2;
			//match_val.ctv4.dst_port = i;
			//match_val.ctv4.src_port = i + 2;
			//match_val.ctv4.protocol = i;

			rule_ct_attr.attr = &rule_attr;
			rule_ct_attr.bi_direction = false;
			ret = mlx5dv_hws_ct_rule_create(matcher, &match_val, &actions_data,
							&rule_ct_attr, rule);
			if (ret)
				TEST_LOG("Failed to create CT rule %d", i);

			queue->pending++;

			poll_for_comp(ctx, queue->queue_id, &queue->pending,
				      BURST_TH, &queue->miss_count, &queue->bp,
				      false);

			queue->total_bp += queue->bp;
			queue->bp = 0;
		}

		/* Drain the queues */
		for (i = 0; i < QUEUE_PER_THREAD; i++) {
			poll_for_comp(ctx,
				      my_th_info->queue[i].queue_id,
				      &my_th_info->queue[i].pending,
				      BURST_TH,
				      &my_th_info->queue[i].miss_count,
				      &my_th_info->queue[i].bp,
				      true);

			queue->total_bp += my_th_info->queue[i].bp;

			my_th_info->queue[i].bp = 0;
			my_th_info->queue[i].miss_count = 0;
			my_th_info->queue[i].pending = 0;
		}

		/* Delete HWS rules */
		for (i = 0; i < NUM_OF_RULES; i++) {
			rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules +
				(i * mlx5dv_hws_rule_get_handle_size()));
			queue = &my_th_info->queue[i % QUEUE_PER_THREAD];

			rule_attr.burst = !((i + 1) % BURST_TH == 0);
			rule_attr.queue_id = queue->queue_id;
			rule_attr.user_data = rule;
			queue->deleted_rules++;

			match_val.ctv4.dst_addr = i;
			//match_val.ctv4.src_addr = i + 2;
			//match_val.ctv4.dst_port = i;
			//match_val.ctv4.src_port = i + 2;
			//match_val.ctv4.protocol = i;

			rule_ct_attr.attr = &rule_attr;
			rule_ct_attr.bi_direction = false;
			ret = mlx5dv_hws_ct_rule_destroy(rule, &match_val, &rule_ct_attr);
			if (ret)
				TEST_LOG("Failed to destroy rule %d", i);

			queue->pending++;
			poll_for_comp(ctx, queue->queue_id, &queue->pending,
				      BURST_TH, &queue->miss_count, &queue->bp,
				      false);
			queue->total_bp += queue->bp;
			queue->bp = 0;
		}

		/* Drain the queue */
		for (i = 0; i < QUEUE_PER_THREAD; i++) {
			poll_for_comp(ctx,
				      my_th_info->queue[i].queue_id,
				      &my_th_info->queue[i].pending,
				      BURST_TH,
				      &my_th_info->queue[i].miss_count,
				      &my_th_info->queue[i].bp,
				      true);

			queue->total_bp += my_th_info->queue[i].bp;

			my_th_info->queue[i].bp = 0;
			my_th_info->queue[i].miss_count = 0;
			my_th_info->queue[i].pending = 0;
		}
	}

	atomic_fetch_add(&global_counter_ct, 1);
	free(rules);
	//destroy_basic_item(&item);
	return NULL;

item_destroy:
	destroy_basic_item(&item);
	atomic_fetch_add(&global_counter_ct, 1);

	return NULL;
}

int test_multi_threaded_ct(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_ct_matcher_attr matcher_ct_attr = {0};
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_table *tbl, *root_tbl;
	struct mlx5dv_hws_matcher *root_matcher;
	struct mlx5dv_hws_action *dest_action;
	struct mlx5dv_hws_matcher **matchers;
	struct mlx5dv_hws_resource_attr res_attr = {0};
	struct mlx5dv_hws_resource *all_resources[4];
	struct mlx5dv_hws_resource *resource_countr;
	struct mlx5dv_hws_action *counter;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_item root_item = {0};
	struct mlx5dv_hws_action *actions[3];
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_at *at;
	struct mlx5dv_hws_mt *mt, *root_mt;
	cpu_set_t t_cpu_set_ct;
	pthread_t *threads;
	uint64_t cpu_mask;
	pthread_t m_th;
	int ret, i, j;
	struct mlx5dv_hws_action *tag;
	struct mlx5dv_hws_table *tbl3;
	struct mlx5dv_hws_action *dest_ft;
	enum mlx5dv_hws_action_flags flags = MLX5DV_HWS_ACTION_FLAG_NIC_RX;


	cpu_set_ct = NULL;
	num_of_cores_ct = 0;
	num_of_threads_ct = 0;

	if (t_info->cpu_mask) {
		/* Get number of cores */
		num_of_cores_ct = set_cpu_cores(&t_cpu_set_ct, t_info->cpu_mask);
		cpu_set_ct = calloc(num_of_cores_ct, sizeof(*cpu_set_ct));
		if (!cpu_set_ct) {
			TEST_LOG("Failed to allocate cpu_set_ct array");
			errno = ENOMEM;
			goto out_err;
		}

		cpu_mask = t_info->cpu_mask;
		i = 0;
		j = 0;

		/* Set different cpu_set_ct for each core */
		while (cpu_mask) {
			if (cpu_mask & 0x1) {
				set_cpu_cores(&cpu_set_ct[j], 1ULL << i);
				j++;
			}
			i++;
			cpu_mask = cpu_mask >> 1;
		}
	} else {
		cpu_set_ct = NULL;
		num_of_cores_ct = DEFAULT_NUM_OF_CORES;
	}

	num_of_threads_ct = num_of_cores_ct * THREADS_PER_CORE;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto free_cpu_set_ct;
	}

	ctx_attr.queues = num_of_threads_ct * QUEUE_PER_THREAD;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	}

	/* Create Root Table */
	tbl_attr.level = 0;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	root_tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!root_tbl) {
		TEST_LOG("Failed to create root table");
		goto ctx_close;
	}

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto root_tbl_destroy;
	}

	/* Create dest table */
	tbl3 = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl3) {
		TEST_LOG("Failed to create table for action");
		return -1;
	}

	printf("Table created\n");

	dest_ft = mlx5dv_hws_action_create_dest_table(ctx, tbl3, flags);
	if (!dest_ft) {
		TEST_LOG("Failed to create table action");
		return -1;
	}

	ret = create_ct_ipv4_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		return -1;
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		goto item_destroy;
	}

	at = mlx5dv_hws_action_template_create(NULL, 0, &at_attr);
	if (!at) {
		TEST_LOG("Failed to create dummy action template");
		goto mt_destroy;
	}

	/* Create ROOT Matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG;
	matcher_attr.priority = 0;

	ret = create_basic_item(&root_item);
	root_mt = mlx5dv_hws_match_template_create(ctx, &root_item, 1, &mt_attr);
	if (!root_mt) {
		TEST_LOG("Failed to create root match template");
		goto item_destroy;
	}

	root_matcher = mlx5dv_hws_matcher_create(root_tbl, &root_mt, 1, &at, 1, &matcher_attr);
	if (!root_matcher) {
		TEST_LOG("Failed to create root matcher");
		goto at_destroy;
	}

	dest_action = mlx5dv_hws_action_create_dest_default_miss(ctx,
								 MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!dest_action) {
		TEST_LOG("Failed to create default miss action\n");
		goto root_matcher_destroy;
	}

	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_HTABLE;
	matcher_attr.table.sz_row_log = ROW_LOG;
	matcher_attr.table.sz_col_log = 0;

	matchers = calloc(num_of_threads_ct, sizeof(*matchers));
	if (!matchers) {
		TEST_LOG("Failed to allocate matchers");
		goto default_miss_destroy;
	}
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_ct_attr.attr = &matcher_attr;
	matcher_ct_attr.ct_type = MLX5DV_HWS_CT_TYPE_V4;

	/* Create HWS Matchers */
	for (i = 0; i < num_of_threads_ct; i++) {
		matcher_attr.priority = i;
		matchers[i] = mlx5dv_hws_ct_matcher_create(tbl, mt, &matcher_ct_attr);
		if (!matchers[i]) {
			TEST_LOG("Failed to create matcher %u", i);
			goto matchers_destroy;
		}
	}

	tag = mlx5dv_hws_action_create_tag(ctx, MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!tag) {
		TEST_LOG("Failed to create tag action");
		return -1;
	}

	res_attr.bulk_log_size = 10;
	res_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER;
	resource_countr = mlx5dv_hws_resource_alloc(ctx, &res_attr);
	if (!resource_countr) {
		TEST_LOG("Failed to create resource CTR!!");
		return -1;
	}
	all_resources[0] = resource_countr;
	counter = mlx5dv_hws_action_create_counter(ctx, all_resources, 1, 1);
	if (!counter) {
		TEST_LOG("Failed to create resource CTR!!");
		return -1;
	}



	/* Dump once all control data created */
	dump_all(ctx);

	/* Run different threads for different matchers */
	atomic_init(&global_counter_ct, 0);
	ret = pthread_create(&m_th, NULL, &manager_thread, NULL);
	if (ret) {
		TEST_LOG("Failed to create manager thread %u", ret);
		goto matcher_destroy;
	}

	threads = calloc(num_of_threads_ct, sizeof(*threads));
	if (!threads) {
		TEST_LOG("Failed to allocate threads");
		goto matcher_destroy;
	}

	th_info_ct = calloc(num_of_threads_ct, sizeof(*th_info_ct));
	if (!th_info_ct) {
		TEST_LOG("Failed to allocate th_info_ct");
		goto threads_destroy;
	}

	actions[0] = tag;
	actions[1] = counter;

	ret = mlx5dv_hws_ct_actions_bind(matchers[0], actions, 2);
	if (ret) {
		TEST_LOG("Failed to bind actions array");
		goto item_destroy;
	}


	for (i = 0; i < num_of_threads_ct; i++) {

		th_info_ct[i].ctx = ctx;
		th_info_ct[i].thread_id = i;
		th_info_ct[i].matcher = matchers[i];
		th_info_ct[i].curr_loop = -1;
		//th_info_ct[i].dest_action = dest_action;
		th_info_ct[i].dest_action = dest_ft;
		th_info_ct[i].counter = counter;
		th_info_ct[i].tag = tag;
		ret = pthread_create(&threads[i], NULL, &run_loop_ct, &th_info_ct[i]);
		if (ret) {
			TEST_LOG("Failed to create thread #0x%x %u", i, ret);
			goto th_info_destroy;
		}
	}

	for (i = 0; i < num_of_threads_ct; i++)
		pthread_join(threads[i], NULL);

	pthread_join(m_th, NULL);

	for (i = 0; i < num_of_threads_ct; i++)
		mlx5dv_hws_matcher_destroy(matchers[i]);

	free(th_info_ct);
	free(threads);
	free(matchers);
	mlx5dv_hws_action_destroy(dest_action);
	mlx5dv_hws_matcher_destroy(root_matcher);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_match_template_destroy(mt);
	//destroy_basic_item(&item);
	mlx5dv_hws_table_destroy(root_tbl);
	mlx5dv_hws_table_destroy(tbl);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);
	free(cpu_set_ct);

	return 0;

th_info_destroy:
	free(th_info_ct);
threads_destroy:
	free(threads);
matcher_destroy:
	for (i = 0; i < num_of_threads_ct; i++) {
		if (matchers[i])
			mlx5dv_hws_matcher_destroy(matchers[i]);
	}
matchers_destroy:
	free(matchers);
default_miss_destroy:
	mlx5dv_hws_action_destroy(dest_action);
root_matcher_destroy:
	mlx5dv_hws_matcher_destroy(root_matcher);
at_destroy:
	mlx5dv_hws_action_template_destroy(at);
mt_destroy:
	mlx5dv_hws_match_template_destroy(mt);
item_destroy:
	destroy_basic_item(&item);
	mlx5dv_hws_table_destroy(tbl);
root_tbl_destroy:
	mlx5dv_hws_table_destroy(root_tbl);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
free_cpu_set_ct:
	free(cpu_set_ct);
out_err:
	TEST_LOG("Test failed errno=%d", errno);
	return -errno;
}
