/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"

#define NUM_OF_QUEUES 16
#define QUEUE_SIZE 256
#define BURST_TH 1

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			TEST_LOG("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

int create_sampler_item(struct mlx5dv_hws_item *item, enum mlx5dv_hws_field_name fname,
			enum mlx5dv_hws_field_header field_header)
{

	struct mlx5dv_hws_field *fields;
	void *mask, *match;

	fields = calloc(1, sizeof(struct mlx5dv_hws_field));
	if (!fields) {
		TEST_LOG("Failed to allocate fields");
		return -ENOMEM;
	}

	match = calloc(1, DEVX_ST_SZ_BYTES(my_basic_item));
	if (!match) {
		TEST_LOG("Failed to allocate match");
		free(fields);
		return -ENOMEM;
	}

	mask = calloc(1, DEVX_ST_SZ_BYTES(my_basic_item));
	if (!mask) {
		TEST_LOG("Failed to allocate mask");
		free(match);
		free(fields);
		return -ENOMEM;
	}

	item->num_of_fields = 1;
	item->fields = fields;
	item->data.mask = mask;
	item->data.match = match;

	fields[0].fname = fname;
	fields[0].header = field_header;
	fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[0].value.bit_off = CALC_BIT_OFF(my_basic_item, sampler);
	DEVX_SET(my_basic_item, mask, sampler, -1);
	DEVX_SET(my_basic_item, match, sampler, 0x01020304);

	return 0;
}

/**
 * Example of a New IP Protocol (from the PRM) - MY_IP
 *
 * We want to add another header (MY_IP) that will follow the MAC header with
 * eth_type 0x799, and will proceed to UDP/TCP according the next header value.
 * In this example, we will create the parse graph and match on the magic
 * number. The sampler currently supports only 32-bit granularity.
 * Hence, we will match the first 32 bits of MY_IP header.
 *
 * The parse graph layout (example):
 * MAC -> IPv4, IPv6, MY_IP
 * IPv4 -> TCP, UDP
 * IPv6 -> TCP, UDP
 * MY_IP -> TCP, UDP
 *
 * The MY_IP protocol header is structured as follows:
 * First 32 bits:
 * payload_length (6 bits)
 * reserved (10 bits)
 * next_header (8 bits)
 * magic (8 bits)
 * Next 32 bits:
 * source_address (32 bits)
 * Next 32 bits:
 * destination_address (32 bits)
 */
int test_parser(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	enum mlx5dv_hws_field_name parser_fname, parser_ok_fname;
	struct mlx5dv_hws_parser_sampler_cfg sampler_cfg = {0};
	struct mlx5dv_hws_parser_arc_cfg arc_tcp_cfg = {0};
	struct mlx5dv_hws_parser_arc_cfg arc_udp_cfg = {0};
	struct mlx5dv_hws_parser_arc_cfg arc_mac_cfg = {0};
	struct mlx5dv_hws_parser_node_cfg my_ip_cfg = {0};
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_parser_sampler *sampler = NULL;
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_parser_graph *graph = NULL;
	struct mlx5dv_hws_parser_arc *arc_tcp = NULL;
	struct mlx5dv_hws_parser_arc *arc_udp = NULL;
	struct mlx5dv_hws_parser_arc *arc_mac = NULL;
	struct mlx5dv_hws_parser_node *my_ip = NULL;
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	enum mlx5dv_hws_field_header field_header;
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_action *dest_action;
	struct mlx5dv_hws_context *ctx = NULL;
	struct mlx5dv_hws_matcher *matcher;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_rule *rule;
	struct mlx5dv_hws_table *tbl;
	uint32_t modify_header_field;
	uint32_t pending_rules = 0;
	struct mlx5dv_hws_at *at;
	struct mlx5dv_hws_mt *mt;
	uint32_t miss_count = 0;
	uint32_t hw_bp = 0;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	}

	graph = mlx5dv_hws_parser_graph_create(ctx);
	if (!graph) {
		TEST_LOG("Failed to create graph");
		goto ctx_close;
	}

       /* Payload length configuration - how to calculate header length */
	my_ip_cfg.hdr_len.mode = MLX5DV_HWS_PARSER_HDR_LEN_MODE_FIELD;
	my_ip_cfg.hdr_len.field.field.bit_offset = 0;
	my_ip_cfg.hdr_len.field.field.bit_length = 6; /* Max bit length */
	my_ip_cfg.hdr_len.field.multiplier = 1;

	/* Next Header - where is the next header pointer located */
	my_ip_cfg.selection_field.bit_offset = 16;
	my_ip_cfg.selection_field.bit_length = 8;

	my_ip_cfg.has_selection_field = true;
	my_ip = mlx5dv_hws_parser_node_create(graph, &my_ip_cfg);
	if (!my_ip) {
		TEST_LOG("Failed to create parser node");
		goto parser_graph_destroy;
	}

	/* MAC -> MY_IP */
	arc_mac_cfg.src_node.type = MLX5DV_HWS_PARSER_NODE_TYPE_NATIVE;
	arc_mac_cfg.src_node.native.node = MLX5DV_HWS_PARSER_NATIVE_NODE_MAC;
	arc_mac_cfg.selection_field_data = 0x799;

	arc_mac_cfg.dst_node.type = MLX5DV_HWS_PARSER_NODE_TYPE_FLEX;
	arc_mac_cfg.dst_node.flex.node = my_ip;
	arc_mac = mlx5dv_hws_parser_arc_create(graph, &arc_mac_cfg);
	if (!arc_mac) {
		TEST_LOG("Failed to create arc_mac");
		goto parser_node_destroy;
	}

	/* MY_IP -> TCP */
	arc_tcp_cfg.src_node.type = MLX5DV_HWS_PARSER_NODE_TYPE_FLEX;
	arc_tcp_cfg.src_node.flex.node = my_ip;
	arc_tcp_cfg.dst_node.type = MLX5DV_HWS_PARSER_NODE_TYPE_NATIVE;
	arc_tcp_cfg.dst_node.native.node = MLX5DV_HWS_PARSER_NATIVE_NODE_TCP;
	arc_tcp_cfg.selection_field_data = 0x06; /* TCP */

	arc_tcp = mlx5dv_hws_parser_arc_create(graph, &arc_tcp_cfg);
	if (!arc_tcp) {
		TEST_LOG("Failed to create arc_tcp");
		goto parser_arc_mac_destroy;
	}

	/* MY_IP -> UDP */
	arc_udp_cfg.src_node.type = MLX5DV_HWS_PARSER_NODE_TYPE_FLEX;
	arc_udp_cfg.src_node.flex.node = my_ip;
	arc_udp_cfg.dst_node.type = MLX5DV_HWS_PARSER_NODE_TYPE_NATIVE;
	arc_udp_cfg.dst_node.native.node = MLX5DV_HWS_PARSER_NATIVE_NODE_UDP;
	arc_udp_cfg.selection_field_data = 0x11; /* UDP */

	arc_udp = mlx5dv_hws_parser_arc_create(graph, &arc_udp_cfg);
	if (!arc_udp) {
		TEST_LOG("Failed to create arc_udp");
		goto parser_arc_tcp_destroy;
	}

	sampler_cfg.share_hw_resources = 0; /* Not supported */
	sampler_cfg.field_cfg.bit_length = 32; /* Support only 32 */
	sampler_cfg.field_cfg.bit_offset = 0; /* First DW */
	sampler_cfg.node_encap_level = 0; /* OUTER */

	sampler = mlx5dv_hws_parser_sampler_create(my_ip, &sampler_cfg);
	if (!sampler) {
		TEST_LOG("Failed to create sampler");
		goto parser_arc_udp_destroy;
	}

	/* Now the graph applied */
	if (mlx5dv_hws_parser_graph_bind(graph)) {
		TEST_LOG("Failed to create bind graph");
		goto parser_sampler_destroy;
	} else {
		sleep(1);
	}

	/* HWS attributes are known only after graph bind and queried after */
	mlx5dv_hws_parser_sampler_query(sampler, &parser_fname, &parser_ok_fname,
					&field_header, &modify_header_field);
	if (create_sampler_item(&item, parser_fname, field_header)) {
		TEST_LOG("Failed to create items");
		goto parser_graph_unbind;
	}

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto parser_destroy_item;
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		goto parser_destroy_tbl;
	}

	at = mlx5dv_hws_action_template_create(NULL, 0, &at_attr);
	if (!at) {
		TEST_LOG("Failed to create empty action template");
		goto parser_destroy_mt;
	}

	/* Create HWS Matcher */
	matcher_attr.optimize_using_rule_idx = 1;
	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &matcher_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher");
		goto parser_destroy_at;
	}

	dest_action = mlx5dv_hws_action_create_dest_default_miss(ctx,
								 MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!dest_action) {
		TEST_LOG("Failed to create default miss action\n");
		goto parser_matcher_destroy;
	}

	rule = calloc(1, mlx5dv_hws_rule_get_handle_size());
	if (!rule) {
		TEST_LOG("Failed to create rule");
		goto parser_default_miss_destroy;
	}

	rule_attr.user_data = rule;

	if (mlx5dv_hws_rule_create(matcher, 0, &item.data, 0, NULL, dest_action,
				   &rule_attr, rule)) {
		TEST_LOG("Failed to create rule");
		goto parser_free_rule;
	}

	pending_rules++;
	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);

	if (mlx5dv_hws_rule_destroy(rule, &rule_attr))
		TEST_LOG("Failed to destroy rule");

	pending_rules++;
	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);

	free(rule);
	mlx5dv_hws_action_destroy(dest_action);
	mlx5dv_hws_matcher_destroy(matcher);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_match_template_destroy(mt);
	mlx5dv_hws_table_destroy(tbl);
	destroy_basic_item(&item);
	mlx5dv_hws_parser_graph_unbind(graph);
	mlx5dv_hws_parser_sampler_destroy(sampler);
	mlx5dv_hws_parser_arc_destroy(arc_mac);
	mlx5dv_hws_parser_arc_destroy(arc_udp);
	mlx5dv_hws_parser_arc_destroy(arc_tcp);
	mlx5dv_hws_parser_node_destroy(my_ip);
	mlx5dv_hws_parser_graph_destroy(graph);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return 0;

parser_free_rule:
	free(rule);
parser_default_miss_destroy:
	mlx5dv_hws_action_destroy(dest_action);
parser_matcher_destroy:
	mlx5dv_hws_matcher_destroy(matcher);
parser_destroy_at:
	mlx5dv_hws_action_template_destroy(at);
parser_destroy_mt:
	mlx5dv_hws_match_template_destroy(mt);
parser_destroy_tbl:
	mlx5dv_hws_table_destroy(tbl);
parser_destroy_item:
	destroy_basic_item(&item);
parser_graph_unbind:
	mlx5dv_hws_parser_graph_unbind(graph);
parser_sampler_destroy:
	mlx5dv_hws_parser_sampler_destroy(sampler);
parser_arc_udp_destroy:
	mlx5dv_hws_parser_arc_destroy(arc_udp);
parser_arc_tcp_destroy:
	mlx5dv_hws_parser_arc_destroy(arc_tcp);
parser_arc_mac_destroy:
	mlx5dv_hws_parser_arc_destroy(arc_mac);
parser_node_destroy:
	mlx5dv_hws_parser_node_destroy(my_ip);
parser_graph_destroy:
	mlx5dv_hws_parser_graph_destroy(graph);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	return -1;
}
