/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"
#include "hws_internal.h"

#define NUM_OF_QUEUES 16
#define QUEUE_SIZE 256

int test_pool(struct ibv_context *ibv_ctx)
{
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct pool_attr pool_attr = {0};
	struct mlx5dv_hws_context *ctx;
	struct pool_chunk stc_0 = {0};
	struct pool_chunk stc_1 = {0};
	struct pool_chunk stc_2 = {0};
	struct pool *pool;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	} else {
		printf("Context created\n");
	}

	pool_attr.flags = POOL_FLAGS_FOR_STC_POOL;
	pool_attr.pool_type = POOL_TYPE_STC;
	pool_attr.alloc_log_sz = max(POOL_STC_LOG_SZ, ctx->caps->stc_alloc_log_gran);
	pool = pool_create(ctx, &pool_attr);
	if (!pool) {
		TEST_LOG("Failed to create pool");
		goto ctx_close;
	} else {
		printf("Pool created\n");
	}

	if (pool_chunk_alloc(pool, &stc_0)) {
		TEST_LOG("Failed to allocate single STC");
		goto destroy_pool;
	} else {
		printf("Got chunk at index: 0x%x\n", stc_0.offset);
	}

	if (pool_chunk_alloc(pool, &stc_1)) {
		TEST_LOG("Failed to allocate single STC");
		goto free_stc_0;
	} else {
		printf("Got chunk at index: 0x%x\n", stc_1.offset);
	}

	if (pool_chunk_alloc(pool, &stc_2)) {
		TEST_LOG("Failed to allocate single STC");
		goto free_stc_1;
	} else {
		printf("Got chunk at index: 0x%x\n", stc_2.offset);
	}

	if (stc_0.offset != 0 || stc_1.offset != 1 ||
	    stc_2.offset != 2) {
		TEST_LOG("Incorrect order");
		goto free_stc_2;
	}

	pool_chunk_free(pool, &stc_0);
	printf("Put chunk at index: 0x%x\n", stc_0.offset);
	pool_chunk_free(pool, &stc_1);
	printf("Put chunk at index: 0x%x\n", stc_1.offset);

	if (pool_chunk_alloc(pool, &stc_1)) {
		TEST_LOG("Failed to allocate single STC");
		pool_chunk_free(pool, &stc_2);
		goto destroy_pool;
	} else {
		printf("Got chunk at index: 0x%x\n", stc_1.offset);
	}

	if (stc_1.offset != 0) {
		TEST_LOG("Incorrect order");
		pool_chunk_free(pool, &stc_2);
		goto destroy_pool;
	}

	pool_chunk_free(pool, &stc_1);
	pool_chunk_free(pool, &stc_2);
	pool_destroy(pool);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return 0;

free_stc_2:
	pool_chunk_free(pool, &stc_2);
free_stc_1:
	pool_chunk_free(pool, &stc_1);
free_stc_0:
	pool_chunk_free(pool, &stc_0);
destroy_pool:
	pool_destroy(pool);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	return -errno;
}
