/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"
#include "../mlx5dv_hws.h"

#define NUM_OF_QUEUES 16
#define ROW_LOG 2
#define NUM_OF_RULES (1 << ROW_LOG)
#define QUEUE_SIZE 256
#define BURST_TH 1

#define RANGE_ITEM_NUM_OF_FIELDS 2

static int check_range_support(struct ibv_context *ibv_ctx, bool *has_supp)
{
	uint32_t out[DEVX_ST_SZ_DW(query_hca_cap_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(query_hca_cap_in)] = {0};
	uint16_t supp_ste_format_gen_wqe;
	int ret;

	DEVX_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
	DEVX_SET(query_hca_cap_in, in, op_mod,
			MLX5_GET_HCA_CAP_OP_MOD_WQE_BASED_FLOW_TABLE |
			MLX5_HCA_CAP_OPMOD_GET_CUR);

	ret = mlx5dv_devx_general_cmd(ibv_ctx, in, sizeof(in), out, sizeof(out));
	if (ret)
		return ret;

	supp_ste_format_gen_wqe =
		DEVX_GET(query_hca_cap_out, out,
			 capability.wqe_based_flow_table_cap.ste_format_gen_wqe);

	*has_supp = supp_ste_format_gen_wqe & (1 << MLX5_IFC_RTC_STE_FORMAT_RANGE);

	return 0;
}

static int create_range_item(struct mlx5dv_hws_item *item)
{
	struct mlx5dv_hws_field *fields;
	void *mask, *match, *last, *buff;

	fields = calloc(RANGE_ITEM_NUM_OF_FIELDS, sizeof(struct mlx5dv_hws_field));
	if (!fields) {
		TEST_LOG("Failed to allocate fields");
		return -ENOMEM;
	}

	buff = calloc(3, DEVX_ST_SZ_BYTES(my_basic_item));
	if (!buff) {
		TEST_LOG("Failed to allocate mask and match");
		free(fields);
		return -ENOMEM;
	}

	match = buff;
	mask = buff + DEVX_ST_SZ_BYTES(my_basic_item);
	last = buff + 2 * DEVX_ST_SZ_BYTES(my_basic_item);

	item->num_of_fields = RANGE_ITEM_NUM_OF_FIELDS;
	item->fields = fields;
	item->data.mask = mask;
	item->data.match = match;
	item->data.last = last;

	fields[0].fname = MLX5DV_HWS_FNAME_IPV4_SRC;
	fields[0].header = MLX5DV_HWS_HEADER_OUTER;
	fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[0].value.bit_off = CALC_BIT_OFF(my_basic_item, sip);
	DEVX_SET(my_basic_item, mask, sip, -1);
	DEVX_SET(my_basic_item, match, sip, 0x01010101);
	DEVX_SET(my_basic_item, last, sip, 0x02020202);

	fields[1].fname = MLX5DV_HWS_FNAME_IPV4_DST;
	fields[1].header = MLX5DV_HWS_HEADER_OUTER;
	fields[1].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[1].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[1].value.bit_off = CALC_BIT_OFF(my_basic_item, dip);
	DEVX_SET(my_basic_item, mask, dip, -1);
	DEVX_SET(my_basic_item, match, dip, 0x03030303);
	DEVX_SET(my_basic_item, last, dip, 0x04040404);

	return 0;
}

static void destroy_range_item(struct mlx5dv_hws_item *item)
{
	free((void *)item->data.match);
	free(item->fields);
}

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			printf("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

int test_range(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_action *dest_action;
	struct mlx5dv_hws_matcher *matcher;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	struct mlx5dv_hws_table *tbl;
	uint32_t pending_rules = 0;
	struct mlx5dv_hws_at *at;
	struct mlx5dv_hws_mt *mt;
	uint32_t miss_count = 0;
	uint16_t queue_id = 0;
	uint32_t hw_bp = 0;
	bool fw_supp;
	int ret, i;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	} else {
		printf("Context created\n");
	}

	ret = check_range_support(ibv_ctx, &fw_supp);
	if (ret) {
		TEST_LOG("Fail to check cap");
		errno = EINVAL;
		goto ctx_close;
	}

	if (!fw_supp) {
		TEST_LOG("Range matching not supported");
		mlx5dv_hws_context_close(ctx);
		ibv_dealloc_pd(ctx_attr.pd);
		return 0;
	}

	/* Create HWS Table */
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	tbl_attr.level = 1;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto ctx_close;
	} else {
		printf("Table created\n");
	}

	ret = create_range_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto tbl_destroy;
	} else {
		printf("Item created\n");
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create range match template");
		goto item_destroy;
	} else {
		printf("Match template created\n");
	}

	at = mlx5dv_hws_action_template_create(NULL, 0, &at_attr);
	if (!at) {
		TEST_LOG("Failed to create dummy action template");
		goto mt_destroy;
	} else {
		printf("Dummy action template created\n");
	}

	/* Create ROOT Matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &matcher_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher");
		goto at_destroy;
	} else {
		printf("Matcher created\n");
	}

	dest_action = mlx5dv_hws_action_create_dest_default_miss(ctx,
								 MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!dest_action) {
		TEST_LOG("Failed to create default miss action\n");
		goto matcher_destroy;
	}

	/* Dump once all control data created */
	dump_all(ctx);

	rules = calloc(NUM_OF_RULES, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		goto default_miss_destroy;
	}

	rule_attr.queue_id = queue_id;

	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules + i * mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		DEVX_SET(my_basic_item, (void *)item.data.match, sip, i);
		ret = mlx5dv_hws_rule_create(matcher, 0, &item.data, 0, NULL, dest_action,
					     &rule_attr, rule);
		if (ret)
			TEST_LOG("Failed to create rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Insertion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules + i * mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
		if (ret)
			TEST_LOG("Failed to destroy rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Deletion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	free(rules);
	mlx5dv_hws_action_destroy(dest_action);
	mlx5dv_hws_matcher_destroy(matcher);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_match_template_destroy(mt);
	destroy_range_item(&item);
	mlx5dv_hws_table_destroy(tbl);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return 0;

default_miss_destroy:
	mlx5dv_hws_action_destroy(dest_action);
matcher_destroy:
	mlx5dv_hws_matcher_destroy(matcher);
at_destroy:
	mlx5dv_hws_action_template_destroy(at);
mt_destroy:
	mlx5dv_hws_match_template_destroy(mt);
item_destroy:
	destroy_range_item(&item);
tbl_destroy:
	mlx5dv_hws_table_destroy(tbl);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	TEST_LOG("Test failed errno=%d", errno);
	return -errno;
}
