/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test.h"

#define NUM_OF_QUEUES 16
#define ROW_LOG 20
#define NUM_OF_RULES 1 << ROW_LOG
#define QUEUE_SIZE 256
#define BURST_TH 32

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			TEST_LOG("Failed during poll queue");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

int test_resize(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_matcher *src_matcher, *dst_matcher;
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_table *tbl, *root_tbl;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_action *drop;
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	uint32_t pending_rules = 0;
	struct mlx5dv_hws_at *at;
	struct mlx5dv_hws_mt *mt;
	uint32_t miss_count = 0;
	uint16_t queue_id = 0;
	uint32_t hw_bp = 0;
	int ret, i;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	} else {
		TEST_LOG("Context created");
	}

	/* Create Root Table */
	tbl_attr.level = 0;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	root_tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!root_tbl) {
		TEST_LOG("Failed to create root table");
		goto ctx_close;
	} else {
		TEST_LOG("Root Table created");
	}

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto root_tbl_destroy;
	} else {
		TEST_LOG("Table created");
	}

	ret = create_basic_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto tbl_destroy;
	} else {
		TEST_LOG("Item created");
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		goto item_destroy;
	} else {
		TEST_LOG("Match template created");
	}

	at = mlx5dv_hws_action_template_create(NULL, 0, &at_attr);
	if (!at) {
		TEST_LOG("Failed to create dummy action template");
		goto mt_destroy;
	} else {
		TEST_LOG("Dummy action template created");
	}

	drop = mlx5dv_hws_action_create_dest_drop(ctx, MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!drop) {
		TEST_LOG("Failed to create drop action");
		goto at_destroy;
	}


	/* Create HWS SRC Matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;
	matcher_attr.resizable = true;
	src_matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &matcher_attr);
	if (!src_matcher) {
		TEST_LOG("Failed to create SRC matcher");
		goto drop_action_destroy;
	} else {
		TEST_LOG("SRC Matcher created");
	}

	/* Create HWS DST Matcher */
	matcher_attr.rule.num_log = ROW_LOG + 1;
	matcher_attr.priority = 1;
	dst_matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &matcher_attr);
	if (!dst_matcher) {
		TEST_LOG("Failed to create DST matcher");
		goto src_matcher_destroy;
	} else {
		TEST_LOG("DST Matcher created");
	}

	/* Dump once all control data created */
	dump_all(ctx);

	rules = calloc(NUM_OF_RULES, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		goto dst_matcher_destroy;
	}

	rule_attr.queue_id = queue_id;

	/* Create NUM_OF_RULES */
	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules + i * mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		DEVX_SET(my_basic_item, (void*)item.data.match, sip, i);
		ret = mlx5dv_hws_rule_create(src_matcher, 0, &item.data, 0, NULL, drop, &rule_attr, rule);
		if (ret) {
			TEST_LOG("Failed to create rule %d", i);
		}

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	TEST_LOG("Created %d rules, on source matcher", NUM_OF_RULES);
	TEST_LOG("Total misses: %u (out of: %u) HW bp %u",
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	/* Move NUM_OF_RULES to dst_matcher */
	ret = mlx5dv_hws_matcher_resize_set_target(src_matcher, dst_matcher);
	if (ret) {
		TEST_LOG("Failed to set matcher resize ret=%d", ret);
		goto dst_matcher_destroy;
	} else {
		TEST_LOG("Call resize set target");
	}

	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules + i * mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		ret = mlx5dv_hws_matcher_resize_rule_move(src_matcher, rule, &rule_attr);
		if (ret) {
			TEST_LOG("Failed to move rule %d", i);
		}

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	TEST_LOG("Moved %d rules, to destination matcher", NUM_OF_RULES);
	TEST_LOG("Total misses: %u (out of: %u) HW bp %u",
		miss_count,
		NUM_OF_RULES,
		hw_bp);
	
	/* Delete NUM_OF_RULES from dst_matcher */
	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules + i * mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
		if (ret) {
			TEST_LOG("Failed to destroy rule %d", i);
		}

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	TEST_LOG("Deleted %d rules, from destination matcher", NUM_OF_RULES);
	TEST_LOG("Total misses: %u (out of: %u) HW bp %u",
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	free(rules);
	mlx5dv_hws_matcher_destroy(dst_matcher);
	mlx5dv_hws_matcher_destroy(src_matcher);
	mlx5dv_hws_action_destroy(drop);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_match_template_destroy(mt);
	destroy_basic_item(&item);
	mlx5dv_hws_table_destroy(root_tbl);
	mlx5dv_hws_table_destroy(tbl);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return 0;

dst_matcher_destroy:
	mlx5dv_hws_matcher_destroy(dst_matcher);
src_matcher_destroy:
	mlx5dv_hws_matcher_destroy(src_matcher);
drop_action_destroy:
	mlx5dv_hws_action_destroy(drop);
at_destroy:
	mlx5dv_hws_action_template_destroy(at);
mt_destroy:
	mlx5dv_hws_match_template_destroy(mt);
item_destroy:
	destroy_basic_item(&item);
tbl_destroy:
	mlx5dv_hws_table_destroy(tbl);
root_tbl_destroy:
	mlx5dv_hws_table_destroy(root_tbl);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	TEST_LOG("Test failed errno=%d", errno);
	return -errno;
}
