/* Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test_resource.h"

struct test_resource_thread_args {
	struct mlx5dv_hws_resource *resource;
	int thread_id;
	uint8_t *offset_taken;
	int *error;
};

static struct mlx5dv_hws_devx_obj *
buf_mkey_obj_create(struct ibv_context *ibv_ctx, uint32_t pd_id, uint32_t umem_id,
		    uint8_t *buf, size_t buf_size)
{
	uint32_t out[DEVX_ST_SZ_DW(create_mkey_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_mkey_in)] = {0};
	void *mkc = DEVX_ADDR_OF(create_mkey_in, in, memory_key_mkey_entry);
	size_t pgsize = sysconf(_SC_PAGESIZE);
	uint32_t log_page_size = __builtin_ffs(pgsize) - 1;
	struct mlx5dv_hws_devx_obj *mkey;

	mkey = malloc(sizeof(*mkey));
	if (!mkey) {
		TEST_LOG("Failed to allocate mkey object data");
		errno = ENOMEM;
		return NULL;
	}

	DEVX_SET(create_mkey_in, in, opcode, MLX5_CMD_OP_CREATE_MKEY);
	DEVX_SET(create_mkey_in, in, mkey_umem_id, umem_id);
	DEVX_SET(mkc, mkc, access_mode_1_0, 0x1);
	DEVX_SET(mkc, mkc, lw, 0x1);
	DEVX_SET(mkc, mkc, lr, 0x1);
	DEVX_SET(mkc, mkc, qpn, 0xffffff);
	DEVX_SET(mkc, mkc, pd, pd_id);
	DEVX_SET(mkc, mkc, mkey_7_0, umem_id & 0xFF);
	DEVX_SET(mkc, mkc, log_page_size, log_page_size);
	DEVX_SET64(mkc, mkc, start_addr, (uint64_t)buf);
	DEVX_SET64(mkc, mkc, len, buf_size);

	mkey->obj = mlx5dv_devx_obj_create(ibv_ctx, in, sizeof(in), out, sizeof(out));
	if (!mkey->obj) {
		TEST_LOG("%s failed %d", __func__, errno);
		free(mkey);
		return NULL;
	}

	mkey->id = (DEVX_GET(create_mkey_out, out, mkey_index) << 8) |
		   (umem_id & 0xFF);
	return mkey;
}

static void *buf_aligned_alloc(uint64_t size, uint64_t align)
{
	uint64_t al_size = (size + align - 1) & ~(align - 1);
	void *buf;
	int err;

	assert(al_size % sizeof(void *) == 0);
	err = posix_memalign(&buf, align, al_size);
	if (err)
		return NULL;

	memset(buf, 0, al_size);

	return buf;
}

int buf_create(struct mlx5dv_hws_context *ctx, struct buf_sp *sp,
	       uint8_t opmod)
{
	struct ibv_context *ibv_ctx = context_get_local_ibv(ctx);
	size_t size = BUF_SIZE_BYTES;
	struct mlx5dv_devx_umem *umem;
	uint8_t *buf;

	buf = buf_aligned_alloc(size, CACHE_LINE_SIZE);
	if (!buf) {
		TEST_LOG("%s alloc failed", __func__);
		return -1;
	}

	umem = mlx5dv_devx_umem_reg(ibv_ctx, buf, size, IBV_ACCESS_LOCAL_WRITE);
	if (!umem) {
		TEST_LOG("%s umem reg failed", __func__);
		goto err_free_buf;
	}

	sp->mkey = buf_mkey_obj_create(ctx->ibv_ctx,
				       ctx->pd_num,
				       umem->umem_id,
				       buf, size);
	if (!sp->mkey) {
		TEST_LOG("%s mkey create failed", __func__);
		goto err_dereg_umem;
	}

	sp->umem = umem;
	sp->buf = buf;

	return 0;

err_dereg_umem:
	mlx5dv_devx_umem_dereg(umem);
err_free_buf:
	free(buf);
	return -1;
}

void buf_destroy(struct buf_sp *sp)
{
	mlx5dv_devx_obj_destroy(sp->mkey->obj);
	free(sp->mkey);
	mlx5dv_devx_umem_dereg(sp->umem);
	free(sp->buf);
}

void aso_fill_condition_cseg(struct aso_cseg_input *aso_ctrl,
			     struct send_wqe_aso_ctrl_seg *aso_cseg)
{
	aso_cseg->operand_masks = htobe32(aso_ctrl->operand_masks);
	aso_cseg->condition_0_data = htobe32(aso_ctrl->condition_0_data);
	aso_cseg->condition_0_mask = htobe32(aso_ctrl->condition_0_mask);
	aso_cseg->condition_1_data = htobe32(aso_ctrl->condition_1_data);
	aso_cseg->condition_1_mask = htobe32(aso_ctrl->condition_1_mask);
	aso_cseg->bitwise_data = htobe64(aso_ctrl->bitwise_data);
	aso_cseg->data_mask = htobe64(aso_ctrl->data_mask);
}

void aso_fill_send_cseg(uint64_t mr_adrr, uint32_t lkey,
			struct send_wqe_aso_ctrl_seg *aso_cseg,
			bool read)
{
	if (read) {
		aso_cseg->va_r = htobe64(mr_adrr | 1); /* Must be aligned to cache size (6b) */
		aso_cseg->lkey = htobe32(lkey);
	} else {
		aso_cseg->va_r = 0;
		aso_cseg->lkey = 0;
	}
}

void *test_resource_thread(void *arg)
{
	struct test_resource_thread_args *args = (struct test_resource_thread_args *)arg;
	struct mlx5dv_hws_resource *resource = args->resource;
	uint32_t offset;
	int i;

	for (i = 0; i < (1 << BUF_LOG_SIZE) / NUM_OF_QUEUES; i++) {
		mlx5dv_hws_resource_get_offset(resource, args->thread_id,
					       &offset);

		if (offset < 0 || args->offset_taken[offset])
			*args->error = 1;

		args->offset_taken[offset] = 1;
		sched_yield();
		args->offset_taken[offset] = 0;
		if (mlx5dv_hws_resource_put_offset(resource, args->thread_id,
						   offset))
			*args->error = 1;
	}

	return NULL;
}

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t pending_req)
{
	struct mlx5dv_hws_queue_op_result res[QUEUE_SIZE];
	int ret = 0;
	int j;

	while (pending_req) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, res, pending_req);
		if (ret < 0) {
			TEST_LOG("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			pending_req -= ret;
			for (j = 0; j < ret; j++) {
				if (res[j].status == MLX5DV_HWS_QUEUE_OP_ERROR) {
					TEST_LOG("Status: MLX5DV_HWS_QUEUE_OP_ERROR\n");
					return -1;
				}
			}
		}
	}

	return 0;
}

static int test_resource_enqueue_query_config(struct ibv_pd *pd,
					      struct mlx5dv_hws_context *ctx)
{
	struct mlx5dv_hws_resource_enqueue_aso_attr enqueue_attr = {0};
	struct mlx5dv_hws_resource_attr resource_attr = {0};
	struct aso_cseg_input aso_cseg_in = {0};
	struct mlx5dv_hws_resource *resource;
	struct aso_wqe aso_wqe = {0};
	struct buf_sp sp = {0};
	size_t wqe_len;
	int i;

	buf_create(ctx, &sp, MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK);

	resource_attr.bulk_log_size = BUF_LOG_SIZE;
	resource_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK;
	resource = mlx5dv_hws_resource_alloc(ctx, &resource_attr);
	if (!resource) {
		TEST_LOG("Failed to allocate resource!");
		goto destroy_ctr_buf;
	}

	mlx5dv_hws_resource_get_offset(resource, 0, &enqueue_attr.offset);
	enqueue_attr.burst = false;
	enqueue_attr.user_data = &aso_wqe;

	/* Read_enable = 0 (Write Only) */
	aso_cseg_in.operand_masks =
			(ASO_BYTEWISE_64BYTE << ASO_CSEG_DATA_MASK_MODE_OFFSET) |
			(ASO_OPER_LOGICAL_OR << ASO_CSEG_COND_OPER_OFFSET) |
			(ASO_OP_ALWAYS_TRUE << ASO_CSEG_COND_1_OPER_OFFSET) |
			(ASO_OP_ALWAYS_TRUE << ASO_CSEG_COND_0_OPER_OFFSET);
	aso_cseg_in.data_mask = UINT64_MAX;
	for (i = 0; i < sizeof(aso_wqe.bytewise_data); i++)
		aso_wqe.bytewise_data[i] = i;
	aso_fill_send_cseg(0, 0, &aso_wqe.aso_ctrl, false);
	aso_fill_condition_cseg(&aso_cseg_in, &aso_wqe.aso_ctrl);
	wqe_len = sizeof(struct aso_wqe); /* ASO Control + ASO Data */
	mlx5dv_hws_resource_enqueue_aso_wqe(resource, 0, (__be32 *)&aso_wqe, wqe_len,
					    &enqueue_attr);
	if (poll_for_comp(ctx, 0, 1))
		goto out_err;

	/* Read_enable = 1 (Read Only) */
	aso_fill_send_cseg((uint64_t)sp.buf, sp.mkey->id, &aso_wqe.aso_ctrl, true);
	aso_cseg_in.operand_masks = 0;
	aso_cseg_in.data_mask = 0;
	wqe_len = sizeof(struct send_wqe_aso_ctrl_seg); /* ASO Control */
	mlx5dv_hws_resource_enqueue_aso_wqe(resource, 0, (__be32 *)&aso_wqe, wqe_len,
					    &enqueue_attr);
	if (poll_for_comp(ctx, 0, 1))
		goto out_err;

	for (i = 0; i < sizeof(aso_wqe.bytewise_data); i++)
		if (sp.buf[i] != aso_wqe.bytewise_data[i]) {
			TEST_LOG("sp.buf[%d] 0x%x != aso_wqe.bytewise_data[%d] 0x%x\n",
				i, sp.buf[i], i, aso_wqe.bytewise_data[i]);
			goto out_err;
		}

	mlx5dv_hws_resource_put_offset(resource, 0, enqueue_attr.offset);
	mlx5dv_hws_resource_free(resource);
	buf_destroy(&sp);

	return 0;
out_err:
	mlx5dv_hws_resource_free(resource);
destroy_ctr_buf:
	buf_destroy(&sp);
	return -1;
}

static int test_resource_create_resource(struct ibv_pd *pd, struct mlx5dv_hws_context *ctx)
{
	uint8_t resource_type[] = {
		MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER,
		MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK,
		MLX5DV_HWS_RESOURCE_TYPE_ASO_FLOW_HIT,
		MLX5DV_HWS_RESOURCE_TYPE_ASO_METER,
		MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B,
		MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_128B,
		MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_256B,
	};
	struct mlx5dv_hws_resource_attr resource_attr = {0};
	struct mlx5dv_hws_resource *resource;
	struct mlx5dv_hws_devx_obj devx_obj;
	int i;

	for (i = 0; i < sizeof(resource_type); i++) {
		resource_attr.bulk_log_size =
			resource_type[i] == MLX5DV_HWS_RESOURCE_TYPE_ASO_FLOW_HIT ?
					    0 : BUF_LOG_SIZE;
		resource_attr.type = resource_type[i];
		resource = mlx5dv_hws_resource_alloc(ctx, &resource_attr);
		if (!resource)
			return -1;
		if (mlx5dv_hws_resource_get_devx_obj(resource, &devx_obj.obj, &devx_obj.id)) {
			mlx5dv_hws_resource_free(resource);
			return -1;
		}

		mlx5dv_hws_resource_free(resource);
	}

	return 0;
}

static int test_resource_sequential_get_put(struct ibv_pd *pd, struct mlx5dv_hws_context *ctx)
{
	int offset_in_qid[NUM_OF_QUEUES][BUF_SIZE / NUM_OF_QUEUES] = {{0}, {0}};
	struct mlx5dv_hws_resource_attr resource_attr = {0};
	struct mlx5dv_hws_resource *resource;
	bool offset_taken[BUF_SIZE] = {0};
	uint32_t offset;
	int i, ret, qid;

	resource_attr.bulk_log_size = BUF_LOG_SIZE;
	resource_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK;
	resource = mlx5dv_hws_resource_alloc(ctx, &resource_attr);
	if (!resource) {
		printf("Failed to allocate resource\n");
		return -1;
	}

	for (qid = 0; qid < NUM_OF_QUEUES; qid++) {
		for (i = 0; i < BUF_SIZE / NUM_OF_QUEUES; i++) {
			ret = mlx5dv_hws_resource_get_offset(resource, qid, &offset);
			if (ret)
				goto out_err;

			if (offset_taken[offset])
				goto out_err;

			offset_in_qid[qid][i] = offset;
			offset_taken[offset] = 1;
		}
	}

	for (qid = 0; qid < NUM_OF_QUEUES; qid++) {
		for (i = 0; i < BUF_SIZE / NUM_OF_QUEUES; i++) {
			ret = mlx5dv_hws_resource_put_offset(resource,
							     qid,
							     offset_in_qid[qid][i]);
			if (ret)
				goto out_err;
		}
	}

	mlx5dv_hws_resource_free(resource);
	return 0;

out_err:
	mlx5dv_hws_resource_free(resource);
	return -1;
}

static int test_resource_multithread_get_put(struct ibv_pd *pd, struct mlx5dv_hws_context *ctx)
{
	struct test_resource_thread_args args[NUM_OF_QUEUES];
	struct mlx5dv_hws_resource_attr resource_attr = {0};
	struct mlx5dv_hws_resource *resource;
	pthread_t threads[NUM_OF_QUEUES];
	uint8_t *offset_taken;
	int *error, ret;
	int i;

	offset_taken = calloc((1 << BUF_LOG_SIZE), sizeof(*offset_taken));
	error = calloc(1, sizeof(*error));
	resource_attr.bulk_log_size = BUF_LOG_SIZE;
	resource_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK;
	resource = mlx5dv_hws_resource_alloc(ctx, &resource_attr);
	if (!resource) {
		printf("Failed to allocate resource\n");
		return -1;
	}

	for (i = 0; i < NUM_OF_QUEUES; ++i) {
		args[i].resource = resource;
		args[i].thread_id = i;
		args[i].offset_taken = offset_taken;
		args[i].error = error;

		if (pthread_create(&threads[i], NULL, test_resource_thread, &args[i]) != 0) {
			printf("Failed to create thread %d\n", i);
			return -1;
		}
	}

	for (i = 0; i < NUM_OF_QUEUES; ++i)
		pthread_join(threads[i], NULL);

	mlx5dv_hws_resource_free(resource);
	free(offset_taken);
	ret = *error;
	free(error);

	return ret;
}

int test_resource(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_context *ctx;
	struct ibv_pd *pd;

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	pd = ibv_alloc_pd(ibv_ctx);
	if (!pd) {
		TEST_LOG("Failed to allocate PD\n");
		goto out_err;
	}

	ctx_attr.pd = pd;
	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context\n");
		goto dealloc_pd;
	}

	if (test_resource_create_resource(pd, ctx)) {
		TEST_LOG("test_resource_create_resource failed!\n");
		goto close_ctx;
	}

	if (test_resource_enqueue_query_config(pd, ctx)) {
		TEST_LOG("test_resource_enqueue_query_config failed!\n");
		goto close_ctx;
	}

	if (test_resource_sequential_get_put(pd, ctx)) {
		TEST_LOG("test_resource_sequential_get_put failed!\n");
		goto close_ctx;
	}

	if (test_resource_multithread_get_put(pd, ctx)) {
		TEST_LOG("test_resource_multithread_get_put failed!\n");
		goto close_ctx;
	}

	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(pd);

	return 0;

close_ctx:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(pd);
out_err:
	return -1;
}
