/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test_resource.h"

#define NUM_OF_QUEUES 16
#define ROW_LOG 10
#define NUM_OF_RULES (1 << ROW_LOG)
#define CTR_BULK_LOG ROW_LOG
#define NUM_OF_CTRS (1 << CTR_BULK_LOG)
#define QUEUE_SIZE 256
#define BURST_TH 32
#define CTR_QUEUE_ID 1

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
		ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			printf("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

static void read_aso_counter_resource(struct mlx5dv_hws_context *ctx,
				      struct mlx5dv_hws_resource *resource,
				      struct buf_sp *sp,
				      uint32_t offset)
{
	struct mlx5dv_hws_resource_enqueue_aso_attr enqueue_attr = {0};
	struct mlx5_ifc_traffic_counter aso_ctr_data = {0};
	struct aso_cseg_input aso_cseg_in = {0};
	struct aso_wqe aso_wqe = {0};
	uint64_t packets_count = 0;
	uint64_t octets_count = 0;
	uint32_t pending_ctr = 0;
	uint32_t miss_count = 0;
	uint32_t user_data = 0;
	uint32_t hw_bp = 0;

	/* Fill ASO WQE control segment */
	aso_cseg_in.operand_masks =
		(ASO_BYTEWISE_64BYTE << ASO_CSEG_DATA_MASK_MODE_OFFSET) |
		(ASO_OPER_LOGICAL_OR << ASO_CSEG_COND_OPER_OFFSET) |
		(ASO_OP_ALWAYS_FALSE << ASO_CSEG_COND_1_OPER_OFFSET) |
		(ASO_OP_ALWAYS_FALSE << ASO_CSEG_COND_0_OPER_OFFSET);
	aso_cseg_in.data_mask = UINT64_MAX;

	aso_fill_send_cseg((uint64_t)sp->buf, sp->mkey->id, &aso_wqe.aso_ctrl, true);
	aso_fill_condition_cseg(&aso_cseg_in, &aso_wqe.aso_ctrl);

	enqueue_attr.burst = false;
	enqueue_attr.user_data = &user_data;
	/* ASO Counter is of granularity of 4, align offset */
	enqueue_attr.offset = offset & 0xfffffffc;

	if (mlx5dv_hws_resource_enqueue_aso_wqe(resource, CTR_QUEUE_ID, (__be32 *)&aso_wqe,
						sizeof(struct aso_wqe),
						&enqueue_attr)) {
		TEST_LOG("Failed to read counter at offset 0x%x", offset);
		return;
	}

	pending_ctr++;

	poll_for_comp(ctx, CTR_QUEUE_ID, &pending_ctr, BURST_TH, &miss_count, &hw_bp, true);

	/* ASO Counter is granularity of 4 so use % to get the relevant offset */
	memcpy(&aso_ctr_data,
	       sp->buf + ((offset % 4) * sizeof(struct mlx5_ifc_traffic_counter)),
	       sizeof(struct mlx5_ifc_traffic_counter));

	packets_count = ((uint64_t)be32toh(aso_ctr_data.packets_63_32) << 32) |
			be32toh(aso_ctr_data.packets_31_0);
	octets_count = ((uint64_t)be32toh(aso_ctr_data.octets_63_32) << 32) |
			be32toh(aso_ctr_data.octets_31_0);

	printf("counter 0x%x results\n", offset);
	printf("Total number of packets counted 0x%lx\n", packets_count);
	printf("Total number of bytes counted  0x%lx\n", octets_count);
}

int test_resource_aso_counter(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_resource_attr resource_attr = {0};
	struct mlx5dv_hws_action_data actions_data[1] = {0};
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_resource *all_resources[4];
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_action *dest_action;
	struct mlx5dv_hws_resource *resource;
	struct mlx5dv_hws_matcher *matcher;
	struct mlx5dv_hws_action *aso_ctr;
	struct mlx5dv_hws_item item = {0};
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	struct mlx5dv_hws_table *tbl;
	uint32_t pending_rules = 0;
	struct mlx5dv_hws_at *at;
	struct mlx5dv_hws_mt *mt;
	uint32_t miss_count = 0;
	struct buf_sp sp = {0};
	uint16_t queue_id = 0;
	uint32_t hw_bp = 0;
	int ret, i;

	ctx_attr.pd = ibv_alloc_pd(ibv_ctx);
	if (!ctx_attr.pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	}

	/* First create buffer to read ASO WQE operations from it */
	if (buf_create(ctx, &sp, MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER)) {
		TEST_LOG("Failed to create buffer");
		goto ctx_close;
	}

	/* Allocate ASO counter resource */
	resource_attr.bulk_log_size = CTR_BULK_LOG;
	resource_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER;
	resource = mlx5dv_hws_resource_alloc(ctx, &resource_attr);
	if (!resource) {
		TEST_LOG("Failed to allocate ASO counter resource");
		goto buf_destroy;
	}

	/* Create counter action */
	all_resources[0] = resource;
	aso_ctr = mlx5dv_hws_action_create_counter(ctx, all_resources, 1,
						   MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!aso_ctr) {
		TEST_LOG("Failed to create counter action");
		goto resource_destroy;
	}

	/* Create HWS Table */
	tbl_attr.level = 1;
	tbl_attr.type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
	tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!tbl) {
		TEST_LOG("Failed to create table");
		goto aso_ctr_destroy;
	}

	ret = create_basic_item(&item);
	if (ret) {
		TEST_LOG("Failed to create item");
		goto tbl_destroy;
	}

	mt_attr.comp_mask = 0;
	mt = mlx5dv_hws_match_template_create(ctx, &item, 1, &mt_attr);
	if (!mt) {
		TEST_LOG("Failed to create match template");
		goto item_destroy;
	}

	at = mlx5dv_hws_action_template_create(&aso_ctr, 1, &at_attr);
	if (!at) {
		TEST_LOG("Failed to create dummy action template");
		goto mt_destroy;
	}

	/* Create HWS Matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = ROW_LOG;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;
	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, &matcher_attr);
	if (!matcher) {
		TEST_LOG("Failed to create matcher");
		goto at_destroy;
	}

	dest_action = mlx5dv_hws_action_create_dest_default_miss(ctx,
								 MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!dest_action) {
		TEST_LOG("Failed to create default miss action\n");
		goto matcher_destroy;
	}

	/* Dump once all control data created */
	dump_all(ctx);

	rules = calloc(NUM_OF_RULES, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		goto default_miss_destroy;
	}

	rule_attr.queue_id = queue_id;

	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules +
						  i * mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		/* Attach ASO counter resource at offset i to the counter
		 * action in the rule.
		 */
		actions_data[0].counter.resource_offset = i;

		DEVX_SET(my_basic_item, (void *)item.data.match, sip, i);
		ret = mlx5dv_hws_rule_create(matcher, 0, &item.data, 0, actions_data,
					     dest_action, &rule_attr, rule);
		if (ret)
			TEST_LOG("Failed to create rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules,
			      BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Insertion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	/* Read random counter resource data */
	srand(time(0));
	read_aso_counter_resource(ctx, resource, &sp, ((uint32_t)rand()) % (NUM_OF_CTRS));

	pending_rules = 0;
	miss_count = 0;
	hw_bp = 0;

	for (i = 0; i < NUM_OF_RULES; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules +
						  i * mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = !((i + 1) % BURST_TH == 0);
		rule_attr.rule_idx = i;

		ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
		if (ret)
			TEST_LOG("Failed to destroy rule %d", i);

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules,
			      BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Deletion. Total misses: %u (out of: %u) HW bp %u\n",
		queue_id,
		miss_count,
		NUM_OF_RULES,
		hw_bp);

	free(rules);
	mlx5dv_hws_action_destroy(dest_action);
	mlx5dv_hws_matcher_destroy(matcher);
	mlx5dv_hws_action_template_destroy(at);
	mlx5dv_hws_match_template_destroy(mt);
	destroy_basic_item(&item);
	mlx5dv_hws_table_destroy(tbl);
	mlx5dv_hws_action_destroy(aso_ctr);
	mlx5dv_hws_resource_free(resource);
	buf_destroy(&sp);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(ctx_attr.pd);

	return 0;

default_miss_destroy:
	mlx5dv_hws_action_destroy(dest_action);
matcher_destroy:
	mlx5dv_hws_matcher_destroy(matcher);
at_destroy:
	mlx5dv_hws_action_template_destroy(at);
mt_destroy:
	mlx5dv_hws_match_template_destroy(mt);
item_destroy:
	destroy_basic_item(&item);
tbl_destroy:
	mlx5dv_hws_table_destroy(tbl);
aso_ctr_destroy:
	mlx5dv_hws_action_destroy(aso_ctr);
resource_destroy:
	mlx5dv_hws_resource_free(resource);
buf_destroy:
	buf_destroy(&sp);
ctx_close:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(ctx_attr.pd);
out_err:
	TEST_LOG("Test failed errno=%d", errno);
	return -errno;
}
