/* Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_test_resource.h"

#define QUEUE_ID 0
#define BURST_TH 1

struct mlx5dv_hws_resource_enqueue_aso_attr enqueue_attr = {0};
struct mlx5dv_hws_resource *resource;
struct buf_sp sp = {0};

static int poll_for_comp(struct mlx5dv_hws_context *ctx,
			 uint16_t queue_id,
			 uint32_t *pending_rules,
			 uint32_t expected_comp,
			 uint32_t *miss_count,
			 uint32_t *hw_bp,
			 bool drain)
{
	bool queue_full = *pending_rules == QUEUE_SIZE;
	bool got_comp = *pending_rules >= expected_comp;
	struct mlx5dv_hws_queue_op_result comp[BURST_TH];
	bool first_comp = true;
	int ret;
	int j;

	/* Check if there are any completions at all */
	if (!got_comp && !drain)
		return 0;

	while (queue_full || ((got_comp || drain) && *pending_rules)) {
	       ret = mlx5dv_hws_queue_poll(ctx, queue_id, comp, expected_comp);
		if (ret < 0) {
			printf("Failed during poll queue\n");
			return -1;
		}

		if (ret) {
			(*pending_rules) -= ret;
			for (j = 0; j < ret; j++) {
				if (comp[j].status == MLX5DV_HWS_QUEUE_OP_ERROR)
					(*miss_count)++;
			}

			queue_full = false;
		} else if (queue_full && first_comp) {
			first_comp = false;
			(*hw_bp)++;
		}

		got_comp = !!ret;
	}

	return 0;
}

static int create_regc1_item(struct mlx5dv_hws_item *item)
{
	struct mlx5dv_hws_field *fields;
	void *mask, *match;

	fields = calloc(BASIC_ITEM_NUM_OF_FIELDS, sizeof(struct mlx5dv_hws_field));
	if (!fields) {
		TEST_LOG("Failed to allocate fields");
		return ENOMEM;
	}

	match = calloc(1, DEVX_ST_SZ_BYTES(my_basic_item));
	if (!match) {
		TEST_LOG("Failed to allocate match");
		free(fields);
		return ENOMEM;
	}

	mask = calloc(1, DEVX_ST_SZ_BYTES(my_basic_item));
	if (!mask) {
		TEST_LOG("Failed to allocate mask");
		free(match);
		free(fields);
		return ENOMEM;
	}

	item->num_of_fields = BASIC_ITEM_NUM_OF_FIELDS;
	item->fields = fields;
	item->data.mask = mask;
	item->data.match = match;

	/* Match on REGC1 */
	fields[0].fname = MLX5DV_HWS_FNAME_REG_C1;
	fields[0].header = MLX5DV_HWS_HEADER_NONE;
	fields[0].op = MLX5DV_HWS_FIELD_MATCH_OP_EQ;
	fields[0].type = MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
	fields[0].value.bit_off = CALC_BIT_OFF(my_basic_item, sip);
	/* The result will be set in the 8 LSB of the register
	 * according to PRM.
	 */
	DEVX_SET(my_basic_item, mask, sip, 0xff);
	/* Match on 0x2 value, which means was colored green
	 * by the meter according to PRM.
	 */
	DEVX_SET(my_basic_item, match, sip, 0x2);

	return 0;
}

static void destroy_regc1_item(struct mlx5dv_hws_item *item)
{
        free((void*)item->data.match);
        free((void*)item->data.mask);
        free(item->fields);
}

static int create_aso_meter_resource(struct mlx5dv_hws_context *ctx)
{
	uint32_t in[DEVX_ST_SZ_DW(flow_meter_parameter)] = {0};
	struct mlx5dv_hws_resource_attr resource_attr = {0};
	struct aso_cseg_input aso_cseg_in = {0};
	struct aso_wqe aso_wqe = {0};
	uint32_t pending_aso = 0;
	uint32_t miss_count = 0;
	uint32_t user_data = 0;
	uint32_t hw_bp = 0;

	/* ASO meter object creation */

	/* First create buffer to read ASO WQE operations from it */
	if (buf_create(ctx, &sp, MLX5DV_HWS_RESOURCE_TYPE_ASO_METER))
		return -1;

	/* Allocate ASO meter resource  */
	resource_attr.bulk_log_size = BUF_LOG_SIZE;
	resource_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ASO_METER;
	resource = mlx5dv_hws_resource_alloc(ctx, &resource_attr);
	if (!resource) {
		TEST_LOG("Failed to allocate resource!");
		goto destroy_buf;
	}

	/* Fill ASO WQE control segment */
	aso_cseg_in.operand_masks =
		(ASO_BYTEWISE_64BYTE << ASO_CSEG_DATA_MASK_MODE_OFFSET) |
		(ASO_OPER_LOGICAL_OR << ASO_CSEG_COND_OPER_OFFSET) |
		(ASO_OP_ALWAYS_TRUE << ASO_CSEG_COND_1_OPER_OFFSET) |
		(ASO_OP_ALWAYS_TRUE << ASO_CSEG_COND_0_OPER_OFFSET);
	aso_cseg_in.data_mask = UINT64_MAX;

	aso_fill_send_cseg((uint64_t)sp.buf, sp.mkey->id, &aso_wqe.aso_ctrl, true);
	aso_fill_condition_cseg(&aso_cseg_in, &aso_wqe.aso_ctrl);

	/* Fill ASO WQE data segment */
	DEVX_SET(flow_meter_parameter, in, valid, 1);
	DEVX_SET(flow_meter_parameter, in, bucket_overflow, 1);
	/* Start with green */
	DEVX_SET(flow_meter_parameter, in, start_color, 2);
	DEVX_SET(flow_meter_parameter, in, both_buckets_on_green, 0);
	/* Packet bytes are calculated by the device including L2 headers */
	DEVX_SET(flow_meter_parameter, in, meter_mode, 1);
	/* Rate is given in bit per second and is calculated by
	 * 8*(10^9)*cir_mantissa*2^(-1*cir_ex-ponent).
	 * for cir_mantissa=4 and cir_exponent=6 the rate will
	 * be 62.5MBps = 500Mbps
	 */
	DEVX_SET(flow_meter_parameter, in, cir_mantissa, 4);
	DEVX_SET(flow_meter_parameter, in, cir_exponent, 6);

	memcpy(aso_wqe.bytewise_data, in, sizeof(in));

	/* Get empty ASO meter object offset and write the ASO meter
	 * to it, then poll the queue.
	 */
	enqueue_attr.burst = false;
	enqueue_attr.user_data = &user_data;
	if (mlx5dv_hws_resource_get_offset(resource, QUEUE_ID, &enqueue_attr.offset))
		goto free_resource;

	if (mlx5dv_hws_resource_enqueue_aso_wqe(resource, QUEUE_ID,
						(__be32 *)&aso_wqe,
						sizeof(struct aso_wqe),
						&enqueue_attr))
		goto put_resource_offset;

	pending_aso++;
	poll_for_comp(ctx, QUEUE_ID, &pending_aso,
		      BURST_TH, &miss_count, &hw_bp, true);

	return 0;

put_resource_offset:
	mlx5dv_hws_resource_put_offset(resource, 0, enqueue_attr.offset);
free_resource:
	mlx5dv_hws_resource_free(resource);
destroy_buf:
	buf_destroy(&sp);
	return -1;
}

static void destroy_aso_meter_resource()
{
	mlx5dv_hws_resource_put_offset(resource, 0, enqueue_attr.offset);
	mlx5dv_hws_resource_free(resource);
	buf_destroy(&sp);
}

int test_resource_aso_meter(struct ibv_context *ibv_ctx, struct test_info *t_info)
{
	struct mlx5dv_hws_action *dest_table, *default_miss, *aso_meter;
	struct mlx5dv_hws_action_data actions_data[1] = {0};
	struct mlx5dv_hws_matcher_attr matcher_attr = {0};
	struct mlx5dv_hws_matcher *f_matcher, *s_matcher;
	struct mlx5dv_hws_context_attr ctx_attr = {0};
	struct mlx5dv_hws_resource *all_resources[4];
	struct mlx5dv_hws_table_attr tbl_attr = {0};
	struct mlx5dv_hws_rule_attr rule_attr = {0};
	struct mlx5dv_hws_mt *basic_mt, *regc1_mt;
	struct mlx5dv_hws_mt_attr mt_attr = {0};
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_item basic_item = {0};
	struct mlx5dv_hws_item regc1_item = {0};
	struct mlx5dv_hws_table *f_tbl, *s_tbl;
	struct mlx5dv_hws_at *f_at, *s_at;
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_rule *rules;
	struct mlx5dv_hws_rule *rule;
	uint32_t pending_rules = 0;
	uint32_t miss_count = 0;
	uint32_t hw_bp = 0;
	struct ibv_pd *pd;
	int ret, i;

	ctx_attr.queues = NUM_OF_QUEUES;
	ctx_attr.queue_size = QUEUE_SIZE;

	pd = ibv_alloc_pd(ibv_ctx);
	if (!pd) {
		TEST_LOG("Failed to allocate PD\n");
		return -errno;
	}

	ctx_attr.pd = pd;
	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context\n");
		ret = -errno;
		goto dealloc_pd;
	}

	/* Create HWS first table */
	tbl_attr.level = 1;
	f_tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!f_tbl) {
		TEST_LOG("Failed to create first table");
		goto close_ctx;
	}

	/* Create HWS second table */
	tbl_attr.level = 2;
	s_tbl = mlx5dv_hws_table_create(ctx, &tbl_attr);
	if (!s_tbl) {
		TEST_LOG("Failed to create second table");
		goto f_tbl_destroy;
	}

	/* Create MT for IPv4 SRC & DST and IP version*/
	ret = create_basic_item(&basic_item);
	if (ret) {
		TEST_LOG("Failed to create basic item");
		goto s_tbl_destroy;
	}

	mt_attr.comp_mask = 0;
	basic_mt = mlx5dv_hws_match_template_create(ctx, &basic_item, 1, &mt_attr);
	if (!basic_mt) {
		TEST_LOG("Failed to create first match template");
		goto basic_item_destroy;
	}

	f_at = mlx5dv_hws_action_template_create(NULL, 0, &at_attr);
	if (!f_at) {
		TEST_LOG("Failed to create second action template");
		goto basic_mt_destroy;
	}

	/* Create first matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = 0;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;

	f_matcher = mlx5dv_hws_matcher_create(f_tbl, &basic_mt, 1, &f_at, 1, &matcher_attr);
	if (!f_matcher) {
		TEST_LOG("Failed to first create matcher");
		goto f_at_destroy;
	}

	dest_table = mlx5dv_hws_action_create_dest_table(ctx, s_tbl, MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!dest_table) {
		TEST_LOG("Failed to create dest table action\n");
		goto f_matcher_destroy;
	}

	rules = calloc(2, mlx5dv_hws_rule_get_handle_size());
	if (!rules) {
		TEST_LOG("Failed to create rule array");
		goto dest_table_destroy;
	}

	/* First rule creation */
	rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules);
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.rule_idx = 0;
	rule_attr.queue_id = QUEUE_ID;

	ret = mlx5dv_hws_rule_create(f_matcher, 0, &basic_item.data, 0, NULL,
				     dest_table, &rule_attr, rule);
	if (ret) {
		TEST_LOG("Failed to create first rule");
		goto rules_destroy;
	}

	pending_rules++;

	poll_for_comp(ctx, QUEUE_ID, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Insertion. Total misses: %u (out of: 1) HW bp %u\n",
	       QUEUE_ID,
	       miss_count,
	       hw_bp);

	/* Create MT for REGC 1 */
	ret = create_regc1_item(&regc1_item);
	if (ret) {
		TEST_LOG("Failed to create REGC 1 item");
		goto f_rule_destroy;
	}

	/* Create ASO meter resource */
	ret = create_aso_meter_resource(ctx);
	if (ret) {
		TEST_LOG("Failed to create ASO meter resource");
		goto regc1_item_destroy;
	}

	/* Create ASO meter action with reg_id 0x1 which will set the ASO
	 * result at the 8LSB of regc 1.
	 */

	all_resources[0] = resource;
	aso_meter = mlx5dv_hws_action_create_aso_meter(ctx, all_resources, 1, 0x1,
						       MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!aso_meter) {
		TEST_LOG("Failed to create ASO meter action\n");
		goto resource_destroy;
	}

	mt_attr.comp_mask = 0;
	regc1_mt = mlx5dv_hws_match_template_create(ctx, &regc1_item, 1, &mt_attr);
	if (!regc1_mt) {
		TEST_LOG("Failed to create match template");
		goto aso_meter_destroy;
	}

	s_at = mlx5dv_hws_action_template_create(&aso_meter, 1, &at_attr);
	if (!s_at) {
		TEST_LOG("Failed to create second action template");
		goto regc1_mt_destroy;
	}

	/* Create second matcher */
	matcher_attr.comp_mask = 0;
	matcher_attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE;
	matcher_attr.rule.num_log = 0;
	matcher_attr.optimize_using_rule_idx = 1;
	matcher_attr.priority = 0;

	s_matcher = mlx5dv_hws_matcher_create(s_tbl, &regc1_mt, 1, &s_at, 1, &matcher_attr);
	if (!f_matcher) {
		TEST_LOG("Failed to create second matcher");
		goto s_at_destroy;
	}

	default_miss = mlx5dv_hws_action_create_dest_default_miss(ctx, MLX5DV_HWS_ACTION_FLAG_NIC_RX);
	if (!default_miss) {
		TEST_LOG("Failed to create default miss action\n");
		goto s_matcher_destroy;
	}

	/* Second rule creation */
	rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules + mlx5dv_hws_rule_get_handle_size());
	rule_attr.user_data = rule;
	rule_attr.burst = 0;
	rule_attr.rule_idx = 0;
	rule_attr.queue_id = QUEUE_ID;

	actions_data[0].aso_meter.resource_offset = 0;
	actions_data[0].aso_meter.resource_idx = 0;
	actions_data[0].aso_meter.init_color = 2;//Start from green
	actions_data[0].aso_meter.aso_wqe = NULL;

	ret = mlx5dv_hws_rule_create(s_matcher, 0, &basic_item.data, 0, actions_data,
				     default_miss, &rule_attr, rule);
	if (ret) {
		TEST_LOG("Failed to create second rule");
		goto default_miss_destroy;
	}

	pending_rules++;

	poll_for_comp(ctx, QUEUE_ID, &pending_rules, BURST_TH, &miss_count, &hw_bp, true);
	printf("queue [%d] Insertion. Total misses: %u (out of: 1) HW bp %u\n",
	       QUEUE_ID,
	       miss_count,
	       hw_bp);

	dump_all(ctx);

	/* Delete the rules */
	for (i = 0; i < 2; i++) {
		rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules +
						  i * mlx5dv_hws_rule_get_handle_size());
		rule_attr.user_data = rule;
		rule_attr.burst = 0;
		rule_attr.rule_idx = 0;
		rule_attr.queue_id = QUEUE_ID;

		ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
		if (ret) {
			TEST_LOG("Failed to destroy rule %d", i);
			goto default_miss_destroy;
		}

		pending_rules++;

		poll_for_comp(ctx, rule_attr.queue_id, &pending_rules,
			      BURST_TH, &miss_count, &hw_bp, false);
	}

	poll_for_comp(ctx, rule_attr.queue_id, &pending_rules, BURST_TH,
		      &miss_count, &hw_bp, false);
	printf("queue [%d] Deletion. Total misses: %u (out of: %u) HW bp %u\n",
			QUEUE_ID,
			miss_count,
			2,
			hw_bp);


	free(rules);
	mlx5dv_hws_action_destroy(default_miss);
	mlx5dv_hws_matcher_destroy(s_matcher);
	mlx5dv_hws_action_template_destroy(s_at);
	mlx5dv_hws_action_destroy(aso_meter);
	destroy_aso_meter_resource();
	mlx5dv_hws_match_template_destroy(regc1_mt);
	destroy_regc1_item(&regc1_item);
	mlx5dv_hws_action_destroy(dest_table);
	mlx5dv_hws_matcher_destroy(f_matcher);
	mlx5dv_hws_action_template_destroy(f_at);
	mlx5dv_hws_match_template_destroy(basic_mt);
	destroy_basic_item(&basic_item);
	mlx5dv_hws_table_destroy(s_tbl);
	mlx5dv_hws_table_destroy(f_tbl);
	mlx5dv_hws_context_close(ctx);
	ibv_dealloc_pd(pd);

	return 0;

default_miss_destroy:
	mlx5dv_hws_action_destroy(default_miss);
s_matcher_destroy:
	mlx5dv_hws_matcher_destroy(s_matcher);
s_at_destroy:
	mlx5dv_hws_action_template_destroy(s_at);
aso_meter_destroy:
        mlx5dv_hws_action_destroy(aso_meter);
resource_destroy:
        destroy_aso_meter_resource();
regc1_mt_destroy:
	mlx5dv_hws_match_template_destroy(regc1_mt);
regc1_item_destroy:
	destroy_regc1_item(&regc1_item);
f_rule_destroy:
	rule = (struct mlx5dv_hws_rule *)((uint8_t *)rules);
	rule_attr.user_data = rule;
	rule_attr.burst = 1;
	rule_attr.rule_idx = 0;

	ret = mlx5dv_hws_rule_destroy(rule, &rule_attr);
	if (ret) {
		TEST_LOG("Failed to destroy rule %d", i);
	}

	pending_rules++;

	poll_for_comp(ctx, QUEUE_ID, &pending_rules, BURST_TH, &miss_count, &hw_bp, false);
rules_destroy:
	free(rules);
dest_table_destroy:
	mlx5dv_hws_action_destroy(dest_table);
f_matcher_destroy:
	mlx5dv_hws_matcher_destroy(f_matcher);
f_at_destroy:
	mlx5dv_hws_action_template_destroy(f_at);
basic_mt_destroy:
	mlx5dv_hws_match_template_destroy(basic_mt);
basic_item_destroy:
	destroy_basic_item(&basic_item);
s_tbl_destroy:
	mlx5dv_hws_table_destroy(s_tbl);
f_tbl_destroy:
	mlx5dv_hws_table_destroy(f_tbl);
close_ctx:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(pd);
	return -1;
}
