// SPDX-License-Identifier: NVIDIA CORPORATION & AFFILIATES
/*
 * Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include <pthread.h>

#include "hws_test.h"

#define RESOURCE_LOG_SIZE 20
#define NUM_THREADS 4
#define QUEUE_SIZE 256
#define BLOCK_ALLOC 1
#define MAX_LOG_SZ 3

#define BM_NUM_ITERS 5
#define BM_NUM_THREADS 1
#define BM_MAX_LOG_SZ 0

#define PRINT_TIMING_INFO 0

struct offset_with_size {
	uint32_t offset;
	uint32_t log_sz;
};

struct thread_state {
	int thread_id;
	struct mlx5dv_hws_resource *resource;
	pthread_barrier_t *start_barrier;
	bool allocate_everything;
	struct offset_with_size *results;
	int num_results;
	double alloc_time_diff;
	double free_time_diff;
};

struct bm_thread_state {
	int thread_id;
	struct mlx5dv_hws_resource *resource;
	pthread_barrier_t *start_barrier;
	unsigned int rand_seed;
	int max_log_sz;
	struct offset_with_size *results;
	int num_results;
	int num_allocs, num_frees;
	double alloc_time_diff, free_time_diff;
};

/* Test that we respect the minimum block size for small resources. */
static int test_min_block_size(struct mlx5dv_hws_context *ctx)
{
	struct mlx5dv_hws_resource_attr resource_attr = { 0 };
	const u32 min_expected_block_size = 2;
	struct mlx5dv_hws_resource *resource;
	u32 block_size;
	int ret = 0;

	resource_attr.bulk_log_size = 5;
	resource_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B;
	resource_attr.block_alloc_support = 1;
	resource = mlx5dv_hws_resource_alloc(ctx, &resource_attr);
	if (!resource) {
		TEST_LOG("Failed to allocate resource");
		return -1;
	}

	block_size = mlx5dv_hws_resource_max_block_size(resource);
	if (block_size < min_expected_block_size) {
		TEST_LOG("Max block size is %d, expected at least %d", block_size,
			 min_expected_block_size);
		ret = -1;
	}

	mlx5dv_hws_resource_free(resource);

	return ret;
}

/* Test that even with the minimum block size we do not exceed the entire resource size. */
static int test_min_block_size_small_resouce(struct mlx5dv_hws_context *ctx)
{
	struct mlx5dv_hws_resource_attr resource_attr = { 0 };
	struct mlx5dv_hws_resource *resource;
	const u32 resource_log_size = 1;
	u32 block_size;
	int ret = 0;

	resource_attr.bulk_log_size = resource_log_size;
	resource_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B;
	resource_attr.block_alloc_support = 1;
	resource = mlx5dv_hws_resource_alloc(ctx, &resource_attr);
	if (!resource) {
		TEST_LOG("Failed to allocate resource");
		return -1;
	}

	block_size = mlx5dv_hws_resource_max_block_size(resource);
	if (block_size != resource_log_size) {
		TEST_LOG("Max block size is %d, expected %d", block_size, resource_log_size);
		ret = -1;
	}

	mlx5dv_hws_resource_free(resource);

	return ret;
}

/* Allocate all available offsets in chunks of the given size, on a
 * single thread.
 */
static int test_homogeneous_get(struct mlx5dv_hws_resource *resource,
				uint8_t log_sz)
{
	int num_offsets, i, err;
	uint32_t last_offset;
	uint32_t *offsets;

	num_offsets = 1 << (RESOURCE_LOG_SIZE - log_sz);
	offsets = calloc(num_offsets, sizeof(*offsets));
	if (!offsets) {
		printf("Failed to allocate offsets array\n");
		return -1;
	}

	for (i = 0; i < num_offsets; i++) {
		err = mlx5dv_hws_resource_get_offset_block(resource, 0, log_sz,
							   &offsets[i]);
		if (err) {
			TEST_LOG("Failed to allocate offset log_sz %d index %d expected %d",
				 log_sz, i, num_offsets);
			goto put_offsets;
		}
	}

	/* Also test that another get will fail, we have now allocated
	 * the entire address space.
	 */
	err = mlx5dv_hws_resource_get_offset_block(resource, 0, log_sz,
						   &last_offset);
	if (err == 0) {
		TEST_LOG("Allocation should have failed, but did not, log_sz %d offset %d",
			 log_sz, last_offset);
		err = -1; /* So we correctly signal the test as failed. */
		goto put_offsets;
	}

	err = 0;

put_offsets:
	while (--i >= 0) {
		err = mlx5dv_hws_resource_put_offset_block(resource, 0, log_sz,
							   offsets[i]);
		if (err) {
			TEST_LOG("Failed to free offset %d log_sz %d",
				 offsets[i], log_sz);
			break;
		}
	}
	free(offsets);
	return err;
}

static void add_result(struct offset_with_size *results, int *num,
		       uint32_t offset, uint32_t log_sz)
{
	struct offset_with_size *res = &results[*num];

	res->offset = offset;
	res->log_sz = log_sz;
	(*num)++;
}

static void *alloc_thread(void *arg)
{
	struct thread_state *st = arg;
	int err, log_sz, max_log_sz;
	struct timespec start, end;
	unsigned int rand_seed;
	uint32_t offset;

	rand_seed = time(NULL) * (1 + st->thread_id);
	st->num_results = 0;
	max_log_sz = min(mlx5dv_hws_resource_max_block_size(st->resource),
			 MAX_LOG_SZ);

	pthread_barrier_wait(st->start_barrier);
	clock_gettime(CLOCK_REALTIME, &start);

	do {
		log_sz = rand_r(&rand_seed) % (1 + max_log_sz);
		err = mlx5dv_hws_resource_get_offset_block(
			st->resource, st->thread_id, log_sz, &offset);
		if (err == 0)
			add_result(st->results, &st->num_results, offset,
				   log_sz);
	} while (err == 0);

	if (st->allocate_everything) {
		/* Allocation failed for `log_sz`, but allocate
		 * everything possible from lower block sizes.
		 */
		for (log_sz--; log_sz >= 0; log_sz--) {
			do {
				err = mlx5dv_hws_resource_get_offset_block(
					st->resource, st->thread_id, log_sz,
					&offset);
				if (err == 0)
					add_result(st->results,
						   &st->num_results, offset,
						   log_sz);
			} while (err == 0);
		}
	}

	clock_gettime(CLOCK_REALTIME, &end);
	st->alloc_time_diff = timediff(&start, &end);

	return NULL;
}

static void *free_thread(void *arg)
{
	struct thread_state *st = arg;
	struct timespec start, end;
	int i;

	pthread_barrier_wait(st->start_barrier);
	clock_gettime(CLOCK_REALTIME, &start);
	for (i = 0; i < st->num_results; i++)
		mlx5dv_hws_resource_put_offset_block(st->resource,
						     st->thread_id,
						     st->results[i].log_sz,
						     st->results[i].offset);
	clock_gettime(CLOCK_REALTIME, &end);
	st->free_time_diff = timediff(&start, &end);

	return NULL;
}

static void accumulate_coverage(int *cov, struct offset_with_size *results,
				int num_results)
{
	int i, j;

	for (i = 0; i < num_results; i++)
		for (j = 0; j < (1 << results[i].log_sz); j++)
			cov[results[i].offset + j]++;
}

static int check_and_log_coverage(const struct mlx5dv_hws_resource *resource,
				  const int *cov, bool allocate_everything)
{
	const int max_log_sz = mlx5dv_hws_resource_max_block_size(resource);
	const int num_offsets = 1 << RESOURCE_LOG_SIZE;
	int i, sum = 0;

	for (i = 0; i < num_offsets; i++) {
		if (cov[i] > 1) {
			TEST_LOG("Index %d was allocated %d times", i, cov[i]);
			return -1;
		}
		sum += cov[i];
	}

	if (max_log_sz == 0 || allocate_everything) {
		if (sum != num_offsets) {
			TEST_LOG("Expected complete coverage, but failed %d/%d",
				 sum, num_offsets);
			return -1;
		}
	} else {
		printf("Coverage: %g%%\n", (float)sum / num_offsets * 100);
	}

	return 0;
}

static void randomize_order(struct offset_with_size *results, int num_results)
{
	struct offset_with_size tmp;
	int i, src, dst;

	srand(time(NULL));
	for (i = 0; i < num_results; i++) {
		src = rand() % num_results;
		dst = rand() % num_results;
		tmp = results[src];
		results[src] = results[dst];
		results[dst] = tmp;
	}
}

static int test_multi_thread_one_iteration(struct mlx5dv_hws_resource *resource,
					   pthread_t *threads,
					   struct thread_state *thread_states,
					   bool allocate_everything)
{
	pthread_barrier_t *start_barrier = thread_states[0].start_barrier;
	int i, err, *cov;

	err = pthread_barrier_init(start_barrier, NULL, NUM_THREADS + 1);
	if (err != 0) {
		TEST_LOG("Failed to initialize barrier");
		return -1;
	}

	for (i = 0; i < NUM_THREADS; i++) {
		struct thread_state *st = &thread_states[i];

		st->num_results = 0;
		err = pthread_create(&threads[i], NULL, alloc_thread, st);
		if (err != 0) {
			TEST_LOG("Failed to start thread %d", i);
			return -1;
		}
	}

	pthread_barrier_wait(start_barrier);

	for (int i = 0; i < NUM_THREADS; i++) {
		err = pthread_join(threads[i], NULL);
		if (err) {
			TEST_LOG("Failed to join thread %d", i);
			return -1;
		}
	}

	/* Allocations are done, check and log coverage. */
	cov = calloc(1 << RESOURCE_LOG_SIZE, sizeof(*cov));
	if (!cov) {
		TEST_LOG("Failed to allocate coverage data");
		return -1;
	}

	for (int i = 0; i < NUM_THREADS; i++)
		accumulate_coverage(cov, thread_states[i].results,
				    thread_states[i].num_results);
	if (check_and_log_coverage(resource, cov, allocate_everything))
		return -1;

	/* Now free up everything. */
	err = pthread_barrier_init(start_barrier, NULL, NUM_THREADS + 1);
	if (err != 0) {
		TEST_LOG("Failed to initialize barrier");
		return -1;
	}

	for (i = 0; i < NUM_THREADS; i++) {
		randomize_order(thread_states[i].results,
				thread_states[i].num_results);
		err = pthread_create(&threads[i], NULL, free_thread,
				     &thread_states[i]);
		if (err != 0) {
			TEST_LOG("Failed to start thread %d", i);
			return -1;
		}
	}

	pthread_barrier_wait(start_barrier);

	if (PRINT_TIMING_INFO)
		printf("\n");
	for (int i = 0; i < NUM_THREADS; i++) {
		const struct thread_state *st = &thread_states[i];

		err = pthread_join(threads[i], NULL);
		if (err) {
			TEST_LOG("Failed to join thread %d", i);
			return -1;
		}
		if (PRINT_TIMING_INFO) {
			printf("Thread %d: allocs %8d\t alloc_time %g\t ",
			       st->thread_id, st->num_results,
			       st->alloc_time_diff);
			printf("free_time %8g\t total_time %8g\n",
			       st->free_time_diff,
			       st->alloc_time_diff + st->free_time_diff);
		}
	}

	return 0;
}

static int test_multi_thread_get_offset(struct mlx5dv_hws_resource *resource,
					bool allocate_everything)
{
	struct thread_state *thread_states;
	pthread_barrier_t start_barrier;
	pthread_t *threads;
	int i, err;

	threads = calloc(NUM_THREADS, sizeof(*threads));
	thread_states = calloc(NUM_THREADS, sizeof(*thread_states));

	for (i = 0; i < NUM_THREADS; i++) {
		struct thread_state *st = &thread_states[i];

		st->thread_id = i;
		st->resource = resource;
		st->start_barrier = &start_barrier;
		st->allocate_everything = allocate_everything;
		st->results =
			calloc(1 << RESOURCE_LOG_SIZE, sizeof(*st->results));
	}

	for (i = 0; i < 10; i++) {
		err = test_multi_thread_one_iteration(resource, threads,
						      thread_states,
						      allocate_everything);
		if (err)
			goto free_resources;
	}

free_resources:
	for (i = 0; i < NUM_THREADS; i++)
		free(thread_states[i].results);
	free(thread_states);
	free(threads);

	return err;
}

static void bm_alloc_all(struct bm_thread_state *st)
{
	uint32_t offset;
	uint8_t log_sz;
	int err;

	do {
		log_sz = rand_r(&st->rand_seed) % (1 + st->max_log_sz);
		err = mlx5dv_hws_resource_get_offset_block(
			st->resource, st->thread_id, log_sz, &offset);
		if (err != 0)
			break;
		add_result(st->results, &st->num_results, offset, log_sz);
	} while (1);

	st->num_allocs += st->num_results;
}

static void bm_free_all(struct bm_thread_state *st)
{
	int i;

	for (i = 0; i < st->num_results; i++)
		mlx5dv_hws_resource_put_offset_block(st->resource,
						     st->thread_id,
						     st->results[i].log_sz,
						     st->results[i].offset);

	st->num_frees += st->num_results;
	st->num_results = 0;
}

static void *bm_thread(void *arg)
{
	struct bm_thread_state *st = arg;
	struct timespec start, end;
	int i;

	pthread_barrier_wait(st->start_barrier);
	for (i = 0; i < BM_NUM_ITERS; i++) {
		clock_gettime(CLOCK_REALTIME, &start);
		bm_alloc_all(st);
		clock_gettime(CLOCK_REALTIME, &end);
		st->alloc_time_diff += timediff(&start, &end);

		clock_gettime(CLOCK_REALTIME, &start);
		bm_free_all(st);
		clock_gettime(CLOCK_REALTIME, &end);
		st->free_time_diff += timediff(&start, &end);
	}

	return NULL;
}

static int test_benchmark(struct mlx5dv_hws_resource *resource)
{
	struct bm_thread_state *thread_states;
	pthread_barrier_t start_barrier;
	pthread_t *threads;
	int i, err;

	threads = calloc(BM_NUM_THREADS, sizeof(*threads));
	thread_states = calloc(BM_NUM_THREADS, sizeof(*thread_states));

	err = pthread_barrier_init(&start_barrier, NULL, BM_NUM_THREADS + 1);
	if (err != 0) {
		TEST_LOG("Failed to initialize barrier");
		return -1;
	}

	for (i = 0; i < BM_NUM_THREADS; i++) {
		struct bm_thread_state *st = &thread_states[i];

		st->thread_id = i;
		st->resource = resource;
		st->start_barrier = &start_barrier;
		st->results =
			calloc(1 << RESOURCE_LOG_SIZE, sizeof(*st->results));
		st->rand_seed = time(NULL) * (1 + i);
		st->max_log_sz =
			min(mlx5dv_hws_resource_max_block_size(st->resource),
			    BM_MAX_LOG_SZ);
		err = pthread_create(&threads[i], NULL, bm_thread, st);
		if (err != 0) {
			TEST_LOG("Failed to start thread %d", i);
			return -1;
		}
	}

	pthread_barrier_wait(&start_barrier);

	printf("\nBenchmark results, block_alloc %d max_block_size_log %d num_threads %d\n",
	       BLOCK_ALLOC, BM_MAX_LOG_SZ, BM_NUM_THREADS);

	for (int i = 0; i < BM_NUM_THREADS; i++) {
		struct bm_thread_state *st = &thread_states[i];

		err = pthread_join(threads[i], NULL);
		if (err) {
			TEST_LOG("Failed to join thread %d", i);
			return -1;
		}
		printf("Thread %d\t alloc rate %g (%d/%gs)\t free rate %g (%d/%gs)\n",
		       i, st->num_allocs / st->alloc_time_diff, st->num_allocs,
		       st->alloc_time_diff, st->num_frees / st->free_time_diff,
		       st->num_frees, st->free_time_diff);
	}

	return 0;
}

int test_resource_block_alloc(struct ibv_context *ibv_ctx,
			      struct test_info *t_info)
{
	struct mlx5dv_hws_resource_attr resource_attr = { 0 };
	struct mlx5dv_hws_context_attr ctx_attr = { 0 };
	struct mlx5dv_hws_resource *resource;
	struct mlx5dv_hws_context *ctx;
	struct ibv_pd *pd;
	uint8_t log_sz;
	int err = -1;

	ctx_attr.queues = NUM_THREADS;
	ctx_attr.queue_size = QUEUE_SIZE;

	pd = ibv_alloc_pd(ibv_ctx);
	if (!pd) {
		TEST_LOG("Failed to allocate PD");
		goto out_err;
	}

	ctx_attr.pd = pd;
	ctx = mlx5dv_hws_context_open(ibv_ctx, &ctx_attr);
	if (!ctx) {
		TEST_LOG("Failed to create context");
		goto dealloc_pd;
	}

	if (test_min_block_size(ctx)) {
		TEST_LOG("test_min_block_size failed");
		goto close_ctx;
	}

	if (test_min_block_size_small_resouce(ctx)) {
		TEST_LOG("test_min_block_size_small_resource failed");
		goto close_ctx;
	}

	resource_attr.bulk_log_size = RESOURCE_LOG_SIZE;
	resource_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B;
	resource_attr.block_alloc_support = BLOCK_ALLOC;
	resource = mlx5dv_hws_resource_alloc(ctx, &resource_attr);
	if (!resource) {
		TEST_LOG("Failed to allocate resource");
		goto close_ctx;
	}

	for (log_sz = 0; log_sz < mlx5dv_hws_resource_max_block_size(resource);
	     log_sz++) {
		err = test_homogeneous_get(resource, log_sz);
		if (err) {
			TEST_LOG("test_homogeneous_get log_sz %d failed",
				 log_sz);
			goto free_resource;
		}
	}

	if (test_multi_thread_get_offset(resource, false)) {
		TEST_LOG("test_multi_thread_get_offset failed");
		goto free_resource;
	}

	if (test_multi_thread_get_offset(resource, true)) {
		TEST_LOG("test_multi_thread_get_offset exhaustive failed");
		goto free_resource;
	}

	if (test_benchmark(resource)) {
		TEST_LOG("test_benchmark failed");
		goto free_resource;
	}

	err = 0;

free_resource:
	mlx5dv_hws_resource_free(resource);
close_ctx:
	mlx5dv_hws_context_close(ctx);
dealloc_pd:
	ibv_dealloc_pd(pd);
out_err:
	return err;
}
