# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

from pyverbs.providers.mlx5.hws_matcher cimport HwsMatcher
from pyverbs.providers.mlx5.mlx5dv cimport Mlx5DevxObj
from pyverbs.providers.mlx5.hws_rule cimport HwsRule
cimport pyverbs.providers.mlx5.mlx5hws_enums as e
from pyverbs.pyverbs_error import PyverbsError
from pyverbs.base cimport close_weakrefs
from libc.stdlib cimport free, calloc
from libc.stdint cimport uint8_t
from libc.string cimport memcpy
import weakref
import socket
import struct

be64toh = lambda num: struct.unpack('Q'.encode(), struct.pack('!8s'.encode(), num))[0]


cdef copy_data_to_ptr(data, uint8_t **ptr):
    arr = bytearray(data)
    ptr[0] = <uint8_t *>calloc(1, len(arr))
    if ptr[0] == NULL:
        raise MemoryError('Failed to allocate memory')
    for i in range(len(arr)):
        (<uint8_t *>ptr[0])[i] = arr[i]

cdef hws.mlx5dv_hws_resource **alloc_hws_resource_list(resources, num_resources):
    assert(len(resources)==num_resources,
           'Number or resources should be equal to the length of resources list.')
    cdef hws.mlx5dv_hws_resource **resource_list_ptr = \
        <hws.mlx5dv_hws_resource **>calloc(sizeof(hws.mlx5dv_hws_resource*), num_resources)
    if not resource_list_ptr:
        raise MemoryError('Memory allocation failed.')
    for i in range(num_resources):
        p_resource = <HwsResource>resources[i]
        resource_list_ptr[i] = <hws.mlx5dv_hws_resource *>p_resource.resource
    return resource_list_ptr


cdef class HwsAction(PyverbsCM):
    def __init__(self, HwsContext ctx):
        super().__init__()
        ctx.add_ref(self)
        self.context = ctx
        self.rules = weakref.WeakSet()
        self.matchers = weakref.WeakSet()
        self.used_actions = weakref.WeakSet()

    cdef add_ref(self, obj):
        if isinstance(obj, HwsRule):
            self.rules.add(obj)
        elif isinstance(obj, HwsMatcher):
            self.matchers.add(obj)
        elif isinstance(obj, HwsAction):
            self.used_actions.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')

    cdef del_ref(self, obj):
        if isinstance(obj, HwsRule):
            self.rules.remove(obj)
        else:
            raise PyverbsError('Unrecognized object type')

    def enqueue_arg_write(self, HwsResource resource, qid, HwsActionEnqueueArgWriteAttr attr):
        """
        Posting argument write into the queue.
        :param resource: HwsResource object
        :param qid: Queue ID
        :param attr: Argument write attribute
        """
        rc = hws.mlx5dv_hws_action_enqueue_arg_write(self.action, resource.resource, qid,
                                                     &attr.attr)
        if rc:
            raise PyverbsError('Failed to enqueue arg write.', rc)


    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.action != NULL:
            self.logger.debug('Closing HwsAction.')
            close_weakrefs([self.rules, self.matchers, self.used_actions])
            rc = hws.mlx5dv_hws_action_destroy(self.action)
            if rc:
                raise PyverbsError('Failed to destroy HwsAction.', rc)
            self.action = NULL
            self.context = None


cdef class HwsActionDestDrop(HwsAction):
    def __init__(self, HwsContext ctx, flags):
        """
        Initializes a destination drop action.
        :param ctx: HwsContext context
        :param flags: Action flags enum
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_dest_drop(ctx.context, flags)
        if self.action == NULL:
            raise PyverbsError('HwsActionDestDrop creation failed.')


cdef class HwsActionDestTir(HwsAction):
    def __init__(self, HwsContext ctx, Mlx5DevxObj tir, flags, is_local=True):
        """
        Initializes a destination TIR action.
        :param ctx: HwsContext context.
        :param tir: Mlx5DevxObj tir object.
        :param flags: Action flags enum combination.
        :param is_local: Indicates whether the tir object was created on local gvmi or other gvmi.
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_dest_tir(ctx.context, tir.obj, is_local, flags)
        if self.action == NULL:
            raise PyverbsError('HwsActionDestTir creation failed.')
        self.devx_obj = tir
        tir.add_ref(self)


cdef class HwsActionDestMatcher(HwsAction):
    def __init__(self, HwsContext ctx, HwsMatcher matcher, flags,
                 type=e.MLX5DV_HWS_ACTION_DEST_MATCHER_BY_INDEX):
        """
        Initializes a destination matcher action.
        :param ctx: HwsContext context.
        :param matcher: Destination matcher.
        :param flags: Action flags enum combination.
        :param type: Specifies the enum for the Matcher destination type.
                     Currently, the system only supports insertion by index in the isolated matcher.
                     As a result, the type is set to e.MLX5DV_HWS_ACTION_DEST_MATCHER_BY_INDEX.
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_dest_matcher(ctx.context, matcher.matcher, type,
                                                                flags)
        if self.action == NULL:
            raise PyverbsError('HwsActionDestMatcher creation failed.')


cdef class HwsActionTag(HwsAction):
    def __init__(self, HwsContext ctx, flags):
        """
        Initializes action TAG.
        :param ctx: HwsContext context.
        :param flags: Action flags enum combination.
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_tag(ctx.context, flags)
        if self.action == NULL:
            raise PyverbsError('HwsActionTag creation failed.')


cdef class HwsActionDestRoot(HwsAction):
    def __init__(self, HwsContext ctx, priority, flags):
        """
        Initializes a destination root table action.
        :param ctx: HwsContext context.
        :param priority: Matcher's priority on the destination root table.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_dest_root(ctx.context, priority, flags)
        if self.action == NULL:
            raise PyverbsError('HwsActionDestRoot creation failed.')


cdef class HwsActionDestDefaultMiss(HwsAction):
    def __init__(self, HwsContext ctx, flags):
        """
        Initializes a default miss destination action.
        :param ctx: HwsContext context.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_dest_default_miss(ctx.context, flags)
        if self.action == NULL:
            raise PyverbsError('HwsActionDestDefaultMiss creation failed.')


cdef class HwsActionDestTable(HwsAction):
    def __init__(self, HwsContext ctx, HwsTable table, flags):
        """
        Initializes a destination table action.
        :param ctx: HwsContext context.
        :param table: Destination table.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_dest_table(ctx.context, table.table, flags)
        if self.action == NULL:
            raise PyverbsError('HwsActionDestTable creation failed.')
        self.table = table
        table.add_ref(self)


cdef class HwsActionDestVport(HwsAction):
    def __init__(self, HwsContext ctx, ib_port_num, flags):
        """
        Initializes a default miss destination action.
        :param ctx: HwsContext context.
        :param ib_port_num: Destination ib_port number.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_dest_vport(ctx.context, ib_port_num, flags)
        if self.action == NULL:
            raise PyverbsError('HwsActionDestVport creation failed.')


cdef class HwsActionCounter(HwsAction):
    def __init__(self, HwsContext ctx, counters, num_of_counters, flags):
        """
        Initializes a counter action.
        :param ctx: HwsContext context.
        :param counter: List of counters of HwsResource class.
        :param num_of_counters: number of counters.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        cdef hws.mlx5dv_hws_resource **ptr = alloc_hws_resource_list(counters, num_of_counters)
        self.action = hws.mlx5dv_hws_action_create_counter(ctx.context, ptr, num_of_counters, flags)
        free(ptr)
        if self.action == NULL:
            raise PyverbsError('HwsActionCounter creation failed.')
        self.counters = counters
        for counter in counters:
            (<HwsResource>counter).add_ref(self)


cdef class HwsActionPopVlan(HwsAction):
    def __init__(self, HwsContext ctx, flags):
        """
        Initializes action pop VLAN.
        :param ctx: HwsContext context.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_pop_vlan(ctx.context, flags)
        if self.action == NULL:
            raise PyverbsError('HwsActionPopVlan creation failed.')


cdef class HwsActionPushVlan(HwsAction):
    def __init__(self, HwsContext ctx, flags):
        """
        Initializes action push VLAN.
        :param ctx: HwsContext context.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_push_vlan(ctx.context, flags)
        if self.action == NULL:
            raise PyverbsError('HwsActionPushVlan creation failed.')


cdef class HwsActionModify(HwsAction):
      def __init__(self, HwsContext ctx, actions_pattern, flags, arguments=[], num_of_args=0):
        """
        Initializes a modify action.
        :param ctx: HwsContext context.
        :param actions_pattern: List of PRM modify actions.
        :param flags: Action flags enum.
        :param arguments: List of arguments of HwsResource class corresponding to the pattern size.
                          Should be empty list for root.
        :param num_of_args: Number of arguments
        """
        super().__init__(ctx)
        cdef hws.mlx5dv_hws_action_modify_header pattern_ptr
        cdef hws.mlx5dv_hws_resource **ptr = alloc_hws_resource_list(arguments, num_of_args)
        action_size_in_bytes = 8
        pattern_ptr.sz = len(actions_pattern) * action_size_in_bytes
        pattern_ptr.modify_actions = <unsigned long long *>calloc(1, pattern_ptr.sz)
        if pattern_ptr.modify_actions == NULL:
            free(pattern_ptr.modify_actions)
            free(ptr)
            raise MemoryError('Memory allocation failed.')
        for i in range(len(actions_pattern)):
            (<unsigned long long *>pattern_ptr.modify_actions)[i] = \
                be64toh(bytes(actions_pattern[i]))
        self.action = hws.mlx5dv_hws_action_create_modify_header(ctx.context,
                                                                 ptr if arguments else NULL,
                                                                 num_of_args,
                                                                 &pattern_ptr,
                                                                 0,
                                                                 flags)
        free(ptr)
        free(pattern_ptr.modify_actions)
        if self.action == NULL:
            raise PyverbsError('HwsActionModify creation failed.')
        self.arguments = arguments
        for argument in arguments:
            (<HwsResource>argument).add_ref(self)


cdef class HwsActionReformatTnlL2ToL2(HwsAction):
    def __init__(self, HwsContext ctx, flags):
        """
        Initializes action reformat tunnel L2 to L2 (decap).
        :param ctx: HwsContext context.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_reformat_tnl_l2_to_l2(ctx.context, flags)
        if self.action == NULL:
            raise PyverbsError('HwsActionReformatTnlL2ToL2 creation failed.')



cdef class HwsActionReformatL2ToTnlL2(HwsAction):
     def __init__(self, HwsContext ctx, data, flags, resources=[], num_of_resources=0):
        """
        Initializes action reformat L2 to tunnel L2 (encap).
        :param ctx: HwsContext context.
        :param data: Header data information.
        :param flags: Action flags enum.
        :param resources: List of argument resources corresponding to the outer header size.
                          Resource should not be supplied for root actions.
        :param num_of_resources: Number of resources.
        """
        super().__init__(ctx)
        cdef hws.mlx5dv_hws_action_hdr_data data_ptr
        cdef hws.mlx5dv_hws_resource **ptr = alloc_hws_resource_list(resources, num_of_resources)
        data_ptr.hdr_data = <unsigned char *>calloc(1, len(data))
        if data_ptr.hdr_data == NULL:
            free(ptr)
            raise MemoryError('Memory allocation failed.')
        data_ptr.sz = len(data)
        arr = bytearray(data)
        for i in range(len(data)):
            (<uint8_t *>(data_ptr.hdr_data))[i] = arr[i]
        self.action = hws.mlx5dv_hws_action_create_reformat_l2_to_tnl_l2(ctx.context,
                                                                         ptr if resources else NULL,
                                                                         num_of_resources,
                                                                         &data_ptr, flags)
        free(ptr)
        free(data_ptr.hdr_data)
        if self.action == NULL:
            raise PyverbsError('HwsActionReformatL2ToTnlL2 creation failed.')
        self.arguments = resources
        for resource in resources:
            (<HwsResource>resource).add_ref(self)


cdef class HwsActionReformatTnlL3ToL2(HwsAction):
    def __init__(self, HwsContext ctx, data, flags, resources=[], num_of_resources=0):
        """
        Initializes action reformat tunnel L3 to L2 (decap).
        :param ctx: HwsContext context.
        :param data: Header data information.
        :param flags: Action flags enum.
        :param resource: list of argument resources corresponding to the L2 header size.
                         Resource should not be supplied for root actions.
        :param num_of_resources: Number of resources.
        """
        super().__init__(ctx)
        cdef hws.mlx5dv_hws_action_hdr_data data_ptr
        cdef hws.mlx5dv_hws_resource **ptr = alloc_hws_resource_list(resources, num_of_resources)
        data_ptr.hdr_data = <unsigned char *>calloc(1, len(data))
        if data_ptr.hdr_data == NULL:
            free(ptr)
            raise MemoryError('Memory allocation failed.')
        data_ptr.sz = len(data)
        arr = bytearray(data)
        for i in range(len(data)):
            (<uint8_t *>(data_ptr.hdr_data))[i] = arr[i]
        self.action = hws.mlx5dv_hws_action_create_reformat_tnl_l3_to_l2(ctx.context,
                                                                         ptr if resources else NULL,
                                                                         num_of_resources,
                                                                         &data_ptr, flags)
        free(ptr)
        free(data_ptr.hdr_data)
        if self.action == NULL:
            raise PyverbsError('HwsActionReformatTnlL3ToL2 creation failed.')
        self.arguments = resources
        for resource in resources:
            (<HwsResource>resource).add_ref(self)


cdef class HwsActionReformatL2ToTnlL3(HwsAction):
    def __init__(self, HwsContext ctx, data, flags, resources=[], num_of_resources=0):
        """
        Initializes action reformat L2 to tunnel L3 (encap).
        :param ctx: HwsContext context.
        :param data: Header data information.
        :param flags: Action flags enum.
        :param resources: List of argument resources corresponding to the outer size.
                         Resource should not be supplied for root actions.
        :param num_of_resources: Number of resources.
        """
        super().__init__(ctx)
        cdef hws.mlx5dv_hws_action_hdr_data data_ptr
        cdef hws.mlx5dv_hws_resource **ptr = alloc_hws_resource_list(resources, num_of_resources)
        data_ptr.hdr_data = <unsigned char *>calloc(1, len(data))
        if data_ptr.hdr_data == NULL:
            free(ptr)
            raise MemoryError('Memory allocation failed.')
        data_ptr.sz = len(data)
        arr = bytearray(data)
        for i in range(len(data)):
            (<uint8_t *>(data_ptr.hdr_data))[i] = arr[i]
        self.action = hws.mlx5dv_hws_action_create_reformat_l2_to_tnl_l3(ctx.context,
                                                                         ptr if resources else NULL,
                                                                         num_of_resources,
                                                                         &data_ptr, flags)
        free(ptr)
        free(data_ptr.hdr_data)
        if self.action == NULL:
            raise PyverbsError('HwsActionReformatL2ToTnlL3 creation failed.')
        self.arguments = resources
        for resource in resources:
            (<HwsResource>resource).add_ref(self)


cdef class HwsActionInsertHeaderAttr(PyverbsCM):
    def __init__(self, data, anchor, offset=0, encap=False, push_esp=False, skip_reparse=False):
        """
        Initializes action Insert Header struct (attributes).
        :param anchor: PRM start anchor to which header will be inserted.
        :param offset: Header insertion offset from anchor in bytes.
        :param encap: Indicates if header insertion adds encapsulation header to the packet.
        :param push_esp: It must be set when adding ESP header.
        :param skip_reparse: Skip re-evaluation of packet headers after modifications.
        """
        super().__init__()
        self.hdr.anchor = anchor
        self.hdr.offset = offset
        self.hdr.encap = encap
        self.hdr.push_esp = push_esp
        self.hdr.skip_reparse = skip_reparse
        self.hdr.hdr_data.sz = len(data)
        copy_data_to_ptr(data, &self.hdr.hdr_data.hdr_data)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.hdr.hdr_data.hdr_data != NULL:
            free(self.hdr.hdr_data.hdr_data)
            self.hdr.hdr_data.hdr_data = NULL


cdef class HwsActionInsertHeader(HwsAction):
    def __init__(self, HwsContext ctx, resources, num_of_resources,
                 HwsActionInsertHeaderAttr hdr, flags):
        """
        Initializes action Insert Header.
        :param ctx: HwsContext context.
        :param resources: List of argument resources corresponding to the input header size.
                         Resource should not be supplied for root actions.
        :param num_of_resources: Number of resources.
        :param hdr: Header insert attributes.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        cdef hws.mlx5dv_hws_resource **ptr = alloc_hws_resource_list(resources, num_of_resources)
        self.action = hws.mlx5dv_hws_action_create_insert_header(ctx.context, ptr, num_of_resources,
                                                                 &(hdr.hdr), flags)
        free(ptr)
        if self.action == NULL:
            raise PyverbsError('HwsActionInsertHeader creation failed.')
        self.arguments = resources
        for resource in resources:
            (<HwsResource>resource).add_ref(self)


cdef class HwsActionRemoveHeaderAttr(PyverbsObject):
    def __init__(self, type_rm, start_anchor, end_anchor=None, decap=False, size=0):
        """
        Initializes action Remove Header struct (attributes).
        :param type_rm: Header remove type, header or offset.
        :param start_anchor: Start anchor from which header will be removed.
        :param end_anchor: End anchor until the header will be removed.
        :param decap: True if remove header decapsulates the packet.
        :param size: Size in words to remove from header.
        """
        super().__init__()
        self.attr.type = type_rm
        if type_rm == e.MLX5DV_HWS_ACTION_REMOVE_HEADER_TYPE_BY_OFFSET:
            self.attr.by_offset.start_anchor = start_anchor
            self.attr.by_offset.size = size
        else:
            self.attr.by_header.start_anchor = start_anchor
            self.attr.by_header.end_anchor = end_anchor
            self.attr.by_header.decap = decap


cdef class HwsActionRemoveHeader(HwsAction):
    def __init__(self, HwsContext ctx, HwsActionRemoveHeaderAttr attr, flags):
        """
        Initializes action Remove Header.
        :param ctx: HwsContext context.
        :param attr: Specifies the remove header type, start/end anchor and size.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        self.action = hws.mlx5dv_hws_action_create_remove_header(ctx.context, &(attr.attr), flags)

        if self.action == NULL:
            raise PyverbsError('HwsActionRemoveHeader creation failed.')


cdef class HwsActionAsoMeter(HwsAction):
    def __init__(self, HwsContext ctx, aso_meters, num_of_resources, return_reg_id, flags):
        """
        Initializes ASO meter action.
        :param ctx: HwsContext context.
        :param aso_meters: List of ASO meters of HwsResource class.
        :param num_of_resources: Number of resources.
        :param return_reg_id: The destination array. Each contains a destination action
 *                  and can have additional actions to be executed prior to it.t.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        cdef hws.mlx5dv_hws_resource **ptr = alloc_hws_resource_list(aso_meters, num_of_resources)
        self.action = hws.mlx5dv_hws_action_create_aso_meter(ctx.context, ptr, num_of_resources,
                                                             return_reg_id, flags)
        free(ptr)
        if self.action == NULL:
            raise PyverbsError('HwsActionAsoMeter creation failed.')
        self.aso_meters = aso_meters
        for aso_meter in aso_meters:
            (<HwsResource>aso_meter).add_ref(self)


cdef class HwsActionAsoCT(HwsAction):
    def __init__(self, HwsContext ctx, resources, num_of_resources, return_reg_id, flags):
        """
        Initializes ASO connection tracking action.
        :param ctx: HwsContext context.
        :param resources: List of ASO objects of HwsResource class.
        :param num_of_resources: Number of resources.
        :param return_reg_id: The destination array. Each contains a destination action
 *                  and can have additional actions to be executed prior to it.t.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        cdef hws.mlx5dv_hws_resource **ptr = alloc_hws_resource_list(resources, num_of_resources)
        self.action = hws.mlx5dv_hws_action_create_aso_ct(ctx.context, ptr, num_of_resources,
                                                          return_reg_id, flags)
        free(ptr)
        if self.action == NULL:
            raise PyverbsError('HwsActionAsoCT creation failed.')
        self.resources = resources
        for resource in resources:
            (<HwsResource>resource).add_ref(self)


cdef class HwsActionDestArrayAttr(PyverbsCM):
    def __init__(self, HwsAction dest_action, reformat_type=None, reformat_data=None):
        """
        Initializes a HwsActionDestArrayAttr object representing
        mlx5dv_hws_action_dest_attr struct.
        :param dest_action: Destination action to forward the packet
        :param reformat_type: Optional reformat type
        :param reformat_data: Optional reformat data
        """
        super().__init__()
        self.attr.dest_action = dest_action.action
        self.dest_action = dest_action
        if reformat_type:
            self.attr.reformat.type = reformat_type
        if reformat_data:
            self.attr.reformat.hdr_data.sz = len(reformat_data)
            copy_data_to_ptr(reformat_data,  &self.attr.reformat.hdr_data.hdr_data)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        self.logger.debug('Closing HwsActionDestArrayAttr.')
        if self.attr.reformat.hdr_data.hdr_data:
            free(self.attr.reformat.hdr_data.hdr_data)
            self.attr.reformat.hdr_data.hdr_data = NULL
        self.dest_action = None


cdef class HwsActionDestArray(HwsAction):
    def __init__(self, HwsContext ctx, num_dest, list dests, action_flag):
        """
        Initializes destination array action.
        :param ctx: HwsContext context.
        :param num_dest: The number of destination in the destination array.
        :param dests: List of HwsActionDestArrayAttr. Each contains a destination action
                      and can have additional reformat actions to be executed prior to it.
        :param flags: Action flags enum.
        """
        super().__init__(ctx)
        assert(num_dest == len(dests),
               'Number of destination must be equla to destinations list lenght')
        self.actions = []
        cdef hws.mlx5dv_hws_action_dest_attr *dests_list_ptr = \
            <hws.mlx5dv_hws_action_dest_attr *>calloc(num_dest, sizeof(hws.mlx5dv_hws_action_dest_attr))
        if dests_list_ptr == NULL:
            raise MemoryError('Memory allocation failed.')
        for i in range(num_dest):
            dest = <HwsActionDestArrayAttr>(dests[i])
            memcpy(&dests_list_ptr[i], &dest.attr, sizeof(hws.mlx5dv_hws_action_dest_attr))
        self.action = hws.mlx5dv_hws_action_create_dest_array(ctx.context, num_dest,
                                                              dests_list_ptr, action_flag)
        free(dests_list_ptr)
        if self.action == NULL:
            raise PyverbsError('HwsActionDestArray creation failed.')
        for d in dests:
            ((<HwsActionDestArrayAttr>d).dest_action).add_ref(self)
            self.actions.append((<HwsActionDestArrayAttr>d).dest_action)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.action != NULL:
            self.logger.debug('Closing HwsActionDestArray.')
            self.actions = []
        super(HwsActionDestArray, self).close()


cdef class HwsActionEnqueueArgWriteAttr(PyverbsCM):
    def __init__(self, comp_mask, user_data, resource_offset, data_size, arg_data, burst):
        """
        Initializes a HwsActionEnqueueArgWriteAttr object representing
        mlx5dv_hws_action_enqueue_arg_write_attr struct.
        :param comp_mask: Compatability mask
        :param user_data: Returned user data in the completion entry
        :param resource_offset: Resource offset
        :param data_size: Size in bytes to write
        :param arg_data: Data to write (can be list for modify actions)
        :param burst: Batching multiple operations
        """
        super().__init__()
        self.attr.comp_mask = comp_mask
        self.attr.user_data =  <void *>user_data if user_data else NULL
        self.attr.resource_offset = resource_offset
        self.attr.data_size = data_size
        self.attr.burst = burst
        if isinstance(arg_data, list):
            arr = bytes()
            for data in arg_data:
                arr += bytes(data)
        else:
            arr = bytes(arg_data)
        copy_data_to_ptr(arr, &self.attr.arg_data)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.attr.arg_data != NULL:
            free(self.attr.arg_data)
            self.attr.arg_data = NULL


cdef class HwsActionTemplateAttr(PyverbsObject):
    def __init__(self, comp_mask=0):
        """
        Initializes a HwsActionTemplateAttr object representing mlx5dv_hws_at_attr
        struct.
        """
        super().__init__()
        self.attr.comp_mask = comp_mask


cdef class HwsActionTemplate(PyverbsCM):
    def __init__(self, actions, num_of_actions, HwsActionTemplateAttr attr=None):
        """
        Initializes a HwsActionTemplate object representing mlx5dv_hws_at struct.
        :param actions: A list of actions based on the order of actions which will be provided
                        with rule actions for rule creation.
        :param num_of_actions: Number of actions.
        :param attr: Action template attributes.
        """
        super().__init__()
        cdef hws.mlx5dv_hws_action **mlx5dv_hws_actions = NULL
        mlx5dv_hws_actions = <hws.mlx5dv_hws_action **>calloc(num_of_actions,
                                                        sizeof(hws.mlx5dv_hws_action*))
        if mlx5dv_hws_actions == NULL:
            raise MemoryError('Memory allocation failed.')
        for i in range(num_of_actions):
            action = <HwsAction>(actions[i])
            mlx5dv_hws_actions[i] = <hws.mlx5dv_hws_action*>(action.action)
        if not attr:
            attr = HwsActionTemplateAttr()
        self.action_template = hws.mlx5dv_hws_action_template_create(mlx5dv_hws_actions,
                                                                     num_of_actions, &attr.attr)
        free(mlx5dv_hws_actions)
        if self.action_template == NULL:
            raise PyverbsError('HwsActionTemplate creation failed.')
        self.actions = actions[:]

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.action_template != NULL:
            self.logger.debug('Closing HwsActionTemplate.')
            rc = hws.mlx5dv_hws_action_template_destroy(self.action_template)
            if rc:
                raise PyverbsError('Failed to destroy HwsActionTemplate.', rc)
            self.action_template = NULL


cdef class HwsActionData(PyverbsCM):
    """
    Class HwsActionData representing mlx5dv_hws_action_data struct.
    Action data to be used for rule creation.
    """
    def __init__(self, **kwargs):
        """
        Initializes the HwsActionData object representing mlx5dr_rule_action struct.
        :param kwargs:
            * *value*
                Tag value.
            * *vlan_hdr*
                Header for push VLAN.
            * *resource_offset*
                Resource offset in the resource bulk.
            * *resource_index*
                Resource index in the resource list.
            * *modify_data*
                List of 8 Bytes Data for modify header action.
            * *init_color*
                Init color for ASO meter action.
            * *aso_wqe*
                ASO WQE for ASO related action to configure the ASO object.
            * *direction*
                Direction for ASO CT action data (initiator or responder).
            * *index*
                index for destination matcher action.
        """
        super().__init__()
        cdef void *data_buf
        tag = socket.htonl(kwargs.get('value', 0))
        resource_offset = kwargs.get('resource_offset', 0)
        resource_index = kwargs.get('resource_index', 0)
        vlan_hdr = socket.htonl(kwargs.get('vlan_hdr', 0))
        mdf_action_data = kwargs.get('modify_data')
        reformat_l2_to_tnl_l2_action_data = kwargs.get('reformat_l2_to_tnl_l2_data')
        reformat_l2_to_tnl_l3_action_data = kwargs.get('reformat_l2_to_tnl_l3_data')
        reformat_tnl_l3_to_l2_action_data = kwargs.get('reformat_tnl_l3_to_l2_data')
        insert_header_action_data = kwargs.get('insert_header')
        init_color = kwargs.get('init_color')
        aso_wqe = kwargs.get('aso_wqe')
        direction = kwargs.get('direction')
        index = kwargs.get('index')
        if tag:
            self.action_data.tag.value = tag
        elif vlan_hdr:
            self.action_data.push_vlan.vlan_hdr = vlan_hdr
        elif mdf_action_data:
            self.action_data.modify_header.resource_offset = resource_offset
            self.action_data.modify_header.resource_idx = resource_index
            action_size_in_bytes = 8
            self.ptr_to_free = calloc(1, len(mdf_action_data) * action_size_in_bytes)
            if self.ptr_to_free == NULL:
                raise MemoryError('Failed to allocate memory')
            for i in range(len(mdf_action_data)):
                (<unsigned long long*>(self.ptr_to_free))[i] = \
                    <unsigned long long>(be64toh(bytes(mdf_action_data[i])))
            self.action_data.modify_header.actions_data = <unsigned long long *>(self.ptr_to_free)
        elif reformat_l2_to_tnl_l2_action_data:
            self.action_data.reformat_l2_to_tnl_l2.resource_offset = resource_offset
            self.action_data.reformat_l2_to_tnl_l2.resource_idx = resource_index
            copy_data_to_ptr(reformat_l2_to_tnl_l2_action_data,
                             &self.action_data.reformat_l2_to_tnl_l2.hdr_data)
            self.ptr_to_free = self.action_data.reformat_l2_to_tnl_l2.hdr_data
        elif reformat_tnl_l3_to_l2_action_data:
            self.action_data.reformat_tnl_l3_to_l2.resource_offset = resource_offset
            self.action_data.reformat_tnl_l3_to_l2.resource_idx = resource_index
            copy_data_to_ptr(reformat_tnl_l3_to_l2_action_data,
                             &self.action_data.reformat_tnl_l3_to_l2.hdr_data)
            self.ptr_to_free = self.action_data.reformat_tnl_l3_to_l2.hdr_data
        elif reformat_l2_to_tnl_l3_action_data:
            self.action_data.reformat_l2_to_tnl_l3.resource_offset = resource_offset
            self.action_data.reformat_l2_to_tnl_l3.resource_idx = resource_index
            copy_data_to_ptr(reformat_l2_to_tnl_l3_action_data,
                             &self.action_data.reformat_l2_to_tnl_l3.hdr_data)
            self.ptr_to_free = self.action_data.reformat_l2_to_tnl_l3.hdr_data
        elif insert_header_action_data:
            self.action_data.insert_header.resource_offset = resource_offset
            self.action_data.insert_header.resource_idx = resource_index
            copy_data_to_ptr(insert_header_action_data,
                             &self.action_data.insert_header.hdr_data)
            self.ptr_to_free = self.action_data.insert_header.hdr_data
        elif init_color:
            self.action_data.aso_meter.resource_offset = resource_offset
            self.action_data.aso_meter.init_color = init_color
            self.action_data.aso_meter.resource_idx = resource_index
        elif direction:
            self.action_data.aso_ct.resource_offset = resource_offset
            self.action_data.aso_ct.direction = direction
            self.action_data.aso_ct.resource_idx = resource_index
        elif resource_offset:
            self.action_data.counter.resource_offset = resource_offset
        elif resource_index:
            self.action_data.counter.resource_idx = resource_index
        elif aso_wqe:
            raise NotImplementedError('ASO WQE is not supported yet')
        elif index:
            self.action_data.dest_matcher.index = index

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.ptr_to_free:
            free(self.ptr_to_free)
            self.ptr_to_free = NULL
