# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

from pyverbs.providers.mlx5.hws_parser cimport HwsParserGraph
from pyverbs.providers.mlx5.hws_resource cimport HwsResource
from pyverbs.providers.mlx5.hws_actions cimport HwsAction
from pyverbs.providers.mlx5.hws_table cimport HwsTable
cimport pyverbs.providers.mlx5.mlx5hws_enums as e
cimport pyverbs.providers.mlx5.libmlx5hws as hws
from pyverbs.pyverbs_error import PyverbsError
from pyverbs.base cimport close_weakrefs
from libc.stdlib cimport free, calloc
from libc.stdint cimport uintptr_t
cimport pyverbs.libibverbs as v
cimport libc.stdio as s
import weakref
import ctypes
import time

POLLING_TIMEOUT = 5


cdef class HwsContextAttr(PyverbsObject):
    def __init__(self, pd, comp_mask=0, queues=1, queue_size=1, shared_ctx=None):
        """
        Initializes a HwsContextAttr object representing mlx5dv_hws_context_attr
        struct.
        :param pd: PD to use
        :param comp_mask: compatibility mask
        :param queues: Number of queues
        :param queue_size: Size of each queue
        :param shared_ctx: Shared Mlx5dvContext to use (if provided)
        """
        super().__init__()
        self.attr.pd = <v.ibv_pd *>pd.pd
        self.attr.comp_mask = comp_mask
        self.attr.queues = queues
        self.attr.queue_size = queue_size
        self.pd = pd
        if shared_ctx:
            self.attr.shared_ibv_ctx = <v.ibv_context *> shared_ctx.context
            self.shared_context = shared_ctx


cdef class HwsContext(PyverbsCM):
    def __init__(self, ibv_context, HwsContextAttr attr):
        """
        Initializes a HwsContext object representing mlx5dv_hws_context struct.
        :param ibv_context: Pyverbs object that represent ibv context.
        :param attr: Attributes for creating HwsContext.
        """
        super().__init__()
        cdef v.ibv_context *ctx_ptr = <v.ibv_context *> ibv_context.context
        self.context = hws.mlx5dv_hws_context_open(ctx_ptr, &attr.attr)
        if self.context == NULL:
            raise PyverbsError('Failed creating HwsContext')
        self.ibv_context = ibv_context
        self.hws_tables = weakref.WeakSet()
        self.hws_actions = weakref.WeakSet()
        self.hws_resources = weakref.WeakSet()
        self.hws_parser_graph = weakref.WeakSet()

    cdef add_ref(self, obj):
        if isinstance(obj, HwsTable):
            self.hws_tables.add(obj)
        elif isinstance(obj, HwsAction):
            self.hws_actions.add(obj)
        elif isinstance(obj, HwsResource):
            self.hws_resources.add(obj)
        elif isinstance(obj, HwsParserGraph):
            self.hws_parser_graph.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')

    def dump(self, filepath):
        """
        Dumps the debug info of the context into a file.
        :param filepath: Path to the file.
        """
        cdef s.FILE *fp
        fp = s.fopen(filepath.encode('utf-8'), 'w+')
        if fp == NULL:
            raise PyverbsError('Opening dump file failed.')
        rc = hws.mlx5dv_hws_debug_dump(self.context, fp)
        if s.fclose(fp) != 0:
            raise PyverbsError('Closing dump file failed.')
        if rc != 0:
            raise PyverbsError('Context dump failed.', rc)

    def poll_send_queue(self, queue_id, res_nb):
        """
        Poll send queue for completions.
        :param queue_id: The queue ID to poll.
        :param res_nb: Number of expected completeions.
        :return: List of HwsQueueOpResult.
        """
        cdef hws.mlx5dv_hws_queue_op_result* res_list_ptr = \
            <hws.mlx5dv_hws_queue_op_result *>calloc(res_nb, sizeof(hws.mlx5dv_hws_queue_op_result))
        if res_list_ptr == NULL:
            raise MemoryError('Failed allocating memory.')
        results = []
        res = hws.mlx5dv_hws_queue_poll(self.context, queue_id, res_list_ptr, res_nb)
        if res > 0:
            for r in range(res):
                if res_list_ptr[r].status == e.MLX5DV_HWS_QUEUE_OP_ERROR:
                    self.logger.warning(f'ERROR completion returned from queue {queue_id}.')
                results.append(HwsQueueOpResult(res_list_ptr[r].status,
                                                <uintptr_t><void*>res_list_ptr[r].user_data))
        free(res_list_ptr)
        if res < 0:
            raise PyverbsError('Polling for completion failed.', res)
        return results

    def check_op_completion(self, queue_id):
        """
        Poll provided queue ID for completion regarding operation on a queue.
        :param: queue_id: Queue ID to poll.
        :return: List of HwsQueueOpResults.
        """
        start_poll_t = time.perf_counter()
        res = []
        while not res and (time.perf_counter() - start_poll_t) < POLLING_TIMEOUT:
            res = self.poll_send_queue(queue_id, 1)
        if not res:
            raise PyverbsError(f'Got timeout on polling queue id {queue_id}.')
        if <HwsQueueOpResult> (res[0]).status != e.MLX5DV_HWS_QUEUE_OP_SUCCESS:
            raise PyverbsError(f'ERROR completion returned from queue ID: {queue_id} '
                               f'with status: {res[0]).status}.')
        return res

    def queue_execute_op(self, queue_id, queue_op):
        """
        Execute an operation on the queue.
        :param queue_id: The id of the queue to execute the operation.
        :param queue_op: Queue operation to perform (enum mlx5dv_hws_queue_op).
        """
        res = hws.mlx5dv_hws_queue_execute_op(self.context, queue_id, queue_op)
        if res:
            raise PyverbsError('Queue operation execution failed.', res)

    def queue_pending(self, queue_id):
        """
        Query number of queued operations.
        :param queue_id: The id of the queue to query.
        """
        res = hws.mlx5dv_hws_queue_pending(self.context, queue_id)
        if res < 0:
            raise PyverbsError(f'Failed to query queue {queue_id} for num of pending operations.')
        return res

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.context != NULL:
            self.logger.debug('Closing HwsContext.')
            close_weakrefs([self.hws_actions, self.hws_tables, self.hws_resources,
                            self.hws_parser_graph])
            rc = hws.mlx5dv_hws_context_close(<hws.mlx5dv_hws_context *>(self.context))
            if rc:
                raise PyverbsError('Failed to destroy HwsContext.', rc)
            self.context = NULL


cdef class HwsQueueOpResult(PyverbsObject):
    def __init__(self, status, user_data=None):
        """
        Initializes a HwsQueueOpResult object representing mlx5dv_hws_queue_op_result struct.
        :param: status: Returns the status of the operation that this completion signals.
        :param: user_data: The user data that will be returned on the completion events.
        """
        super().__init__()
        self.result.status = status
        self.result.user_data = <void *>user_data if user_data else NULL

    @property
    def status(self):
        return self.result.status

    @property
    def user_data(self):
        return ctypes.cast(<object>self.result.user_data, ctypes.py_object).value
