# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.


from pyverbs.providers.mlx5.hws_context cimport HwsContext
from pyverbs.providers.mlx5.hws_item_data cimport HwsData
cimport pyverbs.providers.mlx5.mlx5hws_enums as e
cimport pyverbs.providers.mlx5.libmlx5hws as hws
from pyverbs.pyverbs_error import PyverbsError
from libc.stdlib cimport free, calloc
from libc.stdint cimport uint32_t
from libc.string cimport memcpy


cdef class HwsField(PyverbsObject):
    def __init__(self, fname, header, op, match_type, field_offset=None, user_conv_data=None,
                 user_conv_func=None, fname_b=None, header_b=None):
        """
        Initializes a HwsField object representing mlx5dv_hws_field struct.
        :param fname: Field name.
        :param header: Field location on the inner/outer header.
        :param op: Comparison operation to execute between field and value (or field_b).
        :param match_type: Matching type (match to value or match to field).
        :param field_offset: The offset of the field within the item data
                             or None if matching type to field.
        :param user_conv_data: Optional user value to convert.
        :param user_conv_func: Optional user value convert function.
        :param fname_b: Field name to use when matching is field-to-field.
        :param header_b: Field header type to use when matching is field-to-field.
        """
        super().__init__()
        self.field.fname = fname
        self.field.header = header
        self.field.op = op
        self.field.type = match_type
        cdef void* temp
        if match_type == e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE:
            self.field.value.bit_off = field_offset
            if user_conv_data and user_conv_func:
                self.field.value.user_conv_data = <void*>user_conv_data
                self.field.value.user_conv_func = NULL #TODO: Add support for user_conv_func.
        else:
            self.field.field.fname_b = fname_b
            self.field.field.header_b = header_b


cdef class HwsItem(PyverbsCM):
    def __init__(self, fields, num_of_fields, HwsData data=None):
        """
        Initializes a HwsItem object representing mlx5dv_hws_item struct.
        :param fields: List of HwsFields.
        :param num_of_fields: Number of fields.
        :param data: The data to match on or None in case of field-to-field matching.
        """
        super().__init__()
        cdef hws.mlx5dv_hws_field field_c_i
        cdef hws.mlx5dv_hws_item_data data_c
        cdef hws.mlx5dv_hws_field *fields_c = \
                        <hws.mlx5dv_hws_field *>calloc(num_of_fields, sizeof(hws.mlx5dv_hws_field))
        if fields_c == NULL:
            raise PyverbsError('Failed allocating memory.')
        for i in range(num_of_fields):
            field_py_i = <HwsField>(fields[i])
            field_c_i = field_py_i.field
            memcpy(&fields_c[i], &field_c_i, sizeof(hws.mlx5dv_hws_field))

        self.item.fields = fields_c
        self.item.num_of_fields = num_of_fields
        if data:
            data_c = <hws.mlx5dv_hws_item_data>(data.data)
            self.item.data.match = data_c.match
            self.item.data.mask = data_c.mask
            self.item.data.last = data_c.last
        self.fields = fields
        self.data = data

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.logger:
            self.logger.debug('Closing HwsItem.')
        if self.item.fields != NULL:
            free(self.item.fields)
            self.item.fields = NULL
        if self.data:
            self.data.close()


cdef class HwsMatchTemplateAttr(PyverbsObject):
    def __init__(self, comp_mask=0):
        """
        Initializes a HwsMatchTemplateAttr object representing mlx5dv_hws_mt_attr struct.
        :param comp_mask: Compatability mask.
        """
        super().__init__()
        self.attr.comp_mask = comp_mask


cdef class HwsMatchTemplate(PyverbsCM):
    def __init__(self, HwsContext context, items, num_of_items,
                 HwsMatchTemplateAttr attr=None):
        """
        Initializes a HwsMatchTemplate object representing mlx5dv_hws_mt struct.
        :param context: HwsContext context.
        :param items: List of HwsItems.
        :param num_of_items: Number of items.
        :param attr: Attributes for creating HwsMatchTemplate.
        """
        super().__init__()
        cdef hws.mlx5dv_hws_item *c_items = \
                        <hws.mlx5dv_hws_item *>calloc(num_of_items, sizeof(hws.mlx5dv_hws_item))
        if c_items == NULL:
            raise PyverbsError('Failed allocating memory.')

        for i in range(num_of_items):
            item_py_i = <HwsItem>(items[i])
            item_c_i_ptr = &(item_py_i.item)
            memcpy(&c_items[i], item_c_i_ptr, sizeof(hws.mlx5dv_hws_item))

        if attr == None:
            attr = HwsMatchTemplateAttr()

        self.match_template = \
                        hws.mlx5dv_hws_match_template_create(context.context,
                                                             c_items if num_of_items else NULL,
                                                             num_of_items,
                                                             <hws.mlx5dv_hws_mt_attr *>&(attr.attr))
        free(c_items)
        if self.match_template == NULL:
            raise PyverbsError('Failed creating HwsMatchTemplate.')
        self.hws_context = context

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.match_template == NULL:
            return
        if self.logger:
            self.logger.debug('Closing HwsMatchTemplate.')
        rc = hws.mlx5dv_hws_match_template_destroy(<hws.mlx5dv_hws_mt *>(self.match_template))
        if rc:
            raise PyverbsError('Failed to destroy HwsMatchTemplate.', rc)
        self.match_template = NULL
        self.hws_context = None
