# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

from pyverbs.providers.mlx5.hws_actions cimport HwsActionTemplate, HwsAction, HwsActionData
from pyverbs.providers.mlx5.hws_match_template cimport HwsMatchTemplate
from pyverbs.providers.mlx5.hws_item_data cimport HwsData
from pyverbs.providers.mlx5.hws_rule cimport HwsRule
cimport pyverbs.providers.mlx5.mlx5hws_enums as e
cimport pyverbs.providers.mlx5.libmlx5hws as hws
from pyverbs.pyverbs_error import PyverbsError
from pyverbs.base cimport close_weakrefs
from libc.stdlib cimport free, calloc
from libc.stdint cimport uint32_t
from libc.string cimport memcpy
import weakref


cdef class HwsMatcherAttr(PyverbsObject):
    def __init__(self, priority, insert_mode=e.MLX5DV_HWS_MATCHER_INSERT_BY_HASH,
                 distribute_mode=e.MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH,
                 resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_HTABLE, row_log=4,
                 col_log=4, rule_log=4, rule_idx_opt=False, resizable=False, isolated=False,
                 always_hit=False, max_num_of_at_attach=0, comp_mask=0):
        """
        Initialize a HwsMatcherAttr object representing hws_matcher_attr C struct.
        :param priority: Processing priority inside table
        :param insert_mode: Define the insertion mode for this matcher
        :param distribute_mode: Define the distribution mode for this matcher
        :param resource_mode: Table resourse mode:
                    MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE - Allocate resources based on
                    number of rules with minimal failure probability.
                    MLX5DV_HWS_MATCHER_RESOURCE_MODE_HTABLE - Allocate fixed size hash
                    table based on given column and rows.
        :param row_log: Hint for the log number of rows to be created
        :param col_log: Hint for the log number of colunms to be created
        :param rule_log: Hint for the log number of rules to be created
        :param rule_idx_opt: Determine if to use rule index optimization
        :param resizable: If True matcher can be used as src or dst for matcher resizing
        :param isolated: If True matcher is not part of the matchers chain of parent table
        :param always_hit: If true ignore match items and perform always hit
        :param max_num_of_at_attach: Max number of additional allowed AT attached to matcher
        :param comp_mask: Compatibility mask
        """
        super().__init__()
        self.attr.comp_mask = comp_mask
        self.attr.priority = priority
        self.attr.insert_mode = insert_mode
        self.attr.distribute_mode = distribute_mode
        self.attr.resource_mode = resource_mode
        if resource_mode == e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_HTABLE:
            self.attr.table.sz_row_log = row_log
            self.attr.table.sz_col_log = col_log
        else:
            self.attr.rule.num_log = rule_log
        self.attr.optimize_using_rule_idx = rule_idx_opt
        self.attr.resizable = resizable
        self.attr.isolated = isolated
        self.attr.always_hit = always_hit
        self.attr.max_num_of_at_attach = max_num_of_at_attach


cdef class HwsMatcher(PyverbsCM):
    def __init__(self, HwsTable table, match_templates, num_of_mt, action_templates, num_of_at,
                 HwsMatcherAttr matcher_attr):
        """
        Initialize a HwsMatcher object representing hws_matcher C struct.
        :param table: Matcher's HwsTable table
        :param match_templates: List of match templates
        :param num_of_mt: Number of match templates
        :param action_templates: List of action templates
        :param num_of_at: Number of action templates
        :param matcher_attr: Attributes for creating HwsMatcher
        """
        super().__init__()
        cdef hws.mlx5dv_hws_mt **mt
        cdef hws.mlx5dv_hws_at **at
        mt = <hws.mlx5dv_hws_mt **>calloc(num_of_mt, sizeof(hws.mlx5dv_hws_mt *))
        if mt == NULL:
            raise MemoryError('Failed allocating memory.')
        at = <hws.mlx5dv_hws_at **>calloc(num_of_at, sizeof(hws.mlx5dv_hws_at *))
        if at == NULL:
            free(mt)
            raise MemoryError('Failed allocating memory.')
        for i in range(num_of_mt):
            match_template = <HwsMatchTemplate>(match_templates[i])
            mt[i] = <hws.mlx5dv_hws_mt *>(match_template.match_template)
        for i in range(num_of_at):
            action_template = <HwsActionTemplate>(action_templates[i])
            at[i] = <hws.mlx5dv_hws_at *>(action_template.action_template)
            for i in range(len(action_template.actions)):
                action = <HwsAction>(action_template.actions[i])
                <HwsAction>action.add_ref(self)
        self.matcher = hws.mlx5dv_hws_matcher_create(table.table, mt, num_of_mt, at, num_of_at,
                                                     &matcher_attr.attr)
        free(mt)
        free(at)
        if self.matcher == NULL:
            raise PyverbsError('Failed creating HwsMatcher.')
        self.match_templates = match_templates[:]
        self.action_templates = action_templates[:]
        table.add_ref(self)
        self.hws_table = table
        self.rules = weakref.WeakSet()

    def get_metric(self, flags):
        return HwsMatcherMetric(self, flags)

    def get_mt_metric(self, HwsMetricMatcherTemplateAttr attr):
        return HwsMatcherTemplateMetric(self, attr)

    cdef add_ref(self, obj):
        if isinstance(obj, HwsRule):
            self.rules.add(obj)
        else:
            raise PyverbsError('Unrecognized object type.')

    cdef del_ref(self, obj):
        if isinstance(obj, HwsRule):
            self.rules.remove(obj)
        else:
            raise PyverbsError('Unrecognized object type.')

    def resize_set_target(self, HwsMatcher dst_matcher):
        """
        Link two matchers and allow moving rules from src matcher to dst matcher.
        Both matchers must be in the same table, must be created with 'resizable'
        property, and should have the same characteristics (e.g. same mt, same at).

        It is the user's responsibility to make sure that the dst matcher
        has enough space for the moved rules (i.e. it was allocated with
        the appropriate size).

        Once the function is completed, the user is:
         - allowed to move rules from src into dst matcher
         - no longer allowed to insert rules to the src matcher

        The user is always allowed to insert rules to the dst matcher and
        to delete rules from any matcher.
        :param dst_matcher: Destination matcher for moving rules
        """
        ret = hws.mlx5dv_hws_matcher_resize_set_target(self.matcher, dst_matcher.matcher)
        if ret:
            raise PyverbsError('Failed setting target for matcher resize.')

    def attach_at(self, HwsActionTemplate at):
        """
        Attach new action template to the matcher.
        :param at: Action template to be attached to the matcher.
        """
        ret = hws.mlx5dv_hws_matcher_attach_at(self.matcher, at.action_template)
        if ret:
            raise PyverbsError('Failed attaching action template to matcher.')
        self.action_templates.append(at)

    def rule_hash_calculate(self, items_data, mt_idx):
        """
        Querying the hash result for a given matcher and a set of items.
        This API can be used to predict the location of the rule in the hash table.
        :param items_data: List of HwsData with values to calculate the hash.
        :param mt_idx: Match template index.
        """
        cdef uint32_t res = 0
        cdef uint32_t res_raw = 0
        cdef hws.mlx5dv_hws_item_data *c_items = \
                        <hws.mlx5dv_hws_item_data *>calloc(len(items_data),
                                                           sizeof(hws.mlx5dv_hws_item_data))
        if c_items == NULL:
            raise PyverbsError('Failed allocating memory.')

        for i in range(len(items_data)):
            item_py_i = <HwsData>(items_data[i])
            item_c_i_ptr = &(item_py_i.data)
            memcpy(&c_items[i], item_c_i_ptr, sizeof(hws.mlx5dv_hws_item_data))
        rc = hws.mlx5dv_hws_rule_hash_calculate(self.matcher, c_items, mt_idx, &res, &res_raw)
        free(c_items)
        if rc:
            raise PyverbsError('Failed to calculate hash value.')
        return res

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.matcher != NULL:
            self.logger.debug('Closing HwsMatcher.')
            close_weakrefs([self.rules])
            rc = hws.mlx5dv_hws_matcher_destroy(<hws.mlx5dv_hws_matcher *>(self.matcher))
            if rc:
                raise PyverbsError('Failed to destroy HwsMatcher.', rc)
            self.matcher = NULL
            self.hws_table = None
            self.match_templates = None
            self.action_templates = None

cdef class HwsMatcherMetric(PyverbsObject):
    def __init__(self, HwsMatcher matcher, flags):
        super().__init__()
        rc = hws.mlx5dv_hws_metric_query_matcher(<hws.mlx5dv_hws_matcher*>(matcher.matcher), flags,
                                                 &self.metric)
        if rc:
            raise PyverbsError('Failed querying matcher metrics.')

    def __str__(self):
        return f'STE tables memory in bytes: {self.matcher_tables_mem_sz}\n'\
               f'Matcher id: {self.matcher_id}\n'\
               f'Matcher priority: {self.priority}\n'\
               f'Log number of rows: {self.log_num_of_rows}\n'\
               f'Log number of colunms: {self.log_num_of_columns}\n'\
               f'Collision matcher: {self.col_matcher_exist}\n'\
               f'Collision matcher log number of rows: {self.col_matcher_log_num_of_rows}\n'\
               f'Collision matcher log number of colunms: {self.col_matcher_log_num_of_columns}\n'\
               f'Number of match STEs: {self.num_of_match_stes}\n'\
               f'Number of action STEs: {self.num_of_action_stes}\n'\
               f'Is jumbo: {self.is_jumbo}\n'\
               f'Memory cost: {self.memory_cost}\n'\
               f'Cost of miss packet processing: {self.miss_pkt_processing}\n'

    @property
    def matcher_tables_mem_sz(self):
        return self.metric.info.matcher_tables_mem_sz
    @property
    def matcher_id(self):
        return self.metric.info.matcher_id

    @property
    def priority(self):
        return self.metric.info.priority
    @property
    def log_num_of_rows(self):
        return self.metric.info.log_num_of_rows

    @property
    def log_num_of_columns(self):
        return self.metric.info.log_num_of_columns

    @property
    def col_matcher_exist(self):
        return self.metric.info.col_matcher_exist

    @property
    def col_matcher_log_num_of_rows(self):
        return self.metric.info.col_matcher_log_num_of_rows

    @property
    def col_matcher_log_num_of_columns(self):
        return self.metric.info.col_matcher_log_num_of_columns

    @property
    def num_of_match_stes(self):
        return self.metric.info.num_of_match_stes

    @property
    def num_of_action_stes(self):
        return self.metric.info.num_of_action_stes

    @property
    def is_jumbo(self):
        return self.metric.info.is_jumbo

    @property
    def memory_cost(self):
        return self.metric.cost.memory

    @property
    def miss_pkt_processing(self):
        return self.metric.cost.miss_pkt_processing

cdef class HwsMetricMatcherTemplateAttr(PyverbsCM):
    def __init__(self, flags, HwsAction dest_action, actions_data=[], mt_idx=0, at_idx=0):
        """
        Initialize HwsMetricMatcherTemplateAttr object representing
        mlx5dv_hws_metric_matcher_template_attr C struct.
        :param flags: mlx5dv_hws_metric_flag enum.
        :param dest_action: Destination action.
        :param actions_data: Optional action data array used for more accurate
                             cost calculation.
        :param mt_idx: Index of match template.
        :param at_idx: Index of action template.
        """
        self.attr.flags = flags
        self.attr.dest_action = dest_action.action if dest_action else NULL
        if actions_data:
            self.attr.actions_data = <hws.mlx5dv_hws_action_data *>calloc(len(actions_data),
                                     sizeof(hws.mlx5dv_hws_action_data))
            if not self.attr.actions_data:
                raise MemoryError('Memory allocation failed.')
            for i in range(len(actions_data)):
                self.attr.actions_data[i] = (<HwsActionData>actions_data[i]).action_data
        self.attr.mt_idx = mt_idx
        self.attr.at_idx = at_idx

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.attr.actions_data:
            free(self.attr.actions_data)
            self.attr.actions_data = NULL


cdef class HwsMatcherTemplateMetric(PyverbsObject):
    def __init__(self, HwsMatcher matcher, HwsMetricMatcherTemplateAttr attr):
        """
        Initialize HwsMatcherTemplateMetric object by querying the matcher template metric.
        :param matcher: HwsMatcher matcher to extract the info from.
        :param attr: HwsMetricMatcherTemplateAttr attributes.
        """
        super().__init__()
        rc = hws.mlx5dv_hws_metric_query_matcher_template(matcher.matcher, &(attr.attr),
                                                          &self.metric)
        if rc:
            raise PyverbsError('Failed querying matcher template metrics.')

    @property
    def match_stes(self):
        return self.metric.resource.match_stes

    @property
    def action_stes(self):
        return self.metric.resource.action_stes

    @property
    def counters(self):
        return self.metric.resource.counters

    @property
    def aso(self):
        return self.metric.resource.aso

    @property
    def args(self):
        return self.metric.resource.args

    @property
    def crypto(self):
        return self.metric.resource.crypto

    @property
    def memory_cost(self):
        return self.metric.cost.memory

    @property
    def pkt_processing_cost(self):
        return self.metric.cost.pkt_processing

    @property
    def insertion_cost(self):
        return self.metric.cost.insertion

    def __str__(self):
        return (f'Match STES: {self.match_stes}\n'
                f'Action STES: {self.action_stes}\n'
                f'Counters: {self.counters}\n'
                f'ASO: {self.aso}\n'
                f'Args: {self.args}\n'
                f'Crypto: {self.crypto}\n'
                f'Memory cost: {self.memory_cost}\n'
                f'Packet processing cost: {self.pkt_processing_cost}\n'
                f'Rule insertion cost: {self.insertion_cost}\n')
