# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

cimport pyverbs.providers.mlx5.mlx5hws_enums as e
from pyverbs.pyverbs_error import PyverbsError
from pyverbs.base cimport close_weakrefs
from libc.stdint cimport uint32_t
import weakref


cdef class HwsParserHeaderFieldConfig(PyverbsObject):
    def __init__(self, field_bit_offset, field_bit_length):
        """
        Initialize a HwsParserHeaderFieldConfig object representing mlx5dv_hws_parser_field_cfg
        C struct.
        :param field_bit_offset: Location of the field within the header.
        :param field_bit_length: Size of the field.
        """
        super().__init__()
        self.field_config.bit_offset = field_bit_offset
        self.field_config.bit_length = field_bit_length


cdef class HwsParserHeaderLenConfig(PyverbsObject):
    def __init__(self, mode, hdr_len=0, HwsParserHeaderFieldConfig field_config=None,
                 multiplier=None, addition=None):
        """
        Initialize HwsParserHeaderLenConfig object representing mlx5dv_hws_parser_hdr_len_cfg
        C struct.
        :param mode: Header length mode can be either fixed value or explicit in a field.
        :param hdr_len: Fixed length value if mode is fixed otherwise 0.
        :param field_config: HwsParserHeaderFieldConfig to extract the length value if mode is
                             field, otherwise None.
        :param multiplier: Extracted length is multiplied by this amount if mode is field,
                           otherwise None.
        :param addition: Value to add after the multiplication of the extracted length (if mode is
                         field, otherwise None).
        """
        super().__init__()
        self.len_config.mode = mode
        if mode == e.MLX5DV_HWS_PARSER_HDR_LEN_MODE_FIXED:
            self.len_config.fixed.hdr_len = hdr_len
        else:
            self.len_config.field.field = field_config.field_config
            self.len_config.field.multiplier = multiplier
            self.len_config.field.addition = addition


cdef class HwsParserNodeConfig(PyverbsObject):
    def __init__(self, HwsParserHeaderLenConfig hdr_len, has_options=False,
                 has_selection_field=False, option_offset=None,
                 HwsParserHeaderLenConfig option_len=None,
                 HwsParserHeaderFieldConfig option_type_field=None,
                 HwsParserHeaderFieldConfig selection_field=None):
        """
        Initialize a HwsParserNodeConfig object representing mlx5dv_hws_parser_node_cfg C struct.
        :param hdr_len: HwsParserHeaderLenConfig to calculate the header length.
        :param has_options: Flag to indicate if options (TLVs) are processed for this header.
        :param has_selection_field: Flag to indicate the field is extracted from the header to be
                                    used as the selection value.
        :param option_offset: Offset at which to start parsing options, if has_options flag is true,
                              else None.
        :param option_len: HwsParserHeaderLenConfig to calculate an option length, if has_options
                           flag is true, else None.
        :param option_type_field: HwsParserHeaderFieldConfig configuration of an option's field,
                                  if has_options flag is true, else None.
        :param selection_field: HwsParserHeaderFieldConfig configuration of the selection field,
                                if has_selection_field flag is true, else None.
        """
        super().__init__()
        self.config.hdr_len = <hws.mlx5dv_hws_parser_hdr_len_cfg>(hdr_len.len_config)
        self.config.has_options = has_options
        self.config.has_selection_field = has_selection_field

        if has_options:
            self.config.option_offset = option_offset
            self.config.option_len = <hws.mlx5dv_hws_parser_hdr_len_cfg>(option_len.len_config)
            self.config.option_type_field = option_type_field.field_config

        if has_selection_field:
            self.config.selection_field = selection_field.field_config


cdef class HwsParserArcNode(PyverbsObject):
    def __init__(self, node_type, native_node=None, HwsParserNode flex_node=None):
        """
        Initialize a HwsParserArcNode object representing mlx5dv_hws_parser_arc_node_cfg C struct.
        :param node_type: Enum indicating node type can be either native or flex.
        :param native_node: Enum indicating the native node, if type is native, else None.
        :param flex_node: HwsParserNode indicating the flex node, if type is flex, else None.
        """
        super().__init__()
        self.arc_node.type = node_type
        if node_type == e.MLX5DV_HWS_PARSER_NODE_TYPE_NATIVE:
            self.arc_node.native.node = native_node
        else:
            self.arc_node.flex.node = <hws.mlx5dv_hws_parser_node *>(flex_node.node)


cdef class HwsParserArcConfig(PyverbsObject):
    def __init__(self, HwsParserArcNode src_node, HwsParserArcNode dst_node, src_encap_level=0,
                 selection_field_data=None, is_inner_transition=False):
        """
        Initialize a HwsParserArcConfig object representing mlx5dv_hws_parser_arc_cfg C struct.
        :param src_node: Source HwsParserArcNode node of the arc.
        :param dst_node: Destination HwsParserArcNode node of the arc.
        :param src_encap_level: Encapsulation level of the source node.
        :param selection_field_data: The selection field data/value that selects this specific arc,
                                     Ignored if source node has_selection_field is false.
        :param is_inner_transition: Flag indicating if the arc pass across between encapsulation
                                    levels.
        """
        super().__init__()
        self.config.src_node = src_node.arc_node
        self.config.dst_node = dst_node.arc_node
        self.config.src_encap_level = src_encap_level
        self.config.selection_field_data = selection_field_data
        self.config.is_inner_transition = is_inner_transition


cdef class HwsParserSamplerConfig(PyverbsObject):
    def __init__(self, node_encap_level, HwsParserHeaderFieldConfig field_cfg,
                 sample_from_option=False, option_type=None, share_hw_resources=None):
        """
        Initialize HwsParserSamplerConfig object representing mlx5dv_hws_parser_sampler_cfg struct.
        :param node_encap_level: The encapsulation level to sample from.
        :param field_cfg: The HwsParserHeaderFieldConfig of the field to sample.
        :param sample_from_option: False to sample from main header, True to sample from an option.
        :param option_type: The option type to sample from. None if sample_from_option is False.
        :param share_hw_resources: Optional HwsParserSampler previously created sampler to
                                   share HW resources with.
        """
        super().__init__()
        self.config.node_encap_level = node_encap_level
        self.config.field_cfg = field_cfg.field_config
        self.config.sample_from_option = sample_from_option
        if sample_from_option:
            self.config.option_type = option_type
        if share_hw_resources:
            self.config.share_hw_resources = \
                                    <hws.mlx5dv_hws_parser_sampler *>(share_hw_resources.sampler)


cdef class HwsParserGeneveOptionConfig(PyverbsObject):
    def __init__(self, option_class, option_type, option_data_len=0, option_class_ignore=0,
                 offset_valid=1, sample_offset=0):
        """
        Initialize a HwsParserGeneveOptionConfig object representing-
        mlx5dv_hws_parser_geneve_option_cfg C struct.
        :param option_class: The class of the GENEVE option. Option classes are managed by IANA
                             and define the type of metadata being carried in the option.
                             (as a namespace for the type field, ensuring that options from
                             different classes do not conflict with each other)
        :param option_type: The type of the GENEVE option within the given class.
                            Each class can have multiple types of options.
        :param option_data_len: Data length in granularity of four bytes.
                                Must be <= HCA_CAP.max_geneve_tlv_option_data_len.
                                NOTE - Matching only for the existence of the option is only
                                supported for options when option_data_len is 0x0.
        :param option_class_ignore: If set to 1, matching option_class is not part of the criteria
                                    for the existence of the Option and for the sampling.
                                    Valid if HCA_CAP.geneve_tlv_sample is set.
        :param offset_valid: If set to 1, the sample_offset field is considered valid.
        :param sample_offset: Indicates the offset of the sampled DW from the start of the TLV.
                              Value is given in DW. Value 0 indicates sampling starting from the
                              start of the TLV. Valid if HCA_CAP.geneve_tlv_sample is set.
        """
        super().__init__()
        self.config.option_class = option_class
        self.config.option_type = option_type
        self.config.option_data_len = option_data_len
        self.config.option_class_ignore = option_class_ignore
        self.config.offset_valid = offset_valid
        self.config.sample_offset = sample_offset


cdef class HwsParserGraph(PyverbsCM):
    def __init__(self, HwsContext context):
        """
        Initialize a HwsParserGraph object representing mlx5dv_hws_parser_graph C struct.
        :param context: HwsContext context.
        """
        super().__init__()
        self.graph = hws.mlx5dv_hws_parser_graph_create(context.context)
        if self.graph == NULL:
            raise PyverbsError('Failed creating HwsParserGraph.')
        self.hws_context = context
        context.add_ref(self)
        self.graph_nodes = weakref.WeakSet()
        self.graph_arcs = weakref.WeakSet()

    cdef add_ref(self, obj):
        if isinstance(obj, HwsParserNode):
            self.graph_nodes.add(obj)
        elif isinstance(obj, HwsParserArc):
            self.graph_arcs.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')

    def bind(self):
        rc = hws.mlx5dv_hws_parser_graph_bind(self.graph)
        if rc:
            raise PyverbsError('Failed to bind HwsParserGraph.', rc)

    def unbind(self):
        rc = hws.mlx5dv_hws_parser_graph_unbind(self.graph)
        if rc:
            raise PyverbsError('Failed to unbind HwsParserGraph.', rc)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.graph != NULL:
            self.logger.debug('Closing HwsParserGraph.')
            close_weakrefs([self.graph_arcs, self.graph_nodes])
            rc = hws.mlx5dv_hws_parser_graph_destroy(<hws.mlx5dv_hws_parser_graph *>(self.graph))
            if rc:
                raise PyverbsError('Failed to destroy HwsParserGraph.', rc)
            self.graph = NULL
            self.hws_context = None


cdef class HwsParserNode(PyverbsCM):
    def __init__(self, HwsParserGraph graph, HwsParserNodeConfig node_cfg):
        """
        Initialize a HwsParserNode object representing mlx5dv_hws_parser_node C struct.
        :param graph: Graph object that includes the node.
        :param node_cfg: Node configuration object.
        """
        super().__init__()
        self.node = hws.mlx5dv_hws_parser_node_create(graph.graph, &(node_cfg.config))
        if self.node == NULL:
            raise PyverbsError('Failed creating HwsParserNode.')
        graph.add_ref(self)
        self.node_samplers = weakref.WeakSet()

    cdef add_ref(self, obj):
        if isinstance(obj, HwsParserSampler):
            self.node_samplers.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.node != NULL:
            self.logger.debug('Closing HwsParserNode.')
            close_weakrefs([self.node_samplers])
            rc = hws.mlx5dv_hws_parser_node_destroy(<hws.mlx5dv_hws_parser_node *>(self.node))
            if rc:
                raise PyverbsError('Failed to destroy HwsParserNode.', rc)
            self.node = NULL


cdef class HwsParserArc(PyverbsCM):
    def __init__(self, HwsParserGraph graph, HwsParserArcConfig arc_cfg):
        """
        Initialize a HwsParserArc object representing mlx5dv_hws_parser_arc C struct.
        :param graph: Graph object that includes the arc.
        :param arc_cfg: Arc configuration object.
        """
        super().__init__()
        self.arc = hws.mlx5dv_hws_parser_arc_create(graph.graph, &(arc_cfg.config))
        if self.arc == NULL:
            raise PyverbsError('Failed creating HwsParserArc.')
        graph.add_ref(self)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.arc != NULL:
            self.logger.debug('Closing HwsParserArc.')
            rc = hws.mlx5dv_hws_parser_arc_destroy(<hws.mlx5dv_hws_parser_arc *>(self.arc))
            if rc:
                raise PyverbsError('Failed to destroy HwsParserArc.', rc)
            self.arc = NULL


cdef class HwsParserSampler(PyverbsCM):
    def __init__(self, sampler_cfg, HwsParserNode node=None, HwsContext context=None):
        """
        Initialize a HwsParserSampler object representing mlx5dv_hws_parser_sampler C struct.
        :param sampler_cfg: HwsParserSamplerConfig or HwsParserGeneveOptionConfig object.
        :param node: HwsParserNode to use when sampler_cfg of type HwsParserSamplerConfig.
        :param context: The context in which the new geneve object will be created,
                        Note - sampler_cfg of type HwsParserGeneveOptionConfig.
        """
        super().__init__()
        if isinstance(sampler_cfg, HwsParserSamplerConfig):
            self.sampler = hws.mlx5dv_hws_parser_sampler_create(node.node,
                                                    &((<HwsParserSamplerConfig>sampler_cfg).config))
            if self.sampler:
                node.add_ref(self)
        else:
            self.sampler = hws.mlx5dv_hws_parser_geneve_tlv_options_create(context.context,
                                               &((<HwsParserGeneveOptionConfig>sampler_cfg).config))
        if self.sampler == NULL:
            raise PyverbsError('Failed creating HwsParserSampler.')

    def query(self):
        """
        Query the sampler for its attributes.
        :return parser_fname: FNAME for matching on the sampled field.
                parser_ok_bit_fname: FNAME for indication of successful parsing and sampling.
                field_header: Field header location.
                modify_header_field: Modify header field id that refers to this samplers' value.
        """
        cdef e.mlx5dv_hws_field_name parser_fname
        cdef e.mlx5dv_hws_field_name parser_ok_bit_fname
        cdef e.mlx5dv_hws_field_header field_header
        cdef uint32_t modify_header_field

        rc = hws.mlx5dv_hws_parser_sampler_query(<hws.mlx5dv_hws_parser_sampler *>(self.sampler),
                                                &parser_fname, &parser_ok_bit_fname,
                                                &field_header, &modify_header_field)
        if rc:
            raise PyverbsError('Failed to query HwsParserSampler.', rc)

        return parser_fname, parser_ok_bit_fname, field_header, modify_header_field

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.sampler != NULL:
            self.logger.debug('Closing HwsParserSampler.')
            rc = hws.mlx5dv_hws_parser_sampler_destroy(<hws.mlx5dv_hws_parser_sampler *> \
                                                       (self.sampler))
            if rc:
                raise PyverbsError('Failed to destroy HwsParserSampler.', rc)
            self.sampler = NULL
