# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

from pyverbs.providers.mlx5.mlx5dv cimport Wqe, _prepare_devx_inbox, _prepare_devx_outbox
from pyverbs.providers.mlx5.hws_resource cimport HwsResource
from pyverbs.providers.mlx5.hws_actions cimport HwsAction
from pyverbs.providers.mlx5.hws_item_data cimport HwsData
cimport pyverbs.providers.mlx5.mlx5hws_enums as e
from pyverbs.pyverbs_error import PyverbsError
cimport pyverbs.providers.mlx5.libmlx5 as dv
from pyverbs.base cimport close_weakrefs
from libc.stdlib cimport free, calloc
from libc.stdint cimport uintptr_t
from libc.string cimport memcpy
import weakref
import time


ctypedef unsigned int __be32


cdef class HwsResourceAttr(PyverbsObject):
    def __init__(self, resource_type, bulk_log_size=0, block_alloc_support=0, comp_mask=0):
        """
        Initializes a HwsResourceAttr object representing mlx5dv_hws_resource_attr struct.
        :param resource_type: Resource type.
        :param bulk_log_size: Resource bulk log size.
        :param block_alloc_support: Indicates whether this resource allows allocating contiguous
                                    blocks of power-of-two size offsets.
        :param comp_mask: Compatability mask.
        """
        super().__init__()
        self.attr.type = resource_type
        self.attr.bulk_log_size = bulk_log_size
        self.attr.block_alloc_support = block_alloc_support
        self.attr.comp_mask = comp_mask


cdef class HwsResourceEnqueueAsoAttr(PyverbsObject):
    def __init__(self, burst=False, user_data=None, offset=0, comp_mask=0):
        """
        Initializes a HwsResourceEnqueueAsoAttr object representing
        mlx5dv_hws_resource_enqueue_aso_attr struct.
        :param burst: Notify HW flag.
        :param user_data: Returned user data in the completion entry.
        :param offset: Resource offset in the bulk.
        :param comp_mask: Compatability mask.
        """
        super().__init__()
        self.attr.burst = burst
        self.attr.user_data = <void *>user_data if user_data else NULL
        self.attr.offset = offset
        self.attr.comp_mask = comp_mask
        self.user_data = user_data

    @property
    def user_data(self):
        return self.user_data


cdef class HwsResource(PyverbsCM):
    def __init__(self, HwsContext context, HwsResourceAttr attr):
        """
        Initializes a HwsResource object representing mlx5dv_hws_resource struct.
        :param context: HwsContext
        :param attr: Attribute that defines the resource.
        """
        super().__init__()
        self.resource = hws.mlx5dv_hws_resource_alloc(context.context,
                                                      <hws.mlx5dv_hws_resource_attr *>&(attr.attr))
        if not self.resource:
            raise PyverbsError('Failed to allocate HwsResource.')
        self.context = context
        self.ret_obj = NULL
        self.actions = weakref.WeakSet()
        self.context.add_ref(self)

    cdef add_ref(self, obj):
        if isinstance(obj, HwsAction):
            self.actions.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')

    @property
    def devx_obj_id(self):
        if self.ret_obj:
            return self.ret_obj_id
        else:
            raise PyverbsError('Failed to get a devx object id.')

    @property
    def max_block_size(self):
        return hws.mlx5dv_hws_resource_max_block_size(<hws.mlx5dv_hws_resource *>(self.resource))

    def get_offset(self, qid):
        cdef uint32_t offset

        rc = hws.mlx5dv_hws_resource_get_offset(<hws.mlx5dv_hws_resource *>(self.resource), qid,
                                                &offset)
        if rc < 0:
            raise PyverbsError('Failed to get a single HwsResource.')
        return offset

    def put_offset(self, offset, qid):
        rc = hws.mlx5dv_hws_resource_put_offset(<hws.mlx5dv_hws_resource *>(self.resource), qid,
                                                offset)
        if rc:
            raise PyverbsError('Failed to put a single HwsResource.', rc)

    def get_offset_block(self, qid, log_sz):
        cdef uint32_t offset

        rc = hws.mlx5dv_hws_resource_get_offset_block(<hws.mlx5dv_hws_resource *>(self.resource),
                                                      qid, log_sz, &offset)
        if rc < 0:
            raise PyverbsError('Failed to get block of HwsResources.')
        return offset

    def put_offset_block(self, offset, qid, log_sz):
        rc = hws.mlx5dv_hws_resource_put_offset_block(<hws.mlx5dv_hws_resource *>(self.resource),
                                                      qid, log_sz, offset)
        if rc:
            raise PyverbsError('Failed to put a block of HwsResources.', rc)

    def get_devx_obj(self):
        rc = hws.mlx5dv_hws_resource_get_devx_obj(<hws.mlx5dv_hws_resource *>(self.resource),
                                                  &(self.ret_obj),
                                                  &(self.ret_obj_id))
        if rc:
            raise PyverbsError('Failed to get a devx object.', rc)

    def enqueue_aso_wqe(self, queue_id, Wqe aso_wqe, wqe_len, HwsResourceEnqueueAsoAttr attr):
        """
        Posting ASO WQE into the QP.
        :param queue_id: Queue id to use.
        :param aso_wqe: Raw ASO CTRL and DATA segments for enqueue in BE format.
        :param wqe_len: The size of the wqe to be posted.
        :param attr: HwsResourceEnqueueAsoAttr - enqueue ASO wqe attribute.
        """
        cdef __be32 *aso_wqe_c = <__be32 *>(<uintptr_t>(aso_wqe.address))
        cdef hws.mlx5dv_hws_resource_enqueue_aso_attr *attr_ptr = \
            <hws.mlx5dv_hws_resource_enqueue_aso_attr *>&(attr.attr)

        rc = hws.mlx5dv_hws_resource_enqueue_aso_wqe(<hws.mlx5dv_hws_resource *>(self.resource),
                                                     queue_id, aso_wqe_c, wqe_len, attr_ptr)
        if rc:
            raise PyverbsError('Failed to enqueue ASO WQE.', rc)
        if not attr.attr.burst:
            self.context.check_op_completion(queue_id)

    def query(self, in_, outlen):
        """
        Queries the DevX object of the resource.
        To populate the DevX object info from the resource object, we call get_devx_obj.
        :param in_: Bytes of the obj_query command's input data provided in a
                    device specification format.
                    (Stream of bytes or __bytes__ is implemented)
        :param outlen: Expected output length in bytes
        :return: Bytes of the command's output
        """
        in_bytes = bytes(in_)
        cdef char *in_mailbox = _prepare_devx_inbox(in_bytes)
        cdef char *out_mailbox = _prepare_devx_outbox(outlen)
        self.get_devx_obj()
        rc = dv.mlx5dv_devx_obj_query(self.ret_obj, in_mailbox, len(in_bytes),
                                      out_mailbox, outlen)
        try:
            if rc:
                raise PyverbsError('Failed to query DevX object', rc)
            out = <bytes>out_mailbox[:outlen]
        finally:
            free(in_mailbox)
            free(out_mailbox)
        return out

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.resource != NULL:
            self.logger.debug('Closing HwsResource.')
            close_weakrefs([self.actions])
            rc = hws.mlx5dv_hws_resource_free(<hws.mlx5dv_hws_resource *>(self.resource))
            if rc:
                raise PyverbsError('Failed to destroy HwsResource.', rc)
            self.resource = NULL
            self.context = None
