# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

from pyverbs.providers.mlx5.hws_actions cimport HwsActionData
from pyverbs.providers.mlx5.hws_item_data cimport HwsData
cimport pyverbs.providers.mlx5.mlx5hws_enums as e
from pyverbs.pyverbs_error import PyverbsError
from libc.stdlib cimport free, calloc


cdef class HwsRuleQuery(PyverbsObject):
    def __init__(self, HwsRule rule, comp_mask=0):
        """
        Initialize a HwsRuleQuery object representing mlx5dv_hws_rule_query C struct
        :param comp_mask: Compatibility mask
        """
        super().__init__()
        self.query.comp_mask = comp_mask
        rc = hws.mlx5dv_hws_rule_query(rule.mlx5dv_hws_rule, &self.query)
        if rc:
            raise PyverbsError('Failed to query HwsRule.')

    @property
    def match_tag_sz(self):
        return self.query.match_tag_sz

    @property
    def match_tag(self):
        res = bytearray(self.match_tag_sz)
        for i in range(self.match_tag_sz):
            res[i] = self.query.match_tag[i]
        return res

    def __str__(self):
        return (f'Rule match tag size: {self.match_tag_sz}\n'
                f'Rule match tag: {self.match_tag}')


cdef class HwsRuleAttr(PyverbsObject):
    def __init__(self, queue_id=0, user_data=None, burst=0, rule_idx=0, comp_mask=0):
        """
        Initializes a HwsRuleAttr object representing mlx5dv_hws_rule_attr struct.
        """
        super().__init__()
        self.attr.queue_id = queue_id
        self.attr.user_data = <void *>user_data if user_data else NULL
        self.attr.burst = burst
        self.attr.rule_idx = rule_idx
        self.user_data = user_data
        self.attr.comp_mask = comp_mask

    @property
    def user_data(self):
        return self.user_data


cdef class HwsRule(PyverbsCM):
    def __init__(self, HwsMatcher matcher, mt_idx, item_data, at_idx, action_data,
                 HwsAction dest_action, HwsRuleAttr rule_attr_create,
                 HwsContext ctx=None, HwsRuleAttr rule_attr_destroy=None):
        """
        Initializes a HwsRule object representing mlx5dv_hws_rule struct.
        :param matcher: HwsMatcher to create a rule with.
        :param mt_idx: Index of the matcher template to use.
        :param item_data: List of HwsData items with data for matching.
        :param at_idx: Index of the action template to use.
        :param action_data: List of HwsActionData with actions' data executed on match.
        :param rule_attr_create: Attributes for rule creation.
        :param dest_action: The destination action in case of match.
        :param ctx: If provided poll send queue for rule creation completion.
        :param rule_attr_destroy: Attributes for rule destruction (if not provided,
                                  rule_attr_create is used for destruction instead).
        """
        super().__init__()
        cdef hws.mlx5dv_hws_rule *rule = NULL
        cdef hws.mlx5dv_hws_item_data *item_data_ptr = \
            <hws.mlx5dv_hws_item_data *>calloc(len(item_data), sizeof(hws.mlx5dv_hws_item_data))
        cdef hws.mlx5dv_hws_action_data *actions_ptr = \
            <hws.mlx5dv_hws_action_data *>calloc(len(action_data),
                                                 sizeof(hws.mlx5dv_hws_action_data))

        if not item_data_ptr or not actions_ptr:
            free(item_data_ptr)
            free(actions_ptr)
            raise MemoryError('Memory allocation failed.')

        self.context = ctx
        rule = <hws.mlx5dv_hws_rule *>calloc(1, hws.mlx5dv_hws_rule_get_handle_size())

        if rule == NULL:
            free(item_data_ptr)
            free(actions_ptr)
            raise MemoryError('Memory allocation failed.')
        for i in range(len(action_data)):
            actions_ptr[i] = (<HwsActionData>action_data[i]).action_data
        for i in range(len(item_data)):
            item_data_ptr[i].match = (<HwsData>item_data[i]).data.match
            item_data_ptr[i].mask = (<HwsData>item_data[i]).data.mask
            item_data_ptr[i].last = (<HwsData>item_data[i]).data.last
        rc = hws.mlx5dv_hws_rule_create(matcher.matcher, mt_idx, item_data_ptr, at_idx, actions_ptr,
                                        dest_action.action, &rule_attr_create.attr, rule)
        free(item_data_ptr)
        free(actions_ptr)
        if rc:
            free(rule)
            raise PyverbsError('Failed to create HwsRule.')
        self.dest_action = dest_action
        dest_action.add_ref(self)
        self.mlx5dv_hws_rule = rule
        self.rule_attr_destroy = rule_attr_destroy if rule_attr_destroy else rule_attr_create
        self.matcher = matcher
        matcher.add_ref(self)
        if ctx and not rule_attr_create.attr.burst:
            ctx.check_op_completion(rule_attr_create.attr.queue_id)

    def matcher_resize_rule_move(self, HwsMatcher src_matcher, HwsMatcher dst_matcher,
                                 HwsRuleAttr rule_attr):
        """
        Moving rule from matcher that is being resized to a bigger matcher.
        Moving includes creating a new rule in the destination matcher, and
        deleting the rule from the source matcher.
        This operation creates a single completion.

        Since the user is always allowed to insert rules into the dst matcher,
        it is the user's responsibility to synchronize between moving rules and
        inserting new rules into dst matcher.

        The user is responsible for making sure that the dst matcher has enough
        space for the moved rule, and that the place for the moved rule is
        available (such as when moving rule that was inserted by index - the
        index in the dst matcher should be available).
        :param src_matcher: HwsMatcher that the rule belongs to.
        :param dst_matcher: HwsMatcher that the rule is going to be moved to.
        :param attr: The rule attributes.
        """
        ret = hws.mlx5dv_hws_matcher_resize_rule_move(src_matcher.matcher, self.mlx5dv_hws_rule,
                                                      &(rule_attr.attr))
        if ret:
            raise PyverbsError('Failed moving rule to resized matcher.')
        dst_matcher.add_ref(self)
        src_matcher.del_ref(self)
        self.matcher = dst_matcher

    def update(self, at_idx, action_data, HwsAction dest_action, HwsRuleAttr rule_attr,
               HwsContext ctx=None):
        """
        Update an exiting rule action data with a new action data, rule match stays
        the same but the exectued actions and values differ bas on the at_idx and
        action data.
        """
        cdef hws.mlx5dv_hws_action_data *actions_ptr = \
            <hws.mlx5dv_hws_action_data *>calloc(len(action_data),
                                                 sizeof(hws.mlx5dv_hws_action_data))

        if actions_ptr == NULL:
            raise MemoryError('Memory allocation failed.')
        for i in range(len(action_data)):
            actions_ptr[i] = (<HwsActionData>action_data[i]).action_data
        rc = hws.mlx5dv_hws_rule_action_data_update(self.mlx5dv_hws_rule, at_idx, actions_ptr,
                                                    dest_action.action, &rule_attr.attr)
        free(actions_ptr)
        if rc:
            raise PyverbsError('Failed to update HwsRule.')
        self.dest_action.del_ref(self)
        self.dest_action = dest_action
        dest_action.add_ref(self)

        if ctx and not rule_attr.attr.burst:
            ctx.check_op_completion(rule_attr.attr.queue_id)

    def rule_query(self, comp_mask=0):
        return HwsRuleQuery(self, comp_mask)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.mlx5dv_hws_rule != NULL:
            self.logger.debug('Closing HwsRule.')
            attr = self.rule_attr_destroy
            rc = hws.mlx5dv_hws_rule_destroy(self.mlx5dv_hws_rule,
                                             <hws.mlx5dv_hws_rule_attr *>&(attr.attr))
            if rc:
                raise PyverbsError('Failed to destroy HwsRule.', rc)
            if self.context:
                self.context.check_op_completion(attr.attr.queue_id)
            free(self.mlx5dv_hws_rule)
            self.mlx5dv_hws_rule = NULL
            self.matcher = None
            self.dest_action = None
            self.context = None


cpdef get_rule_handle_size():
    return hws.mlx5dv_hws_rule_get_handle_size()
