# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

from pyverbs.providers.mlx5.hws_context cimport HwsContext
from pyverbs.providers.mlx5.hws_matcher cimport HwsMatcher
from pyverbs.providers.mlx5.hws_actions cimport HwsAction
cimport pyverbs.providers.mlx5.mlx5hws_enums as e
cimport pyverbs.providers.mlx5.libmlx5hws as hws
from pyverbs.pyverbs_error import PyverbsError
from pyverbs.base cimport close_weakrefs
cimport pyverbs.libibverbs as v
import weakref


cdef type_to_str(table_type):
    type_dic = {
        e.MLX5DV_HWS_TABLE_TYPE_NIC_RX: 'MLX5DV_HWS_TABLE_TYPE_NIC_RX',
        e.MLX5DV_HWS_TABLE_TYPE_NIC_TX: 'MLX5DV_HWS_TABLE_TYPE_NIC_TX',
        e.MLX5DV_HWS_TABLE_TYPE_FDB_RX: 'MLX5DV_HWS_TABLE_TYPE_FDB_RX',
        e.MLX5DV_HWS_TABLE_TYPE_FDB_TX: 'MLX5DV_HWS_TABLE_TYPE_FDB_TX',
        e.MLX5DV_HWS_TABLE_TYPE_FDB: 'MLX5DV_HWS_TABLE_TYPE_FDB',
    }
    return type_dic[table_type]


cdef class HwsTableAttr(PyverbsObject):
    def __init__(self, table_type, level, ib_port=None, comp_mask=0):
        """
        Initialize a HwsTableAttr object representing mlx5dv_hws_table_attr C struct.
        :param table_type: Table type.
        :param level: Table level.
        :param ib_port: IB port is valid only for RDMA transport tables.
        :param comp_mask: Comp mask.
        """
        super().__init__()
        self.attr.comp_mask = comp_mask
        self.attr.type = table_type
        self.attr.level = level
        if table_type in [e.MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX,
                          e.MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX]:
            self.attr.ib_port = ib_port


cdef class HwsTableInfo(PyverbsObject):
    def __init__(self, HwsTable table, flags=e.MLX5DV_HWS_METRIC_FLAG_INFO):
        super().__init__()
        if hws.mlx5dv_hws_metric_query_table(table.table, flags, &self.metric):
            raise PyverbsError('Failed query HwsTable info.')

    @property
    def ft_id(self):
        return self.metric.info.ft_id

    @property
    def ft_fw_type(self):
        return self.metric.info.ft_fw_type

    @property
    def level(self):
        return self.metric.info.level

    def __str__(self):
        return f'Flow table id: {hex(self.ft_id)}\n'\
               f'Flow table type: {type_to_str(self.ft_fw_type)}({self.ft_fw_type})\n'\
               f'Flow table level: {self.level}\n'


cdef class HwsTable(PyverbsCM):
    def __init__(self, HwsContext context, HwsTableAttr attr):
        """
        Initialize a HwsTable object representing mlx5dv_hws_table C struct.
        :param context: HwsContext context.
        :param attr: Attributes for creating Mlx5drTable.
        """
        super().__init__()
        self.table = hws.mlx5dv_hws_table_create(context.context,
                                                 <hws.mlx5dv_hws_table_attr *>&(attr.attr))
        if self.table == NULL:
            raise PyverbsError('Failed creating HwsTable.')
        self.hws_context = context
        context.add_ref(self)
        self.hws_actions = weakref.WeakSet()
        self.hws_matchers = weakref.WeakSet()

    cdef add_ref(self, obj):
        if isinstance(obj, HwsAction):
            self.hws_actions.add(obj)
        elif isinstance(obj, HwsMatcher):
            self.hws_matchers.add(obj)
        else:
            raise PyverbsError('Unrecognized object type.')

    def set_default_miss(self, HwsTable miss_tbl=None):
        """
        Set new default miss behaviour to the table.
        :param: miss_tbl: Target (miss) table, or None to remove current miss table.
        """
        rc = hws.mlx5dv_hws_table_set_default_miss(self.table, miss_tbl.table if miss_tbl else NULL)
        if rc:
            raise PyverbsError('Failed to set default miss.')

    def get_info(self, flags=e.MLX5DV_HWS_METRIC_FLAG_INFO):
        return HwsTableInfo(self, flags)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.table != NULL:
            self.logger.debug('Closing HwsTable.')
            close_weakrefs([self.hws_actions, self.hws_matchers])
            rc = hws.mlx5dv_hws_table_destroy(<hws.mlx5dv_hws_table *>(self.table))
            if rc:
                raise PyverbsError('Failed to destroy HwsTable.', rc)
            self.table = NULL
            self.hws_context = None

